/* Linspire Clear KWin window decoration
  Copyright (C) 2004 Lindows, Inc.
            (C) 2004 Sandro Giessl <ceebx@users.sourceforge.net>
   
  based on the KDE style "Plastik":

  Copyright (C) 2003 Sandro Giessl <ceebx@users.sourceforge.net>

  based on the window decoration "Web":
  Copyright (C) 2001 Rik Hemsley (rikkus) <rik@kde.org>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
 */

#include <kconfig.h>

#include "misc.h"
#include "linspireclear.h"
#include "linspireclear.moc"
#include "linspireclearclient.h"

namespace KWinLinspireClear
{

bool LinspireClearHandler::m_animateButtons = true;
bool LinspireClearHandler::m_titleShadow    = true;
bool LinspireClearHandler::m_shrinkBorders  = true;
bool LinspireClearHandler::m_menuClose      = false;
bool LinspireClearHandler::m_reverse        = false;
int  LinspireClearHandler::m_borderSizeSides = 4;
int  LinspireClearHandler::m_borderSizeBottom = 7;
int  LinspireClearHandler::m_titleHeight    = 19;
int  LinspireClearHandler::m_titleHeightTool= 12;
QFont LinspireClearHandler::m_titleFont = QFont();
QFont LinspireClearHandler::m_titleFontTool = QFont();
Qt::AlignmentFlags LinspireClearHandler::m_titleAlign = Qt::AlignHCenter;

LinspireClearHandler::LinspireClearHandler()
{
    // setup pixmap cache...
    m_pixmapCache = new PixmapCache(327680, 2017);
    m_pixmapCache->setAutoDelete(true);

    reset(0);
}

LinspireClearHandler::~LinspireClearHandler()
{
    delete m_pixmapCache;
}

bool LinspireClearHandler::reset(unsigned long changed)
{
    // we assume the active font to be the same as the inactive font since the control
    // center doesn't offer different settings anyways.
    m_titleFont = KDecoration::options()->font(true, false); // not small
    m_titleFontTool = KDecoration::options()->font(true, true); // small

    switch(KDecoration::options()->preferredBorderSize( this )) {
        case BorderLarge:
            m_borderSizeSides = 9;
            m_borderSizeBottom = 13;
            break;
        case BorderVeryLarge:
            m_borderSizeSides = 13;
            m_borderSizeBottom = 17;
            break;
        case BorderHuge:
            m_borderSizeSides = 17;
            m_borderSizeBottom = 21;
            break;
        case BorderVeryHuge:
            m_borderSizeSides = 27;
            m_borderSizeBottom = 30;
            break;
        case BorderOversized:
            m_borderSizeSides = 35;
            m_borderSizeBottom = 40;
            break;
        case BorderNormal:
        default:
            m_borderSizeSides = 4;
            m_borderSizeBottom = 7;
    }

    // check if we are in reverse layout mode
    m_reverse = QApplication::reverseLayout();

    // read in the configuration
    readConfig();

    // Do we need to "hit the wooden hammer" ?
    bool needHardReset = true;
    // TODO: besides the Color and Font settings I can maybe handle more changes
    //       without a hard reset. I will do this later...
    if (changed & SettingColors || changed & SettingFont)
    {
        needHardReset = false;
    }

    if (needHardReset) {
        return true;
    } else {
        resetDecorations(changed);
        return false;
    }
}

KDecoration* LinspireClearHandler::createDecoration( KDecorationBridge* bridge )
{
        return new LinspireClearClient( bridge, this );
}

#if (KDE_VERSION >= KDE_MAKE_VERSION(3,3,89) )
bool LinspireClearHandler::supports( Ability ability )
{
    switch( ability )
    {
        case AbilityAnnounceButtons:
        case AbilityButtonMenu:
        case AbilityButtonOnAllDesktops:
        case AbilityButtonSpacer:
        case AbilityButtonHelp:
        case AbilityButtonMinimize:
        case AbilityButtonMaximize:
        case AbilityButtonClose:
            return true;
        default:
            return false;
    };
}
#endif

void LinspireClearHandler::readConfig()
{
    // create a config object
    KConfig config("kwinlinspireclearrc");
    config.setGroup("General");

    // grab settings
    m_titleShadow    = config.readBoolEntry("TitleShadow", false);

    QFontMetrics fm(m_titleFont);  // active font = inactive font
    int titleHeightMin = 16;
    // The title should strech with bigger font sizes!
    m_titleHeight = QMAX(titleHeightMin, fm.height() + 4); // 4 px for the shadow etc.

    fm = QFontMetrics(m_titleFontTool);  // active font = inactive font
    int titleHeightToolMin = 13;
    // The title should strech with bigger font sizes!
    m_titleHeightTool = QMAX(titleHeightToolMin, fm.height() ); // don't care about the shadow etc.

    QString value = config.readEntry("TitleAlignment", "AlignHCenter");
    if (value == "AlignLeft")         m_titleAlign = Qt::AlignLeft;
    else if (value == "AlignHCenter") m_titleAlign = Qt::AlignHCenter;
    else if (value == "AlignRight")   m_titleAlign = Qt::AlignRight;

    m_animateButtons = config.readBoolEntry("AnimateButtons", true);
    m_menuClose = config.readBoolEntry("CloseOnMenuDoubleClick", true);
}

QColor LinspireClearHandler::getColor(KWinLinspireClear::ColorType type, const bool active)
{
    switch (type) {
        case WindowContour:
            return KDecoration::options()->color(ColorTitleBar, active).dark(190);
        case TitleGradientTo:
            return KDecoration::options()->color(ColorTitleBar, active);
            break;
        case TitleHighlightTop:
        case SideHighlightLeft:
            return alphaBlendColors(KDecoration::options()->color(ColorTitleBar, active),
                    Qt::white, active?150:160);
            break;
        case SideHighlightRight:
        case SideHighlightBottom:
            return alphaBlendColors(KDecoration::options()->color(ColorTitleBar, active),
                    Qt::black, active?150:160);
            break;
        case Border:
            return KDecoration::options()->color(ColorFrame, active);
            break;
        case TitleFont:
            return KDecoration::options()->color(ColorFont, active);
            break;
        case BottomShade:
            return KDecoration::options()->color(ColorTitleBar, active);
            break;
        case Button:
            return KDecoration::options()->color(ColorButtonBg, active);
            break;
        case TitleBlend:
            return KDecoration::options()->color(ColorTitleBlend, active);
            break;
        default:
            return Qt::black;
    }
}

QValueList< LinspireClearHandler::BorderSize >
LinspireClearHandler::borderSizes() const
{
    // the list must be sorted
    return QValueList< BorderSize >() << BorderNormal <<
	BorderLarge << BorderVeryLarge <<  BorderHuge <<
	BorderVeryHuge << BorderOversized;
}

// make the handler accessible to other classes...
static LinspireClearHandler *handler = 0;
LinspireClearHandler* Handler()
{
    return handler;
}

} // KWinLinspireClear

//////////////////////////////////////////////////////////////////////////////
// Plugin Stuff                                                             //
//////////////////////////////////////////////////////////////////////////////

extern "C"
{
    KDecorationFactory *create_factory()
    {
        KWinLinspireClear::handler = new KWinLinspireClear::LinspireClearHandler();
        return KWinLinspireClear::handler;
    }
}
