/* Linspire Clear KWin window decoration
  Copyright (C) 2004 Lindows, Inc.
            (C) 2004 Sandro Giessl <ceebx@users.sourceforge.net>

  based on the KDE style "Plastik":

  Copyright (C) 2003 Sandro Giessl <ceebx@users.sourceforge.net>

  based on the window decoration "Web":
  Copyright (C) 2001 Rik Hemsley (rikkus) <rik@kde.org>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
 */

#include <klocale.h>
#include <kpixmap.h>
#include <kpixmapeffect.h>

#include <qbitmap.h>
#include <qdatetime.h>
#include <qfontmetrics.h>
#include <qimage.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qpainter.h>
#include <qpixmap.h>

#include "linspireclearclient.h"
#include "linspireclearclient.moc"
#include "linspireclearbutton.h"
#include "misc.h"
#include "shadow.h"

// global constants
static const int TOPMARGIN       = 4; // do not change
static const int TOPMARGIN2      = 7; // do not change
static const int TOPMARGIN3      = 4; // do not change
static const int SIDETITLEMARGIN = 10;

// Default button layout
const char default_left[]  = "";
const char default_right[] = "HIAX";

namespace KWinLinspireClear
{

LinspireClearClient::LinspireClearClient(KDecorationBridge* bridge, KDecorationFactory* factory)
    : KDecoration (bridge, factory),
    mainLayout_(0),
    topSpacer_(0), titleSpacer_(0), leftTitleSpacer_(0), rightTitleSpacer_(0),
    decoSpacer_(0), leftSpacer_(0), rightSpacer_(0), bottomSpacer_(0),
    captionBufferDirty(true),
    closing(false),
    s_titleHeight(0),
    s_titleFont(QFont() )
{ }

LinspireClearClient::~LinspireClearClient()
{
    delete aCaptionBuffer;
    delete iCaptionBuffer;

    for (int n=0; n<NumButtons; n++) {
        if (m_button[n]) delete m_button[n];
    }
}

void LinspireClearClient::init()
{
    s_titleHeight = isTool() ?
            LinspireClearHandler::titleHeightTool()
            : LinspireClearHandler::titleHeight();
    s_titleFont = isTool() ?
            LinspireClearHandler::titleFontTool()
            : LinspireClearHandler::titleFont();

    createMainWidget(WNoAutoErase);

    widget()->installEventFilter( this );

    // for flicker-free redraws
    widget()->setBackgroundMode(NoBackground);

    _resetLayout();

    aCaptionBuffer = new QPixmap();
    iCaptionBuffer = new QPixmap();
    captionBufferDirty = true;
    widget()->update(titleSpacer_->geometry());
}

const int SUPPORTED_WINDOW_TYPES_MASK = NET::NormalMask | NET::DesktopMask | NET::DockMask
    | NET::ToolbarMask | NET::MenuMask | NET::DialogMask | NET::OverrideMask | NET::TopMenuMask
    | NET::UtilityMask | NET::SplashMask;

bool LinspireClearClient::isTool()
{
    NET::WindowType type = windowType( SUPPORTED_WINDOW_TYPES_MASK );
    return ((type==NET::Toolbar)||(type==NET::Utility)||(type==NET::Menu));
}

void LinspireClearClient::resizeEvent()
{
    doShape();
    // FIXME: don't update() here! this would result in two paintEvent()s
    // because there is already "something" else triggering the repaint...
//     widget()->update();
}

void LinspireClearClient::paintEvent(QPaintEvent*)
{
    if (captionBufferDirty)
        update_captionBuffer();

    bool active = isActive();

    const QColor& borderColor = LinspireClearHandler::getColor(Border, active);
//     const QColor& bottomShadeColor = LinspireClearHandler::getColor(BottomShade, active);
    const QColor& titleBlendColor = LinspireClearHandler::getColor(TitleBlend, active);
//     const QColor& titleColor = LinspireClearHandler::getColor(TitleGradientTo, active);

    QRect Rtop(topSpacer_->geometry());
    QRect Rtitle(titleSpacer_->geometry());
    QRect Rltitle(leftTitleSpacer_->geometry());
    QRect Rrtitle(rightTitleSpacer_->geometry());
    QRect Rdeco(decoSpacer_->geometry());
    QRect Rleft(leftSpacer_->geometry());
    QRect Rright(rightSpacer_->geometry());
    QRect Rbottom(bottomSpacer_->geometry());
    QRect tempRect;

    QPainter painter(widget() );

    // leftTitleSpacer
    if(Rltitle.width() > 0)
    {
        painter.drawPixmap(Rltitle.left()+1, Rltitle.top(), *getPixmap(TopLeft_Pixmap, active) );
    }

    // Space under the left button group
    painter.drawTiledPixmap(Rltitle.left()+8, Rtitle.top(),
                            Rtitle.left()-1, Rtitle.height()+Rdeco.height(),
                            *getPixmap(TitleBar_Tile_Pixmap, active) );


    // rightTitleSpacer
    if(Rrtitle.width() > 0)
    {
        painter.drawPixmap(Rrtitle.right()-7, Rrtitle.top(), *getPixmap(TopRight_Pixmap, active) );
    }

    // Space under the right button group
    painter.drawTiledPixmap(Rtitle.right()+1, Rtitle.top(),
                            Rrtitle.right()-Rtitle.right()-8, Rtitle.height()+Rdeco.height(),
                            *getPixmap(TitleBar_Tile_Pixmap, active) );

    // topSpacer -- must be painted after left-/rightTitleSpacer because paintings overlap by a few pixels...
    if(Rtop.height() > 0)
    {
        // frame...
        painter.setPen( titleBlendColor );
        painter.drawLine(Rtop.x()+6, Rtop.y(), Rtop.right()-6, Rtop.y());

        painter.drawTiledPixmap(Rtop.left()+2, Rtop.top()+1,
                                Rtop.width()-2*2, Rtop.height()-1,
                                *getPixmap(TitleBar_Tile_Pixmap, active) );

        // edge frame... left. 
        painter.setPen( titleBlendColor );
        painter.drawPoint(Rtop.x()+6, Rtop.y()+0);
        painter.drawPoint(Rtop.x()+5, Rtop.y()+1);
        painter.drawPoint(Rtop.x()+4, Rtop.y()+1);
        painter.drawPoint(Rtop.x()+3, Rtop.y()+2);
        painter.drawPoint(Rtop.x()+2, Rtop.y()+3);
        painter.drawPoint(Rtop.x()+1, Rtop.y()+4);
        painter.drawPoint(Rtop.x()+1, Rtop.y()+5);
        // edge frame... right.
        painter.setPen( titleBlendColor );
        painter.drawPoint(Rtop.right()-6, Rtop.y()+0);
        painter.drawPoint(Rtop.right()-5, Rtop.y()+1);
        painter.drawPoint(Rtop.right()-4, Rtop.y()+1);
        painter.drawPoint(Rtop.right()-3, Rtop.y()+2);
        painter.drawPoint(Rtop.right()-2, Rtop.y()+3);
        painter.drawPoint(Rtop.right()-1, Rtop.y()+4);
        painter.drawPoint(Rtop.right()-1, Rtop.y()+5);
    }

    // titleSpacer
    QPixmap *titleBfrPtr = active ? aCaptionBuffer : iCaptionBuffer;
    if(Rtitle.width() > 0 && titleBfrPtr != 0)
    {
        const int titleMargin = 5; // 5 px between title and buttons

        int tX, tW;
        switch (LinspireClearHandler::titleAlign())
        {
        // AlignCenter
        case Qt::AlignHCenter:
            tX = (titleBfrPtr->width() > Rtitle.width()-2*titleMargin) ?
                    (Rtitle.left()+titleMargin)
                    : Rtitle.left()+(Rtitle.width()- titleBfrPtr->width() )/2;
            tW = (titleBfrPtr->width() > Rtitle.width()-2*titleMargin) ?
                    (Rtitle.width()-2*titleMargin)
                    : titleBfrPtr->width();
            break;
        // AlignRight
        case Qt::AlignRight:
            tX = (titleBfrPtr->width() > Rtitle.width()-2*titleMargin) ?
                    (Rtitle.left()+titleMargin)
                    : Rtitle.right()-titleMargin-titleBfrPtr->width();
            tW = (titleBfrPtr->width() > Rtitle.width()-2*titleMargin) ?
                    (Rtitle.width()-2*titleMargin)
                    : titleBfrPtr->width();
            break;
        // AlignLeft
        default:
            tX = (Rtitle.left()+titleMargin);
            tW = (titleBfrPtr->width() > Rtitle.width()-2*titleMargin) ?
                    (Rtitle.width()-2*titleMargin)
                    : titleBfrPtr->width();
        }

        QRect captionRect(tX, Rtitle.top(), tW, Rtitle.height()+Rdeco.height() );

        if(tW > 0)
        {
            painter.drawTiledPixmap(captionRect,
                                    *titleBfrPtr);
        }

        painter.drawTiledPixmap(Rltitle.right(), captionRect.top(),
                                captionRect.left()-Rltitle.right(), captionRect.height(),
                                *getPixmap(TitleBar_Tile_Pixmap, active) );

        painter.drawTiledPixmap(captionRect.right()+1/*tX+tW*/, captionRect.top(),
                                Rrtitle.left()-(captionRect.right()+1)/* Rtitle.right()-(tX+tW)+1*/, captionRect.height()/* Rtitle.height()+Rdeco.height()*/,
                                *getPixmap(TitleBar_Tile_Pixmap, active) );
    }
    titleBfrPtr = 0;

    // decoSpacer is already drawn previously.

    // leftSpacer
    if(Rleft.width() > 0/* && Rleft.height() > 0*/)
    {
        // frame
        painter.setPen( titleBlendColor );
        painter.drawLine(Rtop.x(), Rtop.y()+6, Rtop.x(), Rbottom.bottom()-6);

        if(Rleft.width() > 1) {
            tempRect.setCoords(Rleft.left()+1, Rleft.top(),
                            Rleft.right(), Rbottom.bottom()-6 );
            painter.fillRect(tempRect, borderColor );
        }
    }

    // rightSpacer
    if(Rright.width() > 0/* && Rright.height() > 0*/)
    {
        // frame
        painter.setPen( titleBlendColor );
        painter.drawLine( Rtop.right(), Rtop.y()+6, Rtop.right(), Rbottom.bottom()-6 );

        if(Rright.width() > 1) {
            tempRect.setCoords(Rright.left(), Rright.top(),
                            Rright.right()-1, Rbottom.bottom()-6 );
            painter.fillRect(tempRect, borderColor );
        }
    }

    // bottomSpacer
    if(Rbottom.height() > 0)
    {
        // frame
        painter.setPen( titleBlendColor );
        painter.drawLine( Rbottom.x()+6, Rbottom.bottom(), Rbottom.right()-6, Rbottom.bottom() );

        int l;
        if(Rleft.width() != 0)
            l = Rbottom.left()+2;
        else
            l = Rbottom.left();
        int r;
        if(Rright.width() != 0)
            r = Rbottom.right()-2;
        else
            r = Rbottom.right();

        painter.drawTiledPixmap(l, Rbottom.top(),
                                r-l, Rbottom.height()-1,
                                *getPixmap(Bottom_Tile_Pixmap, active) );

        // edges...
        painter.drawPixmap(Rbottom.x(), Rbottom.bottom()-10, *getPixmap(BottomLeft_Pixmap, active) );
        painter.drawPixmap(Rbottom.right()-6, Rbottom.bottom()-10, *getPixmap(BottomRight_Pixmap, active) );
    }
}

void LinspireClearClient::mouseDoubleClickEvent(QMouseEvent *e)
{
    if (titleSpacer_->geometry().contains(e->pos()) || decoSpacer_->geometry().contains(e->pos()))
           titlebarDblClickOperation();
}

void LinspireClearClient::doShape()
{
    int w = widget()->width();
    int h = widget()->height();
    int r(w);
    int b(h);

    QRegion mask(0, 0, w, h);

    if(topSpacer_->geometry().height() > 0)
    {
        // Remove top-left corner.
        if(leftTitleSpacer_->geometry().width() > 0)
        {
            mask -= QRegion(0, 0, 6, 1);
            mask -= QRegion(0, 1, 4, 1);
            mask -= QRegion(0, 2, 3, 1);
            mask -= QRegion(0, 3, 2, 1);
            mask -= QRegion(0, 4, 1, 2);
        }
        // Remove top-right corner.
        if(rightTitleSpacer_->geometry().width() > 0)
        {
            mask -= QRegion(r-6, 0, 6, 1);
            mask -= QRegion(r-4, 1, 4, 1);
            mask -= QRegion(r-3, 2, 3, 1);
            mask -= QRegion(r-2, 3, 2, 1);
            mask -= QRegion(r-1, 4, 1, 2);
        }
    }

    // Remove bottom-left corner and bottom-right corner.
    if(bottomSpacer_->geometry().height() > 0)
    {
        mask -= QRegion(0, b  -1, 6, 1);
        mask -= QRegion(0, b-1-1, 4, 1);
        mask -= QRegion(0, b-2-1, 3, 1);
        mask -= QRegion(0, b-3-1, 2, 1);
        mask -= QRegion(0, b-4-2, 1, 2);

        mask -= QRegion(r-6, b  -1, 6, 1);
        mask -= QRegion(r-4, b-1-1, 4, 1);
        mask -= QRegion(r-3, b-2-1, 3, 1);
        mask -= QRegion(r-2, b-3-1, 2, 1);
        mask -= QRegion(r-1, b-4-2, 1, 2);
    }

    setMask( mask );
//     widget()->setMask(mask);
}

void LinspireClearClient::_resetLayout()
{
    // basic layout:
    //  _______________________________________________________________
    // |                         topSpacer                             |
    // |_______________________________________________________________|
    // | leftTitleSpacer | btns | titleSpacer | bts | rightTitleSpacer |
    // |_________________|______|_____________|_____|__________________|
    // |                         decoSpacer                            |
    // |_______________________________________________________________|
    // | |                                                           | |
    // | |                      contentsFake                         | |
    // | |                                                           | |
    // |leftSpacer                                          rightSpacer|
    // |_|___________________________________________________________|_|
    // |                           bottomSpacer                        |
    // |_______________________________________________________________|
    //

    delete mainLayout_;

    delete topSpacer_;
    delete titleSpacer_;
    delete leftTitleSpacer_;
    delete rightTitleSpacer_;
    delete decoSpacer_;
    delete leftSpacer_;
    delete rightSpacer_;
    delete bottomSpacer_;

    mainLayout_ = new QVBoxLayout(widget(), 0, 0);

    topSpacer_        = new QSpacerItem(1, TOPMARGIN, QSizePolicy::Expanding, QSizePolicy::Fixed);
    titleSpacer_      = new QSpacerItem(1, s_titleHeight,
                                        QSizePolicy::Expanding, QSizePolicy::Fixed);
    leftTitleSpacer_  = new QSpacerItem(SIDETITLEMARGIN, s_titleHeight,
                                        QSizePolicy::Fixed, QSizePolicy::Fixed);
    rightTitleSpacer_ = new QSpacerItem(SIDETITLEMARGIN, s_titleHeight,
                                        QSizePolicy::Fixed, QSizePolicy::Fixed);
    decoSpacer_       = new QSpacerItem(1, TOPMARGIN2, QSizePolicy::Expanding, QSizePolicy::Fixed);
    leftSpacer_       = new QSpacerItem(LinspireClearHandler::borderSizeSides(), 1,
                                        QSizePolicy::Fixed, QSizePolicy::Expanding);
    rightSpacer_      = new QSpacerItem(LinspireClearHandler::borderSizeSides(), 1,
                                        QSizePolicy::Fixed, QSizePolicy::Expanding);
    bottomSpacer_     = new QSpacerItem(1, LinspireClearHandler::borderSizeBottom(),
                                        QSizePolicy::Expanding, QSizePolicy::Fixed);

    // top
    mainLayout_->addItem(topSpacer_);

    // title
    QHBoxLayout *titleLayout_ = new QHBoxLayout(mainLayout_, 0, 0);

    // sizeof(...) is calculated at compile time
    memset(m_button, 0, sizeof(LinspireClearButton *) * NumButtons);

    int buttonSize = s_titleHeight;
    if (s_titleHeight > 14)
        buttonSize -= 2;
    else if (s_titleHeight > 11)
        buttonSize -= 1;

    titleLayout_->addItem(LinspireClearHandler::reverseLayout()?rightTitleSpacer_:leftTitleSpacer_);
    addButtons(titleLayout_,
               options()->customButtonPositions() ? options()->titleButtonsLeft() : QString(default_left),
               buttonSize );
    titleLayout_->addItem(titleSpacer_);
    addButtons(titleLayout_,
               options()->customButtonPositions() ? options()->titleButtonsRight() : QString(default_right),
               buttonSize );
    titleLayout_->addItem(LinspireClearHandler::reverseLayout()?leftTitleSpacer_:rightTitleSpacer_);

    // deco
    mainLayout_->addItem(decoSpacer_);

    //Mid
    QHBoxLayout * midLayout   = new QHBoxLayout(mainLayout_, 0, 0);
    midLayout->addItem(LinspireClearHandler::reverseLayout()?rightSpacer_:leftSpacer_);
    if( isPreview())
        midLayout->addWidget(new QLabel( i18n( "<center><b>Linspire Clear preview</b></center>" ), widget()) );
    else
        midLayout->addItem( new QSpacerItem( 0, 0 ));
    midLayout->addItem(LinspireClearHandler::reverseLayout()?leftSpacer_:rightSpacer_);

    //Bottom
    mainLayout_->addItem(bottomSpacer_);


}

void LinspireClearClient::addButtons(QBoxLayout *layout, const QString& s, int buttonSize)
{
    if (s.length() > 0) {
        for (unsigned n=0; n < s.length(); n++) {
            bool added = false; // whether the button actually has been added, used for spacer calculation
            switch (s[n]) {
              case 'M': // Menu button
                  if (!m_button[MenuButton]){
                      m_button[MenuButton] = new LinspireClearButton(this, "menu", i18n("Menu"), MenuButton, buttonSize, LeftButton|RightButton);
                      connect(m_button[MenuButton], SIGNAL(pressed()), SLOT(menuButtonPressed()));
                      connect(m_button[MenuButton], SIGNAL(released()), this, SLOT(menuButtonReleased()));
                      layout->addWidget(m_button[MenuButton], 0, Qt::AlignHCenter | Qt::AlignTop);
                      added = true;
                  }
                  break;
              case 'S': // OnAllDesktops button
                  if (!m_button[OnAllDesktopsButton]){
                      const bool oad = isOnAllDesktops();
                      m_button[OnAllDesktopsButton] = new LinspireClearButton(this, "on_all_desktops", oad?i18n("Not On All Desktops"):i18n("On All Desktops"), OnAllDesktopsButton, buttonSize);
                      m_button[OnAllDesktopsButton]->setOnAllDesktops( oad );
//                       m_button[OnAllDesktopsButton]->setDeco(); // update deco...
                      connect(m_button[OnAllDesktopsButton], SIGNAL(clicked()), SLOT(toggleOnAllDesktops()));
                      layout->addWidget(m_button[OnAllDesktopsButton], 0, Qt::AlignHCenter | Qt::AlignTop);
                      added = true;
                  }
                  break;
              case 'H': // Help button
                  if ((!m_button[HelpButton]) && providesContextHelp()){
                      m_button[HelpButton] = new LinspireClearButton(this, "help", i18n("Help"), HelpButton, buttonSize);
                      connect(m_button[HelpButton], SIGNAL(clicked()), SLOT(showContextHelp()));
                      layout->addWidget(m_button[HelpButton], 0, Qt::AlignHCenter | Qt::AlignTop);
                      added = true;
                  }
                  break;
              case 'I': // Minimize button
                  if ((!m_button[MinButton]) && isMinimizable()){
                      m_button[MinButton] = new LinspireClearButton(this, "minimize", i18n("Minimize"), MinButton, buttonSize);
                      connect(m_button[MinButton], SIGNAL(clicked()), SLOT(minimize()));
                      layout->addWidget(m_button[MinButton], 0, Qt::AlignHCenter | Qt::AlignTop);
                      added = true;
                  }
                  break;
              case 'A': // Maximize button
                  if ((!m_button[MaxButton]) && isMaximizable()){
                      const bool max = maximizeMode()!=MaximizeRestore;
                      m_button[MaxButton] = new LinspireClearButton(this, "maximize", max?i18n("Restore"):i18n("Maximize"), MaxButton, buttonSize, LeftButton|MidButton|RightButton);
                      m_button[MaxButton]->setMaximized( max );
//                       m_button[MaxButton]->setDeco(); // update deco...
                      connect(m_button[MaxButton], SIGNAL(clicked()), SLOT(slotMaximize()));
                      layout->addWidget(m_button[MaxButton], 0, Qt::AlignHCenter | Qt::AlignTop);
                      added = true;
                  }
                  break;
              case 'X': // Close button
                  if ((!m_button[CloseButton]) && isCloseable()){
                      m_button[CloseButton] = new LinspireClearButton(this, "close", i18n("Close"), CloseButton, buttonSize);
                      connect(m_button[CloseButton], SIGNAL(clicked()), SLOT(closeWindow()));
                      layout->addWidget(m_button[CloseButton], 0, Qt::AlignHCenter | Qt::AlignTop);
                      added = true;
                  }
                  break;
              case '_': // Spacer item
                  layout->addSpacing(3); // add a 3 px spacing...
                  added = true;
            }

             // add 1 px spacing between buttons
            if(added && n < (s.length()-1) ) // and only between them!
                layout->addSpacing (2);
        }
    }
}

void LinspireClearClient::captionChange()
{
    captionBufferDirty = true;
    widget()->update(titleSpacer_->geometry());
}

void LinspireClearClient::reset( unsigned long changed )
{
    if (changed & SettingColors)
    {
        // repaint the whole thing
        captionBufferDirty = true;
        widget()->update();
        for (int n=0; n<NumButtons; n++) {
            if (m_button[n]) m_button[n]->update();
        }
    } else if (changed & SettingFont) {
        // font has changed -- update title height and font
        s_titleHeight = isTool() ?
                LinspireClearHandler::titleHeightTool()
                : LinspireClearHandler::titleHeight();
        s_titleFont = isTool() ?
                LinspireClearHandler::titleFontTool()
                : LinspireClearHandler::titleFont();
        // update buttons
        for (int n=0; n<NumButtons; n++) {
            if (m_button[n]) m_button[n]->setSize(s_titleHeight-1);
        }
        // update the spacer
        titleSpacer_->changeSize(1, s_titleHeight,
                QSizePolicy::Expanding, QSizePolicy::Fixed);
        // then repaint
        captionBufferDirty = true;
        widget()->update();
    }
}

LinspireClearClient::Position LinspireClearClient::mousePosition(const QPoint &point) const
{
    const int corner = 18+3*LinspireClearHandler::borderSizeSides()/2;
    Position pos = PositionCenter;

    // often needed coords..
    QRect topRect(topSpacer_->geometry());
    QRect decoRect(decoSpacer_->geometry());
    QRect leftRect(leftSpacer_->geometry());
    QRect leftTitleRect(leftTitleSpacer_->geometry());
    QRect rightRect(rightSpacer_->geometry());
    QRect rightTitleRect(rightTitleSpacer_->geometry());
    QRect bottomRect(bottomSpacer_->geometry());

    if(bottomRect.contains(point)) {
        if      (point.x() <= bottomRect.left()+corner)  pos = PositionBottomLeft;
        else if (point.x() >= bottomRect.right()-corner) pos = PositionBottomRight;
        else                                             pos = PositionBottom;
    }
    else if(leftRect.contains(point)) {
        if      (point.y() <= topRect.top()+corner)       pos = PositionTopLeft;
        else if (point.y() >= bottomRect.bottom()-corner) pos = PositionBottomLeft;
        else                                              pos = PositionLeft;
    }
    else if(leftTitleRect.contains(point)) {
        if      (point.y() <= topRect.top()+corner)       pos = PositionTopLeft;
        else                                              pos = PositionLeft;
    }
    else if(rightRect.contains(point)) {
        if      (point.y() <= topRect.top()+corner)       pos = PositionTopRight;
        else if (point.y() >= bottomRect.bottom()-corner) pos = PositionBottomRight;
        else                                              pos = PositionRight;
    }
    else if(rightTitleRect.contains(point)) {
        if      (point.y() <= topRect.top()+corner)       pos = PositionTopRight;
        else                                              pos = PositionRight;
    }
    else if(topRect.contains(point)) {
        if      (point.x() <= topRect.left()+corner)      pos = PositionTopLeft;
        else if (point.x() >= topRect.right()-corner)     pos = PositionTopRight;
        else                                              pos = PositionTop;
    }
    else if(decoRect.contains(point)) {
        if(point.x() <= leftTitleRect.right()) {
            if(point.y() <= topRect.top()+corner)         pos = PositionTopLeft;
            else                                          pos = PositionLeft;
        }
        else if(point.x() >= rightTitleRect.left()) {
            if(point.y() <= topRect.top()+corner)         pos = PositionTopRight;
            else                                          pos = PositionRight;
        }
    }

    return pos;
}

void LinspireClearClient::iconChange()
{
    // redraw icon in the title...
    captionBufferDirty = true;
    widget()->update(titleSpacer_->geometry());

//     if (m_button[MenuButton])
//     {
//         m_button[MenuButton]->update();
//     }
}

void LinspireClearClient::activeChange()
{
    for (int n=0; n<NumButtons; n++)
        if (m_button[n]) m_button[n]->update();
    widget()->update();
}

void LinspireClearClient::maximizeChange()
{
    if( m_button[MaxButton] ) {
        m_button[MaxButton]->setMaximized( maximizeMode()!=MaximizeRestore);
        m_button[MaxButton]->setTipText( (maximizeMode()==MaximizeRestore) ?
                i18n("Maximize")
                : i18n("Restore"));
//         m_button[MaxButton]->setDeco(); // update the button icon...
    }

    widget()->update();
}

void LinspireClearClient::desktopChange()
{
    if ( m_button[OnAllDesktopsButton] ) {
        m_button[OnAllDesktopsButton]->setOnAllDesktops( isOnAllDesktops() );
        m_button[OnAllDesktopsButton]->setTipText( isOnAllDesktops() ?
                i18n("Not On All Desktops")
                : i18n("On All Desktops"));
//         m_button[OnAllDesktopsButton]->setDeco(); // update icon/deco
    }
}

void LinspireClearClient::slotMaximize()
{
    if (m_button[MaxButton])
    {
#if (KDE_VERSION >= KDE_MAKE_VERSION(3,3,0) ) // I think this requires at least KDE 3.3
        maximize(m_button[MaxButton]->lastMousePress());
#else
        switch (m_button[MaxButton]->lastMousePress())
        {
          case MidButton:
              maximize(maximizeMode() ^ MaximizeVertical );
              break;
          case RightButton:
              maximize(maximizeMode() ^ MaximizeHorizontal );
              break;
          default:
              maximize(maximizeMode() == MaximizeFull ? MaximizeRestore : MaximizeFull );
        }
#endif
        doShape();
    }
}

void LinspireClearClient::menuButtonPressed()
{
    static QTime* t = NULL;
    static LinspireClearClient* lastClient = NULL;
    if (t == NULL)
        t = new QTime;
    bool dbl = (lastClient==this && t->elapsed() <= QApplication::doubleClickInterval());
    lastClient = this;
    t->start();
    if (!dbl || !LinspireClearHandler::menuClose()) {
        QPoint pos = m_button[MenuButton]->mapToGlobal(m_button[MenuButton]->rect().bottomLeft() );
        KDecorationFactory* f = factory();
        showWindowMenu( pos );
        if( !f->exists( this )) // 'this' was deleted
            return;
        m_button[MenuButton]->setDown(false);
    }
    else
        closing = true;
}

void LinspireClearClient::menuButtonReleased()
{
    if(closing)
        closeWindow();
}

void LinspireClearClient::renderSoftBlend(QPainter *p, const QColor &bg, const QColor &fg, int x, int y) const
{
    p->setPen(alphaBlendColors(bg, fg, 245) );
    p->drawPoint(x, y);
    p->setPen(alphaBlendColors(bg, fg, 205) );
    p->drawPoint(x, y+1);
    p->setPen(alphaBlendColors(bg, fg, 165) );
    p->drawPoint(x, y+2);
    p->setPen(alphaBlendColors(bg, fg, 125) );
    p->drawPoint(x, y+3);
}

QPixmap* LinspireClearClient::getPixmap(ClientPixmapType type, bool active) const
{
    PixmapCache& cache = Handler()->pixmapCache();

    const int bgTopHeight = s_titleHeight/3; // height of the frame color-area...

    QPainter painter;

    switch(type) {
        case Top_Tile_Pixmap: {
            QColor borderHighlightColor = active ? LinspireClearHandler::getColor(Border, active).light(102)
                                                 : LinspireClearHandler::getColor(Border, active);
            int height = TOPMARGIN - 1;

            CacheEntry search( cTopTile, 0, height, borderHighlightColor.rgb() );
            int key = search.key();
    
            CacheEntry *cacheEntry;
            if( (cacheEntry = cache.find(key)) ) {
                if( search == *cacheEntry ) { // match! we can return it now...
                    if(cacheEntry->pixmap)
                        return cacheEntry->pixmap;
                } else {
                    // Remove old entry in case of a conflict!
                    // This shouldn't happen very often, see comment in CacheEntry.
                    cache.remove(key);
                }
            }

            QPixmap *result = new QPixmap(1, height);
            painter.begin(result);
            painter.setPen(borderHighlightColor);
            painter.drawLine(0,0, 0,height);
            painter.end();

            // insert into cache using the previously created key.
            CacheEntry *toAdd = new CacheEntry(search);
            toAdd->pixmap = result;
            cache.insert( key, toAdd, result->width()*result->height()*result->depth()/8 );

            return result;
        }

        case TitleBar_Tile_Pixmap: {
            QColor borderHighlightColor = active ? LinspireClearHandler::getColor(Border, active).light(102)
                                                 : LinspireClearHandler::getColor(Border, active);
            QColor borderColor = LinspireClearHandler::getColor(Border, active);
            QColor titleColor = LinspireClearHandler::getColor(TitleGradientTo, active);
            int height = s_titleHeight + TOPMARGIN2;

            CacheEntry search( cTitleBarTile, 0, height, borderColor.rgb(), titleColor.rgb(), borderHighlightColor.rgb() );
            int key = search.key();
            CacheEntry *cacheEntry;
            if( (cacheEntry = cache.find(key)) ) {
                if( search == *cacheEntry ) { // match! we can return it now...
                    if(cacheEntry->pixmap)
                        return cacheEntry->pixmap;
                } else {
                    // Remove old entry in case of a conflict!
                    // This shouldn't happen very often, see comment in CacheEntry.
                    cache.remove(key);
                }
            }

            KPixmap tempPixmap;
            QPixmap *result = new QPixmap(1, height);
            painter.begin(result);
            // the color gradient which "overlapps" the "frame color-area"...
            tempPixmap.resize(1, height - bgTopHeight - TOPMARGIN3 );
            KPixmapEffect::gradient(tempPixmap,
                                    titleColor, alphaBlendColors(borderColor,titleColor,220),
                                    KPixmapEffect::VerticalGradient);
            painter.drawPixmap(0, bgTopHeight+1, tempPixmap);
            // small title bar definition...
            tempPixmap.resize(1, TOPMARGIN3 );
            KPixmapEffect::gradient(tempPixmap,
                                    alphaBlendColors(borderColor,titleColor,142), borderColor,
                                    KPixmapEffect::VerticalGradient);
            painter.drawPixmap(0, height-TOPMARGIN, tempPixmap);
            // the light top of the titlebar
            painter.setPen(borderHighlightColor );
            painter.drawLine(0,0, 0, bgTopHeight);
            // the anti-alias between the overlapping thing and the background...
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 0, bgTopHeight+1-3);
            painter.end();

            // insert into cache using the previously created key.
            CacheEntry *toAdd = new CacheEntry(search);
            toAdd->pixmap = result;
            cache.insert( key, toAdd, result->width()*result->height()*result->depth()/8 );

            return result;
        }

        case Bottom_Tile_Pixmap: {
            QColor borderColor = LinspireClearHandler::getColor(Border, active);
            QColor shadeColor = LinspireClearHandler::getColor(BottomShade, active);
            int height = LinspireClearHandler::borderSizeBottom()-1;

            CacheEntry search( cBottomTile, 0, height, borderColor.rgb(), shadeColor.rgb() );
            int key = search.key();
            CacheEntry *cacheEntry;
            if( (cacheEntry = cache.find(key)) ) {
                if( search == *cacheEntry ) { // match! we can return it now...
                    if(cacheEntry->pixmap)
                        return cacheEntry->pixmap;
                } else {
                    // Remove old entry in case of a conflict!
                    // This shouldn't happen very often, see comment in CacheEntry.
                    cache.remove(key);
                }
            }

            QPixmap *result = new QPixmap(1, height);
            painter.begin(result);

            // end of the frame colored part...
            painter.setPen(borderColor );
            painter.drawLine(0,0, 0, height-5);
            // anti alias...
            painter.setPen(alphaBlendColors(borderColor, shadeColor, 150) );
            painter.drawPoint(0,height-4);
            // the shaded thing...
            painter.setPen(shadeColor);
            painter.drawPoint(0, height-3);
            painter.setPen(shadeColor.dark(105));
            painter.drawPoint(0, height-2);
            painter.setPen(shadeColor.dark(110));
            painter.drawPoint(0, height-1);
            painter.end();

            // insert into cache using the previously created key.
            CacheEntry *toAdd = new CacheEntry(search);
            toAdd->pixmap = result;
            cache.insert( key, toAdd, result->width()*result->height()*result->depth()/8 );

            return result;
        }

        case TopLeft_Pixmap: {
            QColor borderHighlightColor = active ? LinspireClearHandler::getColor(Border, active).light(102)
                                                 : LinspireClearHandler::getColor(Border, active);
            QColor borderColor = LinspireClearHandler::getColor(Border, active);
            QColor titleColor = LinspireClearHandler::getColor(BottomShade, active);
            int height = s_titleHeight + TOPMARGIN2;

            CacheEntry search( cTopLeft, 0, height, borderColor.rgb(), titleColor.rgb(), borderHighlightColor.rgb() );
            int key = search.key();
            CacheEntry *cacheEntry;
            if( (cacheEntry = cache.find(key)) ) {
                if( search == *cacheEntry ) { // match! we can return it now...
                    if(cacheEntry->pixmap)
                        return cacheEntry->pixmap;
                } else {
                    // Remove old entry in case of a conflict!
                    // This shouldn't happen very often, see comment in CacheEntry.
                    cache.remove(key);
                }
            }

            KPixmap tempPixmap;
            QPixmap *result = new QPixmap(7, height);
            painter.begin(result);

            // fill with titlebar background...
            painter.drawTiledPixmap(result->rect(), *getPixmap(TitleBar_Tile_Pixmap, active) );
            // clear background areas...
            painter.setPen(borderHighlightColor );
            painter.drawLine(0, bgTopHeight-2, 0, bgTopHeight+6);
            painter.drawLine(1, bgTopHeight-2, 1, bgTopHeight+4);
            painter.drawLine(2, bgTopHeight-2, 2, bgTopHeight+3);
            painter.drawLine(3, bgTopHeight-2, 3, bgTopHeight+2);
            painter.drawLine(4, bgTopHeight-2, 4, bgTopHeight+1);
            // soft gradient...
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 1, bgTopHeight+2);
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 2, bgTopHeight-0);
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 3, bgTopHeight-1);
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 4, bgTopHeight-1);
            // left anti-alias...
            tempPixmap.resize(1, height - bgTopHeight - 7);
            KPixmapEffect::gradient(tempPixmap,
                        alphaBlendColors( borderHighlightColor, titleColor, 180),
                        borderColor,
                        KPixmapEffect::VerticalGradient);
            painter.drawPixmap(0, bgTopHeight+7, tempPixmap);
            painter.end();

            // insert into cache using the previously created key.
            CacheEntry *toAdd = new CacheEntry(search);
            toAdd->pixmap = result;
            cache.insert( key, toAdd, result->width()*result->height()*result->depth()/8 );

            return result;
        }

        case TopRight_Pixmap: {
            QColor borderHighlightColor = active ? LinspireClearHandler::getColor(Border, active).light(102)
                                                 : LinspireClearHandler::getColor(Border, active);
            QColor borderColor = LinspireClearHandler::getColor(Border, active);
            QColor titleColor = LinspireClearHandler::getColor(BottomShade, active);
            int height = s_titleHeight + TOPMARGIN2;

            CacheEntry search( cTopRight, 0, height, borderColor.rgb(), titleColor.rgb(), borderHighlightColor.rgb() );
            int key = search.key();
            CacheEntry *cacheEntry;
            if( (cacheEntry = cache.find(key)) ) {
                if( search == *cacheEntry ) { // match! we can return it now...
                    if(cacheEntry->pixmap)
                        return cacheEntry->pixmap;
                } else {
                    // Remove old entry in case of a conflict!
                    // This shouldn't happen very often, see comment in CacheEntry.
                    cache.remove(key);
                }
            }

            KPixmap tempPixmap;
            QPixmap *result = new QPixmap(7, height);
            painter.begin(result);

            // * same as TopLeft, but the "6-" stuff reverses the paintings.

            // fill with titlebar background...
            painter.drawTiledPixmap(result->rect(), *getPixmap(TitleBar_Tile_Pixmap, active) );
            // clear background areas...
            painter.setPen(borderHighlightColor );
            painter.drawLine(6-0, bgTopHeight-2, 6-0, bgTopHeight+6);
            painter.drawLine(6-1, bgTopHeight-2, 6-1, bgTopHeight+4);
            painter.drawLine(6-2, bgTopHeight-2, 6-2, bgTopHeight+3);
            painter.drawLine(6-3, bgTopHeight-2, 6-3, bgTopHeight+2);
            painter.drawLine(6-4, bgTopHeight-2, 6-4, bgTopHeight+1);
            // soft gradient...
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 6-1, bgTopHeight+2);
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 6-2, bgTopHeight-0);
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 6-3, bgTopHeight-1);
            renderSoftBlend(&painter, borderHighlightColor, titleColor, 6-4, bgTopHeight-1);
            // left anti-alias...
            tempPixmap.resize(1, height - bgTopHeight - 7);
            KPixmapEffect::gradient(tempPixmap,
                        alphaBlendColors( borderHighlightColor, titleColor, 180),
                        borderColor,
                        KPixmapEffect::VerticalGradient);
            painter.drawPixmap(6-0, bgTopHeight+7, tempPixmap);
            painter.end();

            // insert into cache using the previously created key.
            CacheEntry *toAdd = new CacheEntry(search);
            toAdd->pixmap = result;
            cache.insert( key, toAdd, result->width()*result->height()*result->depth()/8 );

            return result;
        }

        case BottomLeft_Pixmap: {
            QColor borderColor = LinspireClearHandler::getColor(Border, active);
            QColor shadeColor = LinspireClearHandler::getColor(BottomShade, active);
            QColor blendColor = LinspireClearHandler::getColor(TitleBlend, active);

            CacheEntry search( cBottomLeft, 0, 0, borderColor.rgb(), shadeColor.rgb(), blendColor.rgb() );
            int key = search.key();
            CacheEntry *cacheEntry;
            if( (cacheEntry = cache.find(key)) ) {
                if( search == *cacheEntry ) { // match! we can return it now...
                    if(cacheEntry->pixmap)
                        return cacheEntry->pixmap;
                } else {
                    // Remove old entry in case of a conflict!
                    // This shouldn't happen very often, see comment in CacheEntry.
                    cache.remove(key);
                }
            }

            int height = 10,
                width = 7;
            KPixmap tempPixmap;

            QPixmap *result = new QPixmap(width, height);
            painter.begin(result);

            QPixmap *bt = getPixmap(Bottom_Tile_Pixmap, active);
            painter.drawTiledPixmap(0, height - bt->height()/*4*/, 7, bt->height()/*10*/, *bt );
            // frame...
            painter.setPen( blendColor );
            painter.drawLine(0, 0, 0, 4);
            painter.drawLine(9, 6, 9, 7);
            painter.drawPoint(1, 5);
            painter.drawPoint(1, 6);
            painter.drawPoint(2, 7);
            painter.drawPoint(3, 8);
            painter.drawPoint(4, 9);
            painter.drawPoint(5, 9);
            // rest of the panel...
            painter.setPen( borderColor );
            painter.drawLine(1, 0, 6, 0);
            painter.drawLine(2, 1, 6, 1);
            painter.drawLine(2, 2, 6, 2);
            painter.drawLine(3, 3, 6, 3);
            // 1st part from the bottom shade...
            painter.setPen( shadeColor );
            painter.drawPoint(5, 6);
            painter.drawPoint(4, 6);
            painter.drawPoint(3, 5);
            painter.drawPoint(2, 4);
            painter.drawPoint(1, 3);
            // 2nd part from the bottom shade...
            painter.setPen( shadeColor.dark(105) );
            painter.drawPoint(2, 5);
            painter.drawPoint(3, 6);
            // 3rd part from the bottom shade...
            painter.setPen( shadeColor.dark(110) );
            painter.drawPoint(5, 8);
            painter.drawPoint(4, 8);
            painter.drawPoint(3, 7);
            painter.drawPoint(2, 6);
            painter.drawPoint(1, 4);
            // part from the bottom shade anti-alias...
            painter.setPen( alphaBlendColors( borderColor, shadeColor, 190) );
            painter.drawPoint( 5, 5);
            painter.drawPoint( 4, 5);
            painter.drawPoint( 3, 4);
            painter.drawPoint( 2, 3);
            painter.drawPoint( 1, 2);
            painter.drawPoint( 1, 1);
            painter.end();

            // insert into cache using the previously created key.
            CacheEntry *toAdd = new CacheEntry(search);
            toAdd->pixmap = result;
            cache.insert( key, toAdd, result->width()*result->height()*result->depth()/8 );

            return result;
        }

        default: // case default should never be happening, but to make the compiler happy...
        case BottomRight_Pixmap: {
            QColor borderColor = LinspireClearHandler::getColor(Border, active);
            QColor shadeColor = LinspireClearHandler::getColor(BottomShade, active);
            QColor blendColor = LinspireClearHandler::getColor(TitleBlend, active);

            CacheEntry search( cBottomRight, 0, 0, borderColor.rgb(), shadeColor.rgb(), blendColor.rgb() );
            int key = search.key();
            CacheEntry *cacheEntry;
            if( (cacheEntry = cache.find(key)) ) {
                if( search == *cacheEntry ) { // match! we can return it now...
                    if(cacheEntry->pixmap)
                        return cacheEntry->pixmap;
                } else {
                    // Remove old entry in case of a conflict!
                    // This shouldn't happen very often, see comment in CacheEntry.
                    cache.remove(key);
                }
            }

            int height = 10,
                width = 7;
            KPixmap tempPixmap;

            QPixmap *result = new QPixmap(width, height);
            painter.begin(result);

            // * almost the same as TopLeft, but the "6-" stuff reverses the paintings.

            QPixmap *bt = getPixmap(Bottom_Tile_Pixmap, active);
            painter.drawTiledPixmap(0, height - bt->height()/*4*/, 7, bt->height()/*10*/, *getPixmap(Bottom_Tile_Pixmap, active) );
            // frame...
            painter.setPen( blendColor );
            painter.drawLine(6-0, 0, 6-0, 4);
            painter.drawLine(6-9, 6, 6-9, 7);
            painter.drawPoint(6-1, 5);
            painter.drawPoint(6-1, 6);
            painter.drawPoint(6-2, 7);
            painter.drawPoint(6-3, 8);
            painter.drawPoint(6-4, 9);
            painter.drawPoint(6-5, 9);
            // rest of the panel...
            painter.setPen( borderColor );
            painter.drawLine(6-1, 0, 6-6, 0);
            painter.drawLine(6-2, 1, 6-6, 1);
            painter.drawLine(6-2, 2, 6-6, 2);
            painter.drawLine(6-3, 3, 6-6, 3);
            // 1st part from the bottom shade...
            painter.setPen( shadeColor );
            painter.drawPoint(6-5, 6);
            painter.drawPoint(6-4, 6);
            painter.drawPoint(6-3, 5);
            painter.drawPoint(6-2, 4);
            painter.drawPoint(6-1, 3);
            // 2nd part from the bottom shade...
            painter.setPen( shadeColor.dark(105) );
            painter.drawPoint(6-2, 5);
            painter.drawPoint(6-3, 6);
            // 3rd part from the bottom shade...
            painter.setPen( shadeColor.dark(110) );
            painter.drawPoint(6-5, 8);
            painter.drawPoint(6-4, 8);
            painter.drawPoint(6-3, 7);
            painter.drawPoint(6-2, 6);
            painter.drawPoint(6-1, 4);
            // part from the bottom shade anti-alias...
            painter.setPen( alphaBlendColors( borderColor, shadeColor, 190) );
            painter.drawPoint( 6-5, 5);
            painter.drawPoint( 6-4, 5);
            painter.drawPoint( 6-3, 4);
            painter.drawPoint( 6-2, 3);
            painter.drawPoint( 6-1, 2);
            painter.drawPoint( 6-1, 1);
            painter.end();

            // insert into cache using the previously created key.
            CacheEntry *toAdd = new CacheEntry(search);
            toAdd->pixmap = result;
            cache.insert( key, toAdd, result->width()*result->height()*result->depth()/8 );

            return result;
        }
    }
}

void LinspireClearClient::update_captionBuffer()
{
    const uint maxCaptionLength = 300; // truncate captions longer than this!
    QString c(caption() );
    if (c.length() > maxCaptionLength) {
        c.truncate(maxCaptionLength);
        c.append(" [...]");
    }

    QFontMetrics fm(s_titleFont);
    int captionWidth  = fm.width(c);
    int captionHeight = fm.height();

    // used for painting the text
    const int iconSpace = captionHeight+5; // space for the icon...
    QRect textRect(iconSpace,0, captionWidth+2*2,s_titleHeight); // 2*2 px: shadow;

    QPixmap textPixmap;
    QPainter painter;
    if(LinspireClearHandler::titleShadow())
    {
        // prepare the shadow
        textPixmap = QPixmap( textRect.size() ); // 2*2 px shadow space
        textPixmap.fill(QColor(0,0,0));
        textPixmap.setMask( textPixmap.createHeuristicMask(TRUE) );
        painter.begin(&textPixmap);
        painter.setFont(s_titleFont);
        painter.setPen(white);
        painter.drawText(textPixmap.rect(), AlignCenter, c );
        painter.end();
    }

    QImage shadow;
    ShadowEngine se;

    // setup the icon
    KPixmap menuIcon( icon().pixmap( QIconSet::Small, QIconSet::Normal) );
    // set icon to the same size as the text...
    if( menuIcon.height() != captionHeight ) {
        menuIcon.convertFromImage( menuIcon.convertToImage().smoothScale(captionHeight,captionHeight));
    }

    // active
    aCaptionBuffer->resize(iconSpace+textRect.width(), s_titleHeight+decoSpacer_->geometry().height() );
    painter.begin(aCaptionBuffer);
    painter.drawTiledPixmap(aCaptionBuffer->rect(), *getPixmap(TitleBar_Tile_Pixmap, true) );
    // draw icon alongside text
    painter.drawPixmap( 0, (textRect.height()-menuIcon.height())/2, menuIcon );
    if(LinspireClearHandler::titleShadow())
    {
        shadow = se.makeShadow(textPixmap, QColor(0, 0, 0));
        painter.drawImage(textRect.x()+1, textRect.y()+1, shadow);
    }
    painter.setFont(s_titleFont);
    painter.setPen(LinspireClearHandler::getColor(TitleFont,true));
    painter.drawText(textRect, AlignCenter, c );
    painter.end();


    // inactive
    iCaptionBuffer->resize(iconSpace+textRect.width(), s_titleHeight+decoSpacer_->geometry().height() );
    painter.begin(iCaptionBuffer);
    painter.drawTiledPixmap(iCaptionBuffer->rect(), *getPixmap(TitleBar_Tile_Pixmap, false) );
    // draw icon alongside text
    painter.drawPixmap( 0, (textRect.height()-menuIcon.height())/2, KPixmapEffect::toGray(menuIcon ,true) );
    if(LinspireClearHandler::titleShadow())
    {
        painter.drawImage(textRect.x()+1, textRect.y()+1, shadow);
    }
    painter.setFont(s_titleFont);
    painter.setPen(LinspireClearHandler::getColor(TitleFont,false));
    painter.drawText(textRect, AlignCenter, c );
    painter.end();

    captionBufferDirty = false;
}

void LinspireClearClient::borders( int& left, int& right, int& top, int& bottom ) const
{
    // there's something very strange going on with the given references (sometimes top and bottom
    // seem to be the same ref). little work around using normal variables so at least the spacers
    // get the right size...
    int l, r, t, b;

    if ((maximizeMode()==MaximizeFull) && !options()->moveResizeMaximizedWindows()) {
        l = r = b = 0;
        t = s_titleHeight + TOPMARGIN2;

        // update layout etc.
        topSpacer_->changeSize(1, 0, QSizePolicy::Expanding, QSizePolicy::Fixed);
        leftSpacer_->changeSize(l, 1, QSizePolicy::Fixed, QSizePolicy::Expanding);
        leftTitleSpacer_->changeSize(l, s_titleHeight, QSizePolicy::Fixed, QSizePolicy::Fixed);
        rightSpacer_->changeSize(r, 1, QSizePolicy::Fixed, QSizePolicy::Expanding);
        rightTitleSpacer_->changeSize(r, s_titleHeight, QSizePolicy::Fixed, QSizePolicy::Fixed);
        bottomSpacer_->changeSize(1, b, QSizePolicy::Expanding, QSizePolicy::Fixed);
    } else {
        l = r = LinspireClearHandler::borderSizeSides();
        b = LinspireClearHandler::borderSizeBottom();
        t = s_titleHeight + TOPMARGIN2 + TOPMARGIN;

        // update layout etc.
        topSpacer_->changeSize(1, TOPMARGIN, QSizePolicy::Expanding, QSizePolicy::Fixed);
        leftSpacer_->changeSize(l, 1, QSizePolicy::Fixed, QSizePolicy::Expanding);
        leftTitleSpacer_->changeSize(SIDETITLEMARGIN, s_titleHeight,
                QSizePolicy::Fixed, QSizePolicy::Fixed);
        rightSpacer_->changeSize(r, 1, QSizePolicy::Fixed, QSizePolicy::Expanding);
        rightTitleSpacer_->changeSize(SIDETITLEMARGIN, s_titleHeight,
                QSizePolicy::Fixed, QSizePolicy::Fixed);
        bottomSpacer_->changeSize(1, b, QSizePolicy::Expanding, QSizePolicy::Fixed);
    }

    left = l;
    right = r;
    top = t;
    bottom = b;

    // activate updated layout
    widget()->layout()->activate();
}

QSize LinspireClearClient::minimumSize() const
{
    return widget()->minimumSize();
}

void LinspireClearClient::show()
{
    widget()->show();
}

void LinspireClearClient::resize( const QSize& s )
{
    widget()->resize( s );
}

bool LinspireClearClient::eventFilter( QObject* o, QEvent* e )
{
    if( o != widget())
    return false;
    switch( e->type())
    {
    case QEvent::Resize:
        resizeEvent();
        return true;
    case QEvent::Paint:
        paintEvent( static_cast< QPaintEvent* >( e ));
        return true;
    case QEvent::MouseButtonDblClick:
        mouseDoubleClickEvent( static_cast< QMouseEvent* >( e ));
        return true;
    case QEvent::MouseButtonPress:
        processMousePressEvent( static_cast< QMouseEvent* >( e ));
        return true;
    default:
        break;
    }
    return false;
}

} // KWinLinspireClear
