/* Linspire Clear widget style for KDE 3
   Copyright (C) 2004 Lindows, Inc.
             (C) 2004 Sandro Giessl <ceebx@users.sourceforge.net>

   based on the KDE style "Plastik":

   Copyright (C) 2003 Sandro Giessl <ceebx@users.sourceforge.net>

   based on the KDE style "dotNET":

   Copyright (C) 2001-2002, Chris Lee <clee@kde.org>
                            Carsten Pfeiffer <pfeiffer@kde.org>
                            Karol Szwed <gallium@kde.org>
   Drawing routines completely reimplemented from KDE3 HighColor, which was
   originally based on some stuff from the KDE2 HighColor.

   based on drawing routines of the style "Keramik":

   Copyright (c) 2002 Malte Starostik <malte@kde.org>
             (c) 2002,2003 Maksim Orlovich <mo002j@mail.rochester.edu>
   based on the KDE3 HighColor Style
   Copyright (C) 2001-2002 Karol Szwed      <gallium@kde.org>
             (C) 2001-2002 Fredrik Hglund  <fredrik@kde.org>
   Drawing routines adapted from the KDE2 HCStyle,
   Copyright (C) 2000 Daniel M. Duley       <mosfet@kde.org>
             (C) 2000 Dirk Mueller          <mueller@kde.org>
             (C) 2001 Martijn Klingens      <klingens@kde.org>
   Progressbar code based on KStyle,
   Copyright (C) 2001-2002 Karol Szwed <gallium@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
 */

#include <qapplication.h>
#include <qbitmap.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qheader.h>
#include <qimage.h>
#include <qpainter.h>
#include <qpointarray.h>
#include <qprogressbar.h>
#include <qtabbar.h>
#include <qlineedit.h>
#include <qlistbox.h>
#include <qscrollbar.h>
#include <qstylefactory.h>
#include <qstyleplugin.h>
#include <qpushbutton.h>
#include <qtabwidget.h>
#include <qtimer.h>
#include <qtoolbutton.h>
#include <qtoolbar.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qdrawutil.h>
#include <qvariant.h>
#include <qradiobutton.h>
#include <qregion.h>
#include <qsettings.h>
#include <qslider.h>
#include <qwidgetstack.h>

#include "linspireclear.h"
#include "linspireclear.moc"
#include "misc.h"
#include "bitmap.h"

// settings for the animation...
#define ANIMATIONTIMEOUT 35
static const double ANIMATIONSTEPS = 18.0;

// popupmenu item constants...
static const int itemHMargin = 6;
static const int itemFrame = 2;
static const int arrowHMargin = 6;
static const int rightBorder = 12;

// -- Style Plugin Interface -------------------------
class LinspireClearStylePlugin : public QStylePlugin
{
    public:
        LinspireClearStylePlugin() {}
        ~LinspireClearStylePlugin() {}

        QStringList keys() const {
            return QStringList() << "LinspireClear";
        }

        QStyle* create( const QString& key ) {
            if (key.lower() == "linspireclear")
                return new LinspireClearStyle;
            return 0;
        }
};

Q_EXPORT_PLUGIN( LinspireClearStylePlugin )
// -- end --

LinspireClearStyle::LinspireClearStyle() : KStyle( AllowMenuTransparency, /*WindowsStyleScrollBar*/ThreeButtonScrollBar),
    kickerMode(false),
    kornMode(false),
    taskbarMode(false)
{
    hoverWidget = 0;
    hoverTab = 0;
    hoverSubControl = SC_None;
    hoverHeaderId = -1;
    hoverHeaderCurrent = 0;

    // setup pixmap cache...
    pixmapCache = new QIntCache<CacheEntry>(327680, 2017);
    pixmapCache->setAutoDelete(true);

    // read settings
    QSettings settings;
    _contrast = settings.readNumEntry("/Qt/KDE/contrast", 6);
    settings.beginGroup("/linspireclearstyle/Settings");
    _drawToolBarSeparator = settings.readBoolEntry("/drawToolBarSeparator", false);
    _symmetricButtonSurface = settings.readBoolEntry("/symmetricButtonSurface", true);
    _inputFocusHighlight = settings.readBoolEntry("/inputFocusHighlight", true);
    _mouseOver = settings.readBoolEntry("/mouseOver", true);
    _mouseOverAnimation = _mouseOver && settings.readBoolEntry("/mouseOverAnimation", true);
    _toolbarMouseOver = _mouseOver && settings.readBoolEntry("/toolbarMouseOver", false);

    // setup animation timer...
    if( _mouseOverAnimation ) {
        animTimer = new QTimer(this);
        connect( animTimer, SIGNAL(timeout()), this, SLOT(animTimerTimeout()) );
        // will be started/stopped in eventFilter and animTimerTimeout().
    } else {
        animTimer = 0;
    }
}

LinspireClearStyle::~LinspireClearStyle()
{
    delete pixmapCache;
}


void LinspireClearStyle::polish(QApplication* app)
{
    if (!qstrcmp(app->argv()[0], "kicker"))
        kickerMode = true;
    else if (!qstrcmp(app->argv()[0], "korn"))
        kornMode = true;
}

void LinspireClearStyle::polish(QWidget* widget)
{
    if( !strcmp(widget->name(), "__khtml") ) { // is it a khtml widget...?
        khtmlWidgets[widget] = true;
        connect(widget, SIGNAL(destroyed(QObject*)), this, SLOT(khtmlWidgetDestroyed(QObject*)));
    }

    taskbarMode = widget->inherits("TaskContainer"); // kicker's Taskbar...

    // use qt_cast where possible to check if the widget inheits one of the classes. might improve
    // performance compared to QObject::inherits()

    // setup widgets for hover animation (or simple mouse over in case animation is disabled)
    if ( _mouseOver &&
            (::qt_cast<QPushButton*>(widget) || ::qt_cast<QComboBox*>(widget) || ::qt_cast<QSpinWidget*>(widget) ||
            ::qt_cast<QToolButton*>(widget) ) ) {
        widget->installEventFilter(this);
        if( _mouseOverAnimation ) {
            AnimInfo info;
            info.status = Stopped;
            info.step = 0;
            animWidgets[widget] = info;
            connect(widget, SIGNAL(destroyed(QObject*)), this, SLOT(animWidgetDestroyed(QObject*)));
        }
    }

    // line edit focus highlight
    if (::qt_cast<QLineEdit*>(widget)) {
        widget->installEventFilter(this);
    }

    // QCheckBoxes and QRadioButtons are cached heavily internally so no animation is possible for them, use
    // normal highlighting instead...
    if ( _mouseOver && ( ::qt_cast<QCheckBox*>(widget) || ::qt_cast<QRadioButton*>(widget) ||
            widget->inherits("QSplitterHandle") ) )
    {
        widget->installEventFilter(this);
    } else if ( _mouseOver && ( ::qt_cast<QTabBar*>(widget) || ::qt_cast<QHeader*>(widget)
                || ::qt_cast<QScrollBar*>(widget) || ::qt_cast<QSlider*>(widget) ) ) {
        widget->setMouseTracking(true);
        widget->installEventFilter(this);
    }

    // adjust background of some widgets...
    if ( /*::qt_cast<QScrollBar*>(widget) ||*/ ::qt_cast<QPopupMenu*>(widget) ) {
        // set to NoBackground to reduce flicker.
        widget->setBackgroundMode( NoBackground );
    } else if ( !taskbarMode && ( ::qt_cast<QToolButton*>(widget) || ::qt_cast<QMenuBar*>(widget) || widget->inherits("QTableHeader") ) ) {
        // don't change background of Kicker's taskbar which surprisingly inherits QToolButton
        widget->setBackgroundMode( PaletteBackground );
    } else if ( ::qt_cast<QHeader*>(widget)&&!widget->inherits("QTableHeader") ) {
        widget->setBackgroundMode( NoBackground );
    } else if ( widget->parent() && ::qt_cast<QWidgetStack*>(widget->parent()) && !strcmp(widget->parent()->name(), "tab pages") ) {

        // modify the widget's palette to make the background color slightly more light
        // be careful and save the old palette to set it back in unPolish().
        QPalette p = widget->palette();
        oldPalettes[widget] = p;
        QColorGroup cg = p.active();
        QColor newBg = cg.background().light(100+_contrast/4);
//         qDebug("newActiveBg rgb: %d %d %d", newBg.red(), newBg.green(), newBg.blue() );
        cg.setColor( QColorGroup::Background, newBg );
        p.setActive(cg);
        cg = p.inactive();
        newBg = cg.background().light(100+_contrast/4);
//         qDebug("newInactiveBg rgb: %d %d %d", newBg.red(), newBg.green(), newBg.blue() );
        cg.setColor( QColorGroup::Background, newBg );
        p.setInactive(cg);
        cg = p.disabled();
        newBg = cg.background().light(100+_contrast/4);
//         qDebug("newDisabledBg rgb: %d %d %d", newBg.red(), newBg.green(), newBg.blue() );
        cg.setColor( QColorGroup::Background, newBg );
        p.setDisabled(cg);
        widget->setPalette(p);
    }

    if ( !qstrcmp(widget->name(), "kde toolbar widget") ) {
        widget->installEventFilter(this);
    }

//     // debug... what kinds of widgets, background modes etc. go through polish() ?
//     // filter out widgets which aren't of interest at the moment...
//     if(! (widget->inherits("QPushButton")||widget->inherits("QToolButton")||widget->inherits("QHeader")||
//          widget->inherits("QComboBox")||widget->inherits("QLineEdit")||widget->inherits("QSpinBox")||
//          widget->inherits("QTextEdit")||widget->inherits("QCheckBox")||widget->inherits("QRadioButton")||
//          widget->inherits("QSlider")) ) {
//         qDebug( "polish():" );
//         qDebug( "%s: name: %s", widget->className(), widget->name() );
//         if (widget->parentWidget() )
//             qDebug( "%s: parentWidget: %s", widget->className(), widget->parentWidget()->className() );
//         if( widget->backgroundMode() == NoBackground)
//             qDebug( "%s: backgroundMode: NoBackground", widget->className() );
//         else if( widget->backgroundMode() == PaletteBackground)
//             qDebug( "%s: backgroundMode: PaletteBackground", widget->className() );
//         else if( widget->backgroundMode() == PaletteButton)
//             qDebug( "%s: backgroundMode: PaletteButton", widget->className() );
//         else if( widget->backgroundMode() == PaletteBase)
//             qDebug( "%s: backgroundMode: PaletteBase", widget->className() );
//         else
//             qDebug( "%s: backgroundMode: other background: %d", widget->className(), widget->backgroundMode() );
//     }

    KStyle::polish(widget);
}

void LinspireClearStyle::unPolish(QWidget* widget)
{
    if( !strcmp(widget->name(), "__khtml") ) { // is it a khtml widget...?
        khtmlWidgets.remove(widget);
    }

    // use qt_cast to check if the widget inheits one of the classes.
    if ( _mouseOver &&
            (::qt_cast<QPushButton*>(widget) || ::qt_cast<QComboBox*>(widget) || ::qt_cast<QSpinWidget*>(widget) ||
            ::qt_cast<QToolButton*>(widget) ) ) {
        widget->removeEventFilter(this);
        if ( _mouseOverAnimation )
            animWidgets.remove(widget);
    }

    if ( _mouseOver && ( ::qt_cast<QCheckBox*>(widget) || ::qt_cast<QRadioButton*>(widget) ||
            widget->inherits("QSplitterHandle") ) )
    {
        widget->removeEventFilter(this);
    } else if ( _mouseOver && ( ::qt_cast<QTabBar*>(widget) || ::qt_cast<QHeader*>(widget)
                || ::qt_cast<QScrollBar*>(widget) || ::qt_cast<QSlider*>(widget) ) ) {
        widget->setMouseTracking(false);
        widget->removeEventFilter(this);
    }

    // line edit focus highlight
    if (::qt_cast<QLineEdit*>(widget)) {
        widget->removeEventFilter(this);
    }

    if ( /*::qt_cast<QScrollBar*>(widget) || */::qt_cast<QPopupMenu*>(widget) ) {
        widget->setBackgroundMode( PaletteBackground );
    } else if ( !taskbarMode && ( ::qt_cast<QToolButton*>(widget) || ::qt_cast<QMenuBar*>(widget) || widget->inherits("QTableHeader") ) ) {
        widget->setBackgroundMode( PaletteButton );
    } else if ( ::qt_cast<QHeader*>(widget)&&!widget->inherits("QTableHeader") ) {
        widget->setBackgroundMode( PaletteButton );
    } else if ( widget->parent() && ::qt_cast<QWidgetStack*>(widget->parent()) && !strcmp(widget->parent()->name(), "tab pages") ) {
        // restore the original palette...
        if(oldPalettes.contains(widget) ) {
            QPalette p = oldPalettes[widget];
//             QColor bg = p.active().background();
//             qDebug("oldActiveBg rgb: %d %d %d", bg.red(), bg.green(), bg.blue() );
//             bg = p.inactive().background();
//             qDebug("oldInactiveBg rgb: %d %d %d", bg.red(), bg.green(), bg.blue() );
//             bg = p.disabled().background();
//             qDebug("oldDisabledBg rgb: %d %d %d", bg.red(), bg.green(), bg.blue() );
            widget->setPalette(p);
            oldPalettes.remove(widget); // remove the entry...
        }
    }

    if ( !qstrcmp(widget->name(), "kde toolbar widget") ) {
        widget->removeEventFilter(this);
    }

    KStyle::unPolish(widget);
}

void LinspireClearStyle::khtmlWidgetDestroyed(QObject* obj)
{
    khtmlWidgets.remove(static_cast<QWidget*>(obj));
}

void LinspireClearStyle::animWidgetDestroyed(QObject* obj)
{
    animWidgets.remove(static_cast<QWidget*>(obj));
}

void LinspireClearStyle::renderContour(QPainter *p,
                       const QRect &r,
                       const QColor &backgroundColor,
                       const QColor &contour,
                       const uint flags,
                       const EdgeType edge1,
                       const EdgeType edge2,
                       const EdgeType edge3,
                       const EdgeType edge4) const
{
    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    if((r_w <= 0)||(r_h <= 0))
        return;

    const bool drawLeft = flags&Draw_Left;
    const bool drawRight = flags&Draw_Right;
    const bool drawTop = flags&Draw_Top;
    const bool drawBottom = flags&Draw_Bottom;
    const bool disabled = flags&Is_Disabled;
    const bool aBlend = true;

    QColor contourColor;
    if (disabled) {
        contourColor = hsvRelative(contour, 0, - 6*_contrast, 0); // reduce the saturation if disabled...
    } else {
        contourColor = contour;
    }

// sides
    p->setPen( contourColor );
    if(drawTop) {
        int top_left  = (!drawLeft||edge1==Edge_Plain)?
                        r_x
                        : (edge1==Edge_Round3)?r_x+4:r_x+2;
        int top_right = (!drawRight||edge2==Edge_Plain)?
                        r_x2
                        : (edge2==Edge_Round3)?r_x2-4:r_x2-2;
        p->drawLine(top_left, r_y, top_right, r_y);
    }
    if(drawBottom) {
        int bottom_left  = (!drawLeft||edge3==Edge_Plain)?
                        r_x
                        : (edge3==Edge_Round3)?r_x+4:r_x+2;
        int bottom_right = (!drawRight||edge4==Edge_Plain)?
                        r_x2
                        : (edge4==Edge_Round3)?r_x2-4:r_x2-2;
        p->drawLine(bottom_left, r_y2, bottom_right, r_y2);
    }
    if(drawLeft) {
        int left_top    = (!drawTop||edge1==Edge_Plain)?
                        drawTop?r_y+1:r_y // don't overlap with the top line etc.
                        : (edge1==Edge_Round3)?r_y+4:r_y+2;
        int left_bottom = (!drawBottom||edge3==Edge_Plain)?
                        drawBottom?r_y2-1:r_y2
                        : (edge3==Edge_Round3)?r_y2-4:r_y2-2;
        p->drawLine(r_x, left_top, r_x, left_bottom);
    }
    if(drawRight) {
        int right_top    = (!drawTop||edge2==Edge_Plain)?
                        drawTop?r_y+1:r_y
                        : (edge2==Edge_Round3)?r_y+4:r_y+2;
        int right_bottom = (!drawBottom||edge4==Edge_Plain)?
                        drawBottom?r_y2-1:r_y2
                        : (edge4==Edge_Round3)?r_y2-4:r_y2-2;
        p->drawLine(r_x2, right_top, r_x2, right_bottom);
    }


// edges
    const int alphaAA = 110; // the alpha value for anti-aliasing of Edge_Round1...
    // pen color is still contourColor. paint everything that is needed in this color
    // to avoid pen switching, which is quite expensive.
    if(drawLeft && drawTop) {
        switch(edge1) {
            case Edge_Soft:
                p->drawPoint( r_x, r_y+1 );
                p->drawPoint( r_x+1, r_y );
                break;
            case Edge_Round1:
                p->drawPoint(r_x+1, r_y+1);
                break;
            case Edge_Round2:
                p->drawPoint(r_x+1, r_y+1);
                break;
            default:
                break;
        }
    }
    if(drawLeft && drawBottom) {
        switch(edge3) {
            case Edge_Soft:
                p->drawPoint( r_x, r_y2-1 );
                p->drawPoint( r_x+1, r_y2 );
                break;
            case Edge_Round1:
                p->drawPoint(r_x+1, r_y2-1);
                break;
            case Edge_Round2:
                p->drawPoint(r_x+1, r_y2-1);
                break;
            default:
                break;
        }
    }
    if(drawRight && drawTop) {
        switch(edge2) {
            case Edge_Soft:
                p->drawPoint( r_x2, r_y+1 );
                p->drawPoint( r_x2-1, r_y );
                break;
            case Edge_Round1:
                p->drawPoint(r_x2-1, r_y+1);
                break;
            case Edge_Round2:
                p->drawPoint(r_x2-1, r_y+1);
                break;
            default:
                break;
        }
    }
    if(drawRight && drawBottom) {
        switch(edge4) {
            case Edge_Soft:
                p->drawPoint( r_x2, r_y2-1 );
                p->drawPoint( r_x2-1, r_y2 );
                break;
            case Edge_Round1:
                p->drawPoint(r_x2-1, r_y2-1);
                break;
            case Edge_Round2:
                p->drawPoint(r_x2-1, r_y2-1);
                break;
            default:
                break;
        }
    }

    // paint everything with pen color backgroundColor...:
    p->setPen( backgroundColor );
    // if we don't want to do full alpha-blending, fill empty edges...
    if(!aBlend && drawLeft && drawTop) {
        switch(edge1) {
            case Edge_Round1:
                p->drawPoint( r_x, r_y );
                break;
            case Edge_Round2:
                p->drawPoint( r_x, r_y );
                p->drawPoint( r_x+1, r_y );
                p->drawPoint( r_x, r_y+1 );
                break;
            case Edge_Round3:
                p->drawPoint( r_x, r_y );
                break;
            default:
                break;
        }
    }
    if(!aBlend && drawLeft && drawBottom) {
        switch(edge3) {
            case Edge_Round1:
                p->drawPoint( r_x, r_y2 );
                break;
            case Edge_Round2:
                p->drawPoint( r_x, r_y2 );
                p->drawPoint( r_x+1, r_y2 );
                p->drawPoint( r_x, r_y2-1 );
                break;
            case Edge_Round3:
                p->drawPoint( r_x, r_y2 );
                break;
            default:
                break;
        }
    }
    if(!aBlend && drawRight && drawTop) {
        switch(edge2) {
            case Edge_Round1:
                p->drawPoint( r_x2, r_y );
                break;
            case Edge_Round2:
                p->drawPoint( r_x2, r_y );
                p->drawPoint( r_x2-1, r_y );
                p->drawPoint( r_x2, r_y+1 );
                break;
            case Edge_Round3:
                p->drawPoint( r_x2, r_y );
                break;
            default:
                break;
        }
    }
    if(!aBlend && drawRight && drawBottom) {
        switch(edge4) {
            case Edge_Round1:
                p->drawPoint( r_x2, r_y2 );
                break;
            case Edge_Round2:
                p->drawPoint( r_x2, r_y2 );
                p->drawPoint( r_x2-1, r_y2 );
                p->drawPoint( r_x2, r_y2-1 );
                break;
            case Edge_Round3:
                p->drawPoint( r_x2, r_y2 );
                break;
            default:
                break;
        }
    }

    // group renderPixel stuff together...
    if(drawLeft && drawTop) {
        switch(edge1) {
            case Edge_Soft:
                renderPixel(p, r_x, r_y,alphaAA,contourColor,backgroundColor,aBlend);
                break;
            case Edge_Round1:
                // do anti-aliasing
                renderPixel(p, r_x+1, r_y, alphaAA,contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x, r_y+1, alphaAA,contourColor,backgroundColor,aBlend);
                break;
            case Edge_Round3: {
                renderPixel(p, r_x+1,r_y+0, 10, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+2,r_y+0, 121, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+3,r_y+0, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+0,r_y+1, 10, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+1,r_y+1, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+2,r_y+1, 44, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x+3,r_y+1, 1, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x+0,r_y+2, 121, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+1,r_y+2, 44, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x+0,r_y+3, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+1,r_y+3, 1, contourColor/*,backgroundColor,aBlend*/);
                break;
            }
            case Edge_Plain:
            default:
                // do nothing
                break;
        }
    }
    if(drawLeft && drawBottom) {
        switch(edge3) {
            case Edge_Soft:
                renderPixel(p, r_x,r_y2, alphaAA,contourColor,backgroundColor,aBlend);
                break;
            case Edge_Round1:
                // do anti-aliasing
                renderPixel(p, r_x+1,r_y2, alphaAA,contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x,r_y2-1, alphaAA,contourColor,backgroundColor,aBlend);
                break;
            case Edge_Round3: {
                renderPixel(p, r_x+1,r_y2-0, 10, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+2,r_y2-0, 121, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+3,r_y2-0, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+0,r_y2-1, 10, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+1,r_y2-1, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+2,r_y2-1, 44, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x+3,r_y2-1, 1, contourColor/*,backgroundColor,aBlend*/); // always alpha blend "inner" aa
                renderPixel(p, r_x+0,r_y2-2, 121, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+1,r_y2-2, 44, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x+0,r_y2-3, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x+1,r_y2-3, 1, contourColor/*,backgroundColor,aBlend*/);
                break;
            }
            case Edge_Plain:
            default:
                // do nothing
                break;
        }
    }
    if(drawRight && drawTop) {
        switch(edge2) {
            case Edge_Soft:
                renderPixel(p, r_x2,r_y, alphaAA,contourColor,backgroundColor,aBlend);
                break;
            case Edge_Round1:
                // do anti-aliasing
                renderPixel(p,r_x2-1,r_y,alphaAA,contourColor,backgroundColor,aBlend);
                renderPixel(p,r_x2,r_y+1,alphaAA,contourColor,backgroundColor,aBlend);
                break;
            case Edge_Round3: {
                renderPixel(p, r_x2-1,r_y+0, 10, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-2,r_y+0, 121, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-3,r_y+0, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-0,r_y+1, 10, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-1,r_y+1, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-2,r_y+1, 44, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x2-3,r_y+1, 1, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x2-0,r_y+2, 121, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-1,r_y+2, 44, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x2-0,r_y+3, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-1,r_y+3, 1, contourColor/*,backgroundColor,aBlend*/);
                break;
            }
            case Edge_Plain:
            default:
                // do nothing
                break;
        }
    }
    if(drawRight && drawBottom) {
        switch(edge4) {
            case Edge_Soft:
                renderPixel(p,r_x2,r_y2,alphaAA,contourColor,backgroundColor,aBlend);
                break;
            case Edge_Round1:
                // do anti-aliasing
                renderPixel(p,r_x2-1,r_y2,alphaAA,contourColor,backgroundColor,aBlend);
                renderPixel(p,r_x2,r_y2-1,alphaAA,contourColor,backgroundColor,aBlend);
                break;
            case Edge_Round3: {
                renderPixel(p, r_x2-1,r_y2-0, 10, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-2,r_y2-0, 121, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-3,r_y2-0, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-0,r_y2-1, 10, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-1,r_y2-1, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-2,r_y2-1, 44, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x2-3,r_y2-1, 1, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x2-0,r_y2-2, 121, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-1,r_y2-2, 44, contourColor/*,backgroundColor,aBlend*/);
                renderPixel(p, r_x2-0,r_y2-3, 235, contourColor,backgroundColor,aBlend);
                renderPixel(p, r_x2-1,r_y2-3, 1, contourColor/*,backgroundColor,aBlend*/);
                break;
            }
            case Edge_Plain:
            default:
                // do nothing
                break;
        }
    }
}

void LinspireClearStyle::renderMask(QPainter *p,
                              const QRect &r,
                              const QColor &color,
                              const uint flags) const
{
    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    if((r_w <= 0)||(r_h <= 0))
        return;

    const bool roundUpperLeft = flags&Round_UpperLeft;
    const bool roundUpperRight = flags&Round_UpperRight;
    const bool roundBottomLeft = flags&Round_BottomLeft;
    const bool roundBottomRight = flags&Round_BottomRight;

    p->fillRect (QRect(r_x+1, r_y+1, r_w-2, r_h-2) , color);

    p->setPen(color);
    // sides
    p->drawLine(roundUpperLeft?r_x+1:r_x, r_y,
                roundUpperRight?r_x2-1:r_x2, r_y );
    p->drawLine(roundBottomLeft?r_x+1:r_x, r_y2,
                roundBottomRight?r_x2-1:r_x2, r_y2 );
    p->drawLine(r_x, roundUpperLeft?r_y+1:r_y,
                r_x, roundBottomLeft?r_y2-1:r_y2 );
    p->drawLine(r_x2, roundUpperLeft?r_y+1:r_y,
                r_x2, roundBottomLeft?r_y2-1:r_y2 );
}

void LinspireClearStyle::renderSurface(QPainter *p,
                                 const QRect &r,
                                 const QColor &color,
                                 uint flags) const
{
    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    if(r_w <= 0 || r_h <= 0)
        return;

    const bool disabled = flags&Is_Disabled;

    const bool drawSoft = flags&Draw_Soft;
    const bool symmetric = flags&Draw_Symmetric;
//     const bool drawLeft = flags&Draw_Left;
//     const bool drawRight = flags&Draw_Right;
//     const bool drawTop = flags&Draw_Top;
//     const bool drawBottom = flags&Draw_Bottom;
//     const bool roundUpperLeft = flags&Round_UpperLeft;
//     const bool roundUpperRight = flags&Round_UpperRight;
//     const bool roundBottomLeft = flags&Round_BottomLeft;
//     const bool roundBottomRight = flags&Round_BottomRight;
    const bool sunken = flags&Is_Sunken;
    const bool horizontal = flags&Is_Horizontal;
    // FIXME: currently unused... remove?
    bool highlight = false;
//     // only highlight if not sunken & not disabled...
    if(!sunken && !disabled) {
        highlight = (flags&Is_Highlight);
    }

    QColor baseColor = color;
    if( disabled )
        baseColor = hsvRelative(baseColor, 0, - 7*_contrast, 0 );

    const QColor cTop =     sunken?
                                alphaBlendColors(baseColor,Qt::black, 235)
                                : alphaBlendColors(baseColor,Qt::white, 60 /*100*/);
    const QColor cBottom  = sunken?
                                alphaBlendColors(baseColor,Qt::white, 190)
                                : alphaBlendColors(baseColor,Qt::black, 245);
    QColor cBottomBlend = hsvRelative(cBottom, -35, -40, +15);
    const QColor cBlur  = alphaBlendColors( cTop, cBottom, 150 );
    const QColor cSharp = cBottom.dark(115);
    const QColor cSoftBlur = alphaBlendColors( cTop, cBottom, 100 );

    // TODO: maybe respect roundUpperLeft/Right, roundBottomLeft/Right flags,
    //       should working fine this way, though, because the edges are
    //       over-painted by the contour...
// button area...
    if (horizontal)
    // draw surface for horizontal widgets...
    {
            int h1 = symmetric ? r_h/2 : r_h/3;
            if (drawSoft)
                h1 -= 4;
            else
                h1 -= 1;
            if (h1 < 1)
                h1 = 1;
            if(sunken) h1 += pixelMetric(PM_ButtonShiftVertical);
            int h2 = drawSoft?(r_h-h1-4):(r_h-h1-1-1);
            // dark bottom area...
            renderGradient(p, QRect(r_x, drawSoft?(r_y+h1+4):(r_y+h1+1+1), r_w, h2),
                    cBottom, cBottomBlend, true);
            // light top area...
            p->fillRect( r_x, r_y, r_w, h1, cTop );
            if(!drawSoft) {
                // blurring line...
                p->setPen( cBlur );
                p->drawLine( /*drawLeft?r_x+2:*/r_x, r_y+h1, /*drawRight?r_x2-2:*/r_x2, r_y+h1 );
                // the sharpening line...
                p->setPen( cSharp );
                p->drawLine( /*drawLeft?r_x+2:*/r_x, r_y+h1+1, /*drawRight?r_x2-2:*/r_x2, r_y+h1+1 );
            } else {
                renderGradient(p, QRect(r_x, r_y+h1, r_w, 4),
                               cTop, cSoftBlur, true);
            }
    } else
    // vertical, basically rotate the paintings by 90 degres...
    {
            int w1 = symmetric ? r_w/2 : r_w/3;
            if (drawSoft)
                w1 -= 4;
            else
                w1 -= 1;
            if (w1 < 1)
                w1 = 1;
            int w2 = drawSoft?(r_w-w1-4):(r_w-w1-1-1);
            // dark bottom area...
            renderGradient(p, QRect(drawSoft?(r_x+w1+4):(r_x+w1+1+1), r_y, w2, r_h),
                    cBottom, cBottomBlend, false);
            // light top area...
            p->fillRect( r_x, r_y, w1, r_h, cTop );
            if(!drawSoft) {
                // the blurring line...
                p->setPen( cBlur );
                p->drawLine( r_x+w1, /*drawTop?r_y+2:*/r_y, r_x+w1, /*drawBottom?r_y2-2:*/r_y2 );
                // the sharpening line...
                p->setPen( cSharp );
                p->drawLine( r_x+w1+1, /*drawTop?r_y+2:*/r_y, r_x+w1+1, /*drawBottom?r_y2-2:*/r_y2 );
            } else {
                renderGradient(p, QRect(r_x+w1, r_y, 4, r_h),
                               cTop, cSoftBlur, false);
            }
    }
}

void LinspireClearStyle::renderPixel(QPainter *p,
            int x, int y,
            const int alpha,
            const QColor &color,
            const QColor &background,
            bool fullAlphaBlend) const
{
    if(fullAlphaBlend)
    // full alpha blend: paint into an image with alpha buffer and convert to a pixmap ...
    {
        QRgb rgb = color.rgb();
        // generate a quite unique key -- use the unused width field to store the alpha value.
        CacheEntry search(cAlphaDot, alpha, 0, rgb, 0);
        int key = search.key();

        CacheEntry *cacheEntry;
        if( (cacheEntry = pixmapCache->find(key)) ) {
            if( search == *cacheEntry ) { // match! we can draw now...
                if(cacheEntry->pixmap)
                    p->drawPixmap(x, y, *(cacheEntry->pixmap) );
                return;
            } else { //Remove old entry in case of a conflict!
                pixmapCache->remove( key );
            }
        }

        // there wasn't anything matching in the cache, create the pixmap now...
        QImage aImg(1,1,32); // 1x1
        aImg.setAlphaBuffer(true);
        aImg.setPixel(0,0,qRgba(qRed(rgb),qGreen(rgb),qBlue(rgb),alpha));
        QPixmap *result = new QPixmap(aImg);

        p->drawPixmap(x,y, *result);

        // add to the cache...
        CacheEntry *toAdd = new CacheEntry(search);
        toAdd->pixmap = result;
        bool insertOk = pixmapCache->insert( key, toAdd, result->depth()/8);
        if(!insertOk)
            delete result;
    } else
    // don't use an alpha buffer: calculate the resulting color from the alpha value, the fg- and the bg-color.
    {
        QRgb rgb_a = color.rgb();
        QRgb rgb_b = background.rgb();
        int a = alpha;
        if(a>255) a = 255;
        if(a<0) a = 0;
        int a_inv = 255 - a;

        QColor res  = QColor( qRgb(qRed(rgb_b)*a_inv/255 + qRed(rgb_a)*a/255,
                              qGreen(rgb_b)*a_inv/255 + qGreen(rgb_a)*a/255,
                              qBlue(rgb_b)*a_inv/255 + qBlue(rgb_a)*a/255) );
        p->setPen(res);
        p->drawPoint(x,y);
    }
}

void LinspireClearStyle::renderButton(QPainter *p,
                               const QRect &r,
                               const QColorGroup &g,
                               bool sunken,
                               bool mouseOver,
                               bool horizontal,
                               bool enabled,
                               bool khtmlMode,
                               int highlightStep) const
{
    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    const QPen oldPen( p->pen() );

    const WidgetState s = enabled?(sunken?IsPressed:IsEnabled):IsDisabled;

    uint surfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|
                         Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight|Draw_Soft;
    if(_symmetricButtonSurface)
        surfaceFlags|=Draw_Symmetric;
    if(horizontal) surfaceFlags|=Is_Horizontal;
    if(!enabled) surfaceFlags|=Is_Disabled;
    else {
        if(sunken) surfaceFlags|=Is_Sunken;
        else {
            if(mouseOver) {
                surfaceFlags|=Is_Highlight;
//                 if(horizontal) {
//                     surfaceFlags|=Highlight_Top;
//                     surfaceFlags|=Highlight_Bottom;
//                 } else {
//                     surfaceFlags|=Highlight_Left;
//                     surfaceFlags|=Highlight_Right;
//                 }
            }
        }
    }
    renderSurface(p, QRect(r_x+1, r_y+1, r_w-2, r_h-2),
                   g.button().light(100+highlightStep /*(highlightStep*highlightStep)*/), surfaceFlags);

    uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom/*|
            Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight*/;
    if(!enabled) contourFlags|=Is_Disabled;
    if( khtmlMode ) contourFlags|=Draw_AlphaBlend;
    renderContour(p, r, g.background(), getColor(g,ButtonContour,s).light(100+highlightStep), contourFlags);

    p->setPen(oldPen);
}

void LinspireClearStyle::renderDot(QPainter *p,
                             int x, int y,
                             const QColor &baseColor,
                             const bool thick,
                             const bool sunken) const
{
    const QColor topColor = alphaBlendColors(baseColor, sunken?baseColor.dark(130):baseColor.light(150), 70);
    const QColor bottomColor = alphaBlendColors(baseColor, sunken?baseColor.light(150):baseColor.dark(130), 70);
    p->setPen(topColor );
    p->drawLine(x, y, x+1, y);
    p->drawPoint(x, y+1);
    p->setPen(bottomColor );
    if(thick) {
        p->drawLine(x+1, y+2, x+2, y+2);
        p->drawPoint(x+2, y+1);
    } else {
        p->drawPoint(x+1, y+1);
    }
}

void LinspireClearStyle::renderGradient(QPainter *painter,
                        const QRect &rect,
                        const QColor &c1,
                        const QColor &c2,
                        bool horizontal) const
{
    if((rect.width() <= 0)||(rect.height() <= 0))
        return;

    // generate a quite unique key for this surface.
    CacheEntry search(cGradientTile,
                      horizontal ? 0 : rect.width(),
                      horizontal ? rect.height() : 0,
                      c1.rgb(), c2.rgb(), horizontal );
    int key = search.key();

    CacheEntry *cacheEntry;
    if( (cacheEntry = pixmapCache->find(key)) ) {
        if( search == *cacheEntry ) { // match! we can draw now...
            if(cacheEntry->pixmap) {
                painter->drawTiledPixmap(rect, *(cacheEntry->pixmap) );
            }
            return;
        } else {
            // Remove old entry in case of a conflict!
            // This shouldn't happen very often, see comment in CacheEntry.
            pixmapCache->remove(key);
        }
    }

    // there wasn't anything matching in the cache, create the pixmap now...
    QPixmap *result = new QPixmap(horizontal ? 10 : rect.width(),
                                  horizontal ? rect.height() : 10);
    QPainter p(result);

    int r_w = result->rect().width();
    int r_h = result->rect().height();
    int r_x, r_y, r_x2, r_y2;
    result->rect().coords(&r_x, &r_y, &r_x2, &r_y2);

    int rDiff, gDiff, bDiff;
    int rc, gc, bc;

    register int x, y;

    rDiff = ( c2.red())   - (rc = c1.red());
    gDiff = ( c2.green()) - (gc = c1.green());
    bDiff = ( c2.blue())  - (bc = c1.blue());

    register int rl = rc << 16;
    register int gl = gc << 16;
    register int bl = bc << 16;

    int rdelta = ((1<<16) / (horizontal ? r_h : r_w)) * rDiff;
    int gdelta = ((1<<16) / (horizontal ? r_h : r_w)) * gDiff;
    int bdelta = ((1<<16) / (horizontal ? r_h : r_w)) * bDiff;

    // these for-loops could be merged, but the if's in the inner loop
    // would make it slow
    if(horizontal) {
        for ( y = 0; y < r_h; y++ ) {
            rl += rdelta;
            gl += gdelta;
            bl += bdelta;

            p.setPen(QColor(rl>>16, gl>>16, bl>>16));
            p.drawLine(r_x, r_y+y, r_x2, r_y+y);
        }
    } else {
        for( x = 0; x < r_w; x++) {
            rl += rdelta;
            gl += gdelta;
            bl += bdelta;

            p.setPen(QColor(rl>>16, gl>>16, bl>>16));
            p.drawLine(r_x+x, r_y, r_x+x, r_y2);
        }
    }

    p.end();

    // draw the result...
    painter->drawTiledPixmap(rect, *result);

    // insert into cache using the previously created key.
    CacheEntry *toAdd = new CacheEntry(search);
    toAdd->pixmap = result;
    bool insertOk = pixmapCache->insert( key, toAdd, result->width()*result->height()*result->depth()/8 );

    if(!insertOk)
        delete result;
}

void LinspireClearStyle::renderPanel(QPainter *p,
                              const QRect &r,
                              const QColorGroup &g,
                              const bool pseudo3d,
                              const bool sunken/*,
                              const bool thick*/) const
{
    int r_x, r_x2, r_y, r_y2, r_w, r_h;
    r_h = r.height();
    r_w = r_w;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    const QPen oldPen( p->pen() );

    // kicker exception
    if (kickerMode &&
            p->device() && p->device()->devType() == QInternal::Widget &&
            QCString(static_cast<QWidget*>(p->device())->className()) == "FittsLawFrame") {
            if (sunken) {
                const QCOORD corners[] = { r_x2, r_y, r_x2, r_y2, r_x, r_y2, r_x, r_y };
                p->setPen(g.background().dark());
                p->drawConvexPolygon(QPointArray(4, corners));
                p->setPen(g.background().light());
                p->drawPolyline(QPointArray(4, corners), 0, 3);
            } else {
                const QCOORD corners[] = {r_x, r_y2, r_x, r_y, r_x2, r_y, r_x2, r_y2 };
                p->setPen(g.background().dark());
                p->drawPolygon(QPointArray(4, corners));
                p->setPen(g.background().light());
                p->drawPolyline(QPointArray(4, corners), 0, 3);
            }
    } else {
        if(pseudo3d) {
            if (sunken)
                p->setPen(getColor(g, PanelDark) );
            else
                p->setPen(getColor(g, PanelLight) );
            p->drawLine(r_x+2, r_y+1, r_x2-2, r_y+1);
            p->setPen(getColor(g, PanelMid) );
            p->drawLine(r_x+1, r_y+2, r_x+1, r_y2-2);
            p->drawLine(r_x2-1, r_y+2, r_x2-1, r_y2-2);
            if (sunken)
                p->setPen(getColor(g, PanelLight) );
            else
                p->setPen(getColor(g, PanelDark) );
            p->drawLine(r_x+2, r_y2-1, r_x2-2, r_y2-1);
        }
        renderContour(p, r, g.background(),
                getColor(g, PanelContour),
                Draw_Left|Draw_Right|Draw_Top|Draw_Bottom,
                Edge_Round3, Edge_Round3, Edge_Round3, Edge_Round3 );
    }

    p->setPen(oldPen);
}

void LinspireClearStyle::renderTab(QPainter *p,
                            const QRect &r,
                            const QColorGroup &g,
                            bool mouseOver,
                            const bool selected,
                            const bool bottom,
                            const TabPosition pos,
//                             const bool triangular,
                            const bool cornerWidget) const
{
    const bool reverseLayout = QApplication::reverseLayout();

    const bool isFirst = (pos == First) || (pos == Single);
    const bool isLast = (pos == Last);
    const bool isSingle = (pos == Single);

    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    if (selected) {
    // is selected

    // the top part of the tab which is nearly the same for all positions
        QRect Rc; // contour
        if (!bottom) {
            if( isSingle ) {
                Rc = QRect(r_x, r_y, r_w, r_h-1);
            } else if ((isFirst&&!reverseLayout)||(isLast&&reverseLayout)) {
                Rc = QRect(r_x, r_y, r_w-1, r_h-1);
            } else if ((isFirst&&reverseLayout)||(isLast&&!reverseLayout)) {
                Rc = QRect(r_x+1, r_y, r_w-1, r_h-1);
            } else {
                Rc = QRect(r_x+1, r_y, r_w-2, r_h-1);
            }
        } else {
            if( isSingle ) {
                Rc = QRect(r_x, r_y+1, r_w, r_h-1);
            } else if ((isFirst&&!reverseLayout)||(isLast&&reverseLayout)) {
                Rc = QRect(r_x, r_y+1, r_w-1, r_h-1);
            } else if ((isFirst&&reverseLayout)||(isLast&&!reverseLayout)) {
                Rc = QRect(r_x+1, r_y+1, r_w-1, r_h-1);
            } else {
                Rc = QRect(r_x+1, r_y+1, r_w-2, r_h-1);
            }
        }
        const QRect Rs(Rc.x()+2, bottom?Rc.y():Rc.y()+2, Rc.width()-4, Rc.height()-2); // the resulting surface
        // the area where the fake border shoudl appear
        const QRect Rb(r_x, bottom?r_y:Rc.bottom()+1, r_w, r_h-Rc.height() );

        // surface
//         const QColor cTop = alphaBlendColors(g.button(),Qt::white, 130);
//         const QColor cBottom  = alphaBlendColors(g.button(),Qt::black, 245);
//         const QColor cBlur  = alphaBlendColors( cTop, cBottom, 150 );
        const QColor cBottom  = g.background().dark(120);
        const QColor cTop = alphaBlendColors(cBottom,Qt::white, 80);
        const QColor cSoftBlur = alphaBlendColors( cTop, cBottom, 100 ); 
//         int h1 = Rs.height()/3-1;
//         int h2 = Rs.height()-h1-1;
        int h1 = Rs.height()/3 - 2;
        int h2 = Rs.height()-h1;
        if(!bottom) {
            p->fillRect( Rs.x(), Rs.y(), Rs.width(), h1, cTop );
            renderGradient(p, QRect(Rs.x(), Rs.y()+h1, Rs.width(), 4),
                           cTop, cSoftBlur, true); 
            renderGradient(p, QRect(Rs.x(),Rs.y()+h1+4,Rs.width(),h2),
                    cBottom, g.background(), true);
        } else {
//             QColor cBottomBlend = hsvRelative(cBottom, -35, -40, +15);
            QColor cBottomBlend = cBottom.light(115);
            renderGradient(p, QRect(Rs.x(),Rs.y(),Rs.width(),h1),
                    g.background(), cTop, true);
            renderGradient(p, QRect(Rs.x(), Rs.y()+h1, Rs.width(), 4),
                           cTop, cSoftBlur, true); 
            renderGradient(p, QRect(Rs.x(), Rs.y()+h1+4, Rs.width(), h2),
                    cBottom, cBottomBlend, true);
        }

        // contour
        p->setPen( getColor(g, bottom?PanelDark:PanelLight) );
        p->drawLine( Rc.x()+2, bottom?Rc.bottom()-1:Rc.y()+1, Rc.right()-2, bottom?Rc.bottom()-1:Rc.y()+1 );
        p->setPen( getColor(g,PanelMid) );
        p->drawLine( Rc.x()+1, bottom?Rc.y():Rc.y()+1, Rc.x()+1, bottom?Rc.bottom()-1:Rc.bottom() );
        p->drawLine( Rc.right()-1, bottom?Rc.y():Rc.y()+1, Rc.right()-1, bottom?Rc.bottom()-1:Rc.bottom() );
        uint contourFlags = Draw_Left|Draw_Right;
        if(!bottom) {
            contourFlags |= Draw_Top;
        } else {
            contourFlags |= Draw_Bottom;
        }
        renderContour(p, Rc, g.background(), getColor(g,TabPanelContour), contourFlags, Edge_Round3, Edge_Round3, Edge_Round3, Edge_Round3);

        // left connection from the panel border to the tab. :)
        if(isFirst && !reverseLayout && !cornerWidget) {
            p->setPen( getColor(g,TabPanelContour) );
            p->drawLine(Rb.x(), Rb.y(), Rb.x(), Rb.bottom() );
            p->setPen( getColor(g,PanelMid) );
            p->drawLine(Rb.x()+1, Rb.y(), Rb.x()+1, Rb.bottom() );
            // fake parts of the panel border
            p->setPen( getColor(g,bottom?PanelDark:PanelLight) );
            p->drawLine(Rb.right(), bottom?Rb.bottom():Rb.y(), Rc.right()-1, bottom?Rb.bottom():Rb.y() );
        } else if(isFirst && reverseLayout && !cornerWidget) {
            p->setPen( getColor(g,TabPanelContour) );
            p->drawLine(Rb.right(), Rb.y(), Rb.right(), Rb.bottom() );
            p->setPen( getColor(g,PanelMid) );
            p->drawLine(Rb.right()-1, Rb.y(), Rb.right()-1, Rb.bottom() );
            // fake parts of the panel border
            p->setPen( getColor(g,bottom?PanelDark:PanelLight) );
            p->drawLine(Rb.x(), bottom?Rb.bottom():Rb.y(), Rc.x()+1, bottom?Rb.bottom():Rb.y() );
        } else {
            // fake parts of the panel border
            p->setPen( getColor(g,bottom?PanelDark:PanelLight) );
            p->drawLine(Rb.x(), bottom?Rb.bottom():Rb.y(), Rc.x()+1, bottom?Rb.bottom():Rb.y() );
            p->drawLine(Rb.right(), bottom?Rb.bottom():Rb.y(), Rc.right()-1, bottom?Rb.bottom():Rb.y() );
        }

    } else {
    // inactive tabs

    // the top part of the tab which is nearly the same for all positions
        QRect Rc; // contour
        if ( isFirst && !cornerWidget ) {
            // intend inactive tabs at the first position by 1 pixel
            Rc = reverseLayout ? QRect(r_x, r_y+2, r_w-1, r_h-4 )
                               : QRect(r_x+1, r_y+2, r_w-1, r_h-4 );
        } else {
            Rc = QRect(r_x, r_y+2, r_w, r_h-4 );
        }

        QRect Rs; // the resulting surface
        if ( (isFirst&&!reverseLayout) || (isLast&&reverseLayout) ) {
            Rs = QRect(Rc.x()+1, bottom?Rc.y():Rc.y()+1, Rc.width()-2, Rc.height()-1);
        } else {
            Rs = QRect(Rc.x(), bottom?Rc.y():Rc.y()+1, Rc.width()-1, Rc.height()-1);
        }
        // the area where the fake border shoudl appear
        const QRect Rb(r_x, bottom?r_y:Rc.bottom()+1, r_w, 2 );

        // surface
        const WidgetState s = /*enabled?(*/mouseOver?IsHighlighted:IsEnabled/*):IsDisabled*/;
        p->fillRect( Rs, getColor(g, TabSurfaceInactive, s) );

        // contour
        uint contourFlags;
        EdgeType e1 = Edge_Plain, e2 = Edge_Plain, e3 = Edge_Plain, e4 = Edge_Plain;
        if(!bottom) {
            e1 = e2 = Edge_Round2;
            contourFlags = Draw_Right|Draw_Left|Draw_Top;
        } else {
            e3 = e4 = Edge_Round2;
            contourFlags = Draw_Right|Draw_Left|Draw_Bottom;
        }
        renderContour(p, Rc, g.background(), getColor(g, InactiveTabContour), contourFlags, e1, e2, e3, e4);


//     // some "position specific" paintings...
        // fake parts of the panel border
        if(!bottom) {
            p->setPen( getColor(g,TabPanelContour) );
            p->drawLine(Rb.x(), Rb.y(), Rb.right(), Rb.y());
            p->setPen( getColor(g, PanelLight) );
            p->drawLine(Rb.x(), Rb.y()+1, Rb.right(), Rb.y()+1);
//             p->drawLine(Rb.x(), Rb.y(), ((isLast&&!reverseLayout)||(isFirst&&reverseLayout&&cornerWidget))?Rb.right():Rb.right()-1, Rb.y());
        } else {
            p->setPen( getColor(g,TabPanelContour) );
            p->drawLine(Rb.x(), Rb.bottom(), Rb.right(), Rb.bottom());
            p->setPen( getColor(g, PanelDark) );
            p->drawLine(Rb.x(), Rb.bottom()-1, Rb.right(), Rb.bottom()-1);
//             p->drawLine(Rb.x(), Rb.bottom(), ((isLast&&!reverseLayout)||(isFirst&&reverseLayout&&cornerWidget))?Rb.right():Rb.right()-1, Rb.bottom());
        }
        // left connection from the panel border to the tab. :)
        if(isFirst && !reverseLayout && !cornerWidget) {
            p->setPen( getColor(g,TabPanelContour) );
            p->drawLine(Rb.x(), Rb.y(), Rb.x(), Rb.bottom() );
//             p->setPen( getColor(g,PanelMid) );
//             p->drawLine(Rb.x()+1, Rb.y(), Rb.x()+1, Rb.bottom() );
        } else if(isFirst && reverseLayout && !cornerWidget) {
            p->setPen( getColor(g,TabPanelContour) );
            p->drawLine(Rb.right(), Rb.y(), Rb.right(), Rb.bottom() );
//             p->setPen( getColor(g,PanelMid) );
//             p->drawLine(Rb.right()-1, Rb.y(), Rb.right()-1, Rb.bottom() );
        }
    }
}


void LinspireClearStyle::renderMenuBlendPixmap( KPixmap &pix, const QColorGroup &cg,
    const QPopupMenu* /* popup */ ) const
{
    pix.fill( cg.background().light(105) );
}


void LinspireClearStyle::drawKStylePrimitive(KStylePrimitive kpe,
                                      QPainter *p,
                                      const QWidget* widget,
                                      const QRect &r,
                                      const QColorGroup &cg,
                                      SFlags flags,
                                      const QStyleOption& opt) const
{
    // some "global" vars...
    const bool enabled = (flags & Style_Enabled);

    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

//  SLIDER
//  ------
    switch( kpe ) {
        case KPE_SliderGroove: {
            const QSlider* slider = (const QSlider*)widget;
            bool horizontal = slider->orientation() == Horizontal;

            if (horizontal) {
                int center = r_y+r_h/2;
                renderContour(p, QRect(r_x, center-2, r_w, 4),
                              cg.background(), getColor(cg,ButtonContour,enabled) /*cg.background().dark(enabled?150:130)*/,
                              Draw_Left|Draw_Right|Draw_Top|Draw_Bottom, Edge_Soft, Edge_Soft, Edge_Soft, Edge_Soft);
                p->fillRect( r_x+1, center-1, r_w-2, 2, getColor(cg, ButtonSurfaceInactive, enabled) );
            } else {
                int center = r_x+r_w/2;
                renderContour(p, QRect(center-2, r_y, 4, r_h),
                              cg.background(), getColor(cg,ButtonContour,enabled)/*cg.background().dark(enabled?150:130)*/,
                              Draw_Left|Draw_Right|Draw_Top|Draw_Bottom, Edge_Soft, Edge_Soft, Edge_Soft, Edge_Soft);
                p->fillRect( center-1, r_y+1, 2, r_h-2, getColor(cg, ButtonSurfaceInactive, enabled) );
            }
            break;
        }

        case KPE_SliderHandle: {
                const QSlider* slider = (const QSlider*)widget;
                bool horizontal = slider->orientation() == Horizontal;

                bool hover = false;
                if( _mouseOver && hoverSubControl==SC_SliderHandle && hoverWidget==slider )
                    hover = true;

                const bool pressed = (flags&Style_Active);
                const WidgetState s = enabled?(pressed?IsPressed:(hover?IsHighlighted:IsEnabled)):IsDisabled;
                const QColor contour = getColor(cg,DragButtonContour,s),
                             surface = getColor(cg,DragButtonSurface,s);

                int xcenter = (r_x+r_x2) / 2;
                int ycenter = (r_y+r_y2) / 2;

                if (horizontal) {
                    // surface
                    QRegion mask(xcenter-4, ycenter-5, 9, 13);
                    mask -= QRegion(xcenter-4, ycenter+4, 1, 4);
                    mask -= QRegion(xcenter-3, ycenter+5, 1, 3);
                    mask -= QRegion(xcenter-2, ycenter+6, 1, 2);
                    mask -= QRegion(xcenter-1, ycenter+7, 1, 1);
                    mask -= QRegion(xcenter+1, ycenter+7, 1, 1);
                    mask -= QRegion(xcenter+2, ycenter+6, 1, 2);
                    mask -= QRegion(xcenter+3, ycenter+5, 1, 3);
                    mask -= QRegion(xcenter+4, ycenter+4, 1, 4);
                    
                    mask -= QRegion(xcenter+4, ycenter-5, 1, 1);
                    p->setClipRegion(mask);
                    p->setClipping(true);
                    uint surfaceFlags = Draw_Left|Draw_Right|Round_UpperLeft|Round_UpperRight|Draw_Symmetric/*|Draw_Soft*/;
                    if(!enabled)
                        surfaceFlags |= Is_Disabled;
                    renderSurface(p, QRect(xcenter-4, ycenter-5, 9/*+1*/, 13), surface, surfaceFlags); /*// width is 1 pixel wider... stupid hack so that the sharp line of the surface is centered.*/
                    p->setClipping(false);

                    // contour
                    renderContour(p, QRect(xcenter-5, ycenter-6, 11, 10),
                                  cg.background(), contour,
                                  Draw_Left|Draw_Right|Draw_Top, Edge_Round1, Edge_Round1);

                    // manual contour: vertex
                    p->setPen( contour );
                    p->drawPoint(xcenter-5+1, ycenter+4);
                    p->drawPoint(xcenter+5-1, ycenter+4);
                    p->drawPoint(xcenter-5+2, ycenter+5);
                    p->drawPoint(xcenter+5-2, ycenter+5);
                    p->drawPoint(xcenter-5+3, ycenter+6);
                    p->drawPoint(xcenter+5-3, ycenter+6);
                    p->drawPoint(xcenter-5+4, ycenter+7);
                    p->drawPoint(xcenter+5-4, ycenter+7);
                    // anti-aliasing of the contour... sort of. :)
                    p->setPen(alphaBlendColors(cg.background(), contour, 30) );
                    p->drawPoint(xcenter, ycenter+8);
                    p->setPen(alphaBlendColors(cg.background(), contour, 100) );
                    p->drawPoint(xcenter-5, ycenter+4);
                    p->drawPoint(xcenter+5, ycenter+4);
                    p->drawPoint(xcenter-5+1, ycenter+5);
                    p->drawPoint(xcenter+5-1, ycenter+5);
                    p->drawPoint(xcenter-5+2, ycenter+6);
                    p->drawPoint(xcenter+5-2, ycenter+6);
                    p->drawPoint(xcenter-5+3, ycenter+7);
                    p->drawPoint(xcenter+5-3, ycenter+7);
                    p->setPen(alphaBlendColors(cg.background(), contour, 140) );
                    p->drawPoint(xcenter-5+4, ycenter+8);
                    p->drawPoint(xcenter+5-4, ycenter+8);
                } else {
                    // surface
                    QRegion mask(xcenter-5, ycenter-4, 13, 9);
                    mask -= QRegion(xcenter+4, ycenter-4, 4, 1);
                    mask -= QRegion(xcenter+5, ycenter-3, 3, 1);
                    mask -= QRegion(xcenter+6, ycenter-2, 2, 1);
                    mask -= QRegion(xcenter+7, ycenter-1, 1, 1);
                    mask -= QRegion(xcenter+7, ycenter+1, 1, 1);
                    mask -= QRegion(xcenter+6, ycenter+2, 2, 1);
                    mask -= QRegion(xcenter+5, ycenter+3, 3, 1);
                    mask -= QRegion(xcenter+4, ycenter+4, 4, 1);

                    mask -= QRegion(xcenter-5, ycenter+4, 1, 1);
                    p->setClipRegion(mask);
                    p->setClipping(true);
                    uint surfaceFlags = Draw_Top|Draw_Bottom|Round_UpperLeft|Round_BottomLeft|
                                    Round_UpperRight|Is_Horizontal|Draw_Symmetric/*|Draw_Soft*/;
                    if(!enabled)
                        surfaceFlags |= Is_Disabled;
                    renderSurface(p, QRect(xcenter-5, ycenter-4, 13, 9/*+1*/), surface, surfaceFlags); /*// height is 1 pixel higher... stupid hack so that the sharp line of the surface is centered.*/
                    p->setClipping(false);

                    // contour
                    renderContour(p, QRect(xcenter-6, ycenter-5, 10, 11),
                                  cg.background(), contour,
                                  Draw_Left|Draw_Top|Draw_Bottom, Edge_Round1, Edge_Round2, Edge_Round1);

                    // manual contour: vertex
                    p->setPen( contour );
                    p->drawPoint(xcenter+4, ycenter-5+1);
                    p->drawPoint(xcenter+4, ycenter+5-1);
                    p->drawPoint(xcenter+5, ycenter-5+2);
                    p->drawPoint(xcenter+5, ycenter+5-2);
                    p->drawPoint(xcenter+6, ycenter-5+3);
                    p->drawPoint(xcenter+6, ycenter+5-3);
                    p->drawPoint(xcenter+7, ycenter-5+4);
                    p->drawPoint(xcenter+7, ycenter+5-4);
                    // anti-aliasing. ...sort of :)
                    p->setPen(alphaBlendColors(cg.background(), contour, 30) );
                    p->drawPoint(xcenter+8, ycenter);
                    p->setPen(alphaBlendColors(cg.background(), contour, 100) );
                    p->drawPoint(xcenter+4, ycenter-5);
                    p->drawPoint(xcenter+4, ycenter+5);
                    p->drawPoint(xcenter+5, ycenter-5+1);
                    p->drawPoint(xcenter+5, ycenter+5-1);
                    p->drawPoint(xcenter+6, ycenter-5+2);
                    p->drawPoint(xcenter+6, ycenter+5-2);
                    p->drawPoint(xcenter+7, ycenter-5+3);
                    p->drawPoint(xcenter+7, ycenter+5-3);
                    p->setPen(alphaBlendColors(cg.background(), contour, 140) );
                    p->drawPoint(xcenter+8, ycenter-5+4);
                    p->drawPoint(xcenter+8, ycenter+5-4);
                }

                break;
            }

        case KPE_ListViewExpander: {
            int radius = (r_w - 4) / 2;
            int centerx = r_x + r_w/2;
            int centery = r_y + r_h/2;

            renderContour(p, r, cg.base(), cg.mid().dark(115)/*, Draw_Left|Draw_Right|Draw_Top|Draw_Bottom, */);

            // plus or minus
            p->setPen( cg.text() );
            p->drawLine( centerx - radius, centery, centerx + radius, centery );
            if ( flags & Style_On ) // Collapsed = On
                p->drawLine( centerx, centery - radius, centerx, centery + radius );

            break;
        }

        default:
            KStyle::drawKStylePrimitive(kpe, p, widget, r, cg, flags, opt);
    }
}


void LinspireClearStyle::drawPrimitive(PrimitiveElement pe,
                                QPainter *p,
                                const QRect &r,
                                const QColorGroup &cg,
                                SFlags flags,
                                const QStyleOption &opt ) const
{
    bool down   = flags & Style_Down;
    bool on     = flags & Style_On;
    bool sunken = flags & Style_Sunken;
    bool horiz  = flags & Style_Horizontal;
    const bool enabled = flags & Style_Enabled;
    const bool mouseOver = flags & Style_MouseOver;
    const bool hasFocus = flags & Style_HasFocus;

    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    switch(pe) {
    // BUTTONS
    // -------
        case PE_FocusRect: {
            p->drawWinFocusRect( r );
            break;
        }

        case PE_HeaderSection: {
                // little hack: since toolbars seem to be painted disabled all the time in kicker, set enabled
                // always to true when we are in kickerMode
                bool enabled2 = enabled;
                if(kickerMode) enabled2 = true;

                bool hover = false;
                if ( hoverHeaderCurrent == p->device() ) {
                    int id = hoverHeaderCurrent->sectionAt( r_x );
                    hover = (id != -1) && (id == hoverHeaderId);
                }

                const QPen oldPen( p->pen() );

                uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
                // Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight;
                if(!enabled2) contourFlags|=Is_Disabled;
                renderContour(p, r, cg.background(),
                        hover?(getColor(cg,ButtonContour).light(100+(int)ANIMATIONSTEPS)):getColor(cg,ButtonContour),
                        contourFlags, Edge_Plain, Edge_Plain, Edge_Plain, Edge_Plain);

                uint surfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Draw_Soft;
                if(_symmetricButtonSurface)
                    surfaceFlags|=Draw_Symmetric;
                // Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight;
                /*if(horizontal) */surfaceFlags|=Is_Horizontal;
                if(!enabled2) surfaceFlags|=Is_Disabled;
                else if(on||down) surfaceFlags|=Is_Sunken;
                renderSurface(p, QRect(r_x+1, r_y+1, r_w-2, r_h-2),
                        hover?(cg.button().light(100+(int)ANIMATIONSTEPS)):cg.button(),
                        surfaceFlags);

                p->setPen(oldPen);
            break;
        }
        case PE_ButtonBevel:
        case PE_ButtonTool:
        case PE_ButtonDropDown:
        case PE_ButtonCommand: {
            int animStep = 0;
            if( _mouseOverAnimation && animWidgets.contains(opt.widget()) ) {
                animStep = (int)animWidgets[opt.widget()].step;
            } else if ( _mouseOver && (mouseOver || opt.widget()==hoverWidget) ) {
                animStep = (int)ANIMATIONSTEPS;
            }
            bool khtmlMode = khtmlWidgets.contains(opt.widget());
            renderButton(p, r, cg, (on||down), mouseOver, true, enabled, khtmlMode, animStep );
            break;
        }

        case PE_ButtonDefault: {
            p->fillRect(r, cg.background());
            const WidgetState s = enabled?(sunken?IsPressed:IsEnabled):IsDisabled;
            p->setPen( getColor(cg,ButtonContour,s) );
            // sides
            p->drawLine( r_x+3, r_y, r_x2-3, r_y );
            p->drawLine( r_x+3, r_y2, r_x2-3, r_y2 );
            p->drawLine( r_x, r_y+3, r_x, r_y2-3 );
            p->drawLine( r_x2, r_y+3, r_x2, r_y2-3 );
            // top left edge
            p->drawPoint( r_x+1, r_y+2 );
            p->drawPoint( r_x+2, r_y+1 );
            // top right edge
            p->drawPoint( r_x2-1, r_y+2 );
            p->drawPoint( r_x2-2, r_y+1 );
            // bottom left edge
            p->drawPoint( r_x+1, r_y2-2 );
            p->drawPoint( r_x+2, r_y2-1 );
            // top right edge
            p->drawPoint( r_x2-1, r_y2-2 );
            p->drawPoint( r_x2-2, r_y2-1 );

            break;
        }

        case PE_SpinWidgetPlus:
        case PE_SpinWidgetMinus: {
// use previously set pen...
//             p->setPen( cg.buttonText() );

            int l = QMIN( r_w-3, r_h-3 );
            // make the length even so that we get a nice symmetric plus...
            if(l%2 != 0)
                --l;
            QPoint c = r.center();

            p->drawLine( c.x()-l/2, c.y(), c.x()+l/2, c.y() );
            if ( pe == PE_SpinWidgetPlus ) {
                p->drawLine( c.x(), c.y()-l/2, c.x(), c.y()+l/2 );
            }
            break;
        }

        case PE_ScrollBarSlider: {
            bool hover = false;
            if( _mouseOver && hoverSubControl==SC_ScrollBarSlider && hoverWidget==dynamic_cast<QWidget*>(p->device()) )
                hover = true;
            const WidgetState s = enabled?(down?IsPressed:(hover?IsHighlighted:IsEnabled)):IsDisabled;

            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            if(!enabled) contourFlags|=Is_Disabled;
            renderContour(p, r, cg.background(), getColor(cg, DragButtonContour, s),
                    contourFlags, Edge_Plain, Edge_Plain, Edge_Plain, Edge_Plain);

            uint surfaceFlags = Draw_Symmetric|Draw_Soft/* = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom*/;
            if(horiz) surfaceFlags|=Is_Horizontal;
            if(!enabled) surfaceFlags|=Is_Disabled;
//             if(r_h >= 4)
                renderSurface(p, QRect(r_x+1, r_y+1, r_w-2, r_h-2),
                        getColor(cg, DragButtonSurface, s), surfaceFlags);

            break;
        }

        case PE_ScrollBarAddPage:
        case PE_ScrollBarSubPage: {
            QColor base = (on||down) ?
                    getColor(cg, ButtonSurfaceInactive, enabled).dark(100+2*_contrast)
                    : getColor(cg, ButtonSurfaceInactive, enabled);

            if(flags & Style_Horizontal) {
                p->setPen( getColor(cg, DragButtonContour, enabled) );
                p->drawLine(r_x, r_y, r_x2, r_y);
                p->drawLine(r_x, r_y2, r_x2, r_y2);
                p->fillRect(r_x, r_y+1, r_w, r_h-2, base);
            } else {
                p->setPen( getColor(cg, DragButtonContour, enabled) );
                p->drawLine(r_x, r_y, r_x, r_y2);
                p->drawLine(r_x2, r_y, r_x2, r_y2);
                p->fillRect(r_x+1, r_y, r_w-2, r_h, base );
            }

//             p->fillRect(r, QBrush(cg.background().light(), Dense4Pattern));
            break;
        }

    // SCROLLBAR BUTTONS
    // -----------------
        case PE_ScrollBarSubLine: {
            bool hover = false;
            if( _mouseOver && hoverSubControl==SC_ScrollBarSubLine && hoverWidget==dynamic_cast<QWidget*>(p->device()) )
                hover = true;
            const WidgetState s = enabled?(down?IsPressed:(hover?IsHighlighted:IsEnabled)):IsDisabled;
            EdgeType e1 = Edge_Plain, e2 = Edge_Plain, e3 = Edge_Plain, e4 = Edge_Plain;
            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            uint surfaceFlags = /*Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|*/Draw_Soft|Draw_Symmetric;
            if(down) surfaceFlags|=Is_Sunken;
            if(!enabled) {
                contourFlags|=Is_Disabled;
                surfaceFlags|=Is_Disabled;
            }
            if(horiz) {
//                 e1 = e3 = Edge_Round2;
                surfaceFlags |= /*Round_UpperLeft|Round_BottomLeft|*/Is_Horizontal;
//             } else {
//                 e1 = e2 = Edge_Round2;
//                 surfaceFlags |= Round_UpperLeft|Round_UpperRight;
            }
            renderContour(p, r, cg.background(), getColor(cg, DragButtonContour, s),
                    contourFlags, e1, e2, e3, e4);
            renderSurface(p, QRect(r_x+1, r_y+1, r_w-2, r_h-2),
                    getColor(cg, DragButtonSurface, s), surfaceFlags);

//             QImage mask = uic_findImage(horiz?"scroll_arrow_left.png":"scroll_arrow_up.png");
//             int dX = (r_w-mask.width())/2,
//                 dY = (r_h-mask.height())/2;
//             if(down)
//                 ++dY;
//             for(int x = 0; x < mask.width(); ++x) {
//                 for(int y = 0; y < mask.height(); ++y) {
//                     renderPixel(p, r_x+dX+x, r_y+dY+y, qAlpha(mask.pixel(x,y)),
//                             cg.foreground()/*getColor(cg, ButtonArrow)*/, cg.background(), true);
//                 }
//             }

            p->setPen(cg.foreground());
            drawPrimitive((horiz ? PE_ArrowLeft : PE_ArrowUp), p, r, cg, flags);
            break;
        }

        case PE_ScrollBarAddLine: {
            bool hover = false;
            if( _mouseOver && hoverSubControl==SC_ScrollBarAddLine && hoverWidget==dynamic_cast<QWidget*>(p->device()) )
                hover = true;
            const WidgetState s = enabled?(down?IsPressed:(hover?IsHighlighted:IsEnabled)):IsDisabled;
            EdgeType e1 = Edge_Plain, e2 = Edge_Plain, e3 = Edge_Plain, e4 = Edge_Plain;
            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            uint surfaceFlags = /*Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|*/Draw_Soft|Draw_Symmetric;
            if(down) surfaceFlags|=Is_Sunken;
            if(!enabled) {
                contourFlags|=Is_Disabled;
                surfaceFlags|=Is_Disabled;
            }
            if(horiz) {
//                 e2 = e4 = Edge_Round2;
                surfaceFlags |= /*Round_UpperRight|Round_BottomRight|*/Is_Horizontal;
//             } else {
//                 e3 = e4 = Edge_Round2;
//                 surfaceFlags |= Round_BottomLeft|Round_BottomRight;
            }
            renderContour(p, r, cg.background(), getColor(cg, DragButtonContour, s),
                    contourFlags, e1, e2, e3, e4);
            renderSurface(p, QRect(r_x+1, r_y+1, r_w-2, r_h-2),
                    getColor(cg, DragButtonSurface, s), surfaceFlags);

//             QImage mask = uic_findImage(horiz?"scroll_arrow_right.png":"scroll_arrow_down.png");
//             int dX = (r_w-mask.width())/2,
//                 dY = (r_h-mask.height())/2;
//             if(down)
//                 ++dY;
//             for(int x = 0; x < mask.width(); ++x) {
//                 for(int y = 0; y < mask.height(); ++y) {
//                     renderPixel(p, r_x+dX+x, r_y+dY+y, qAlpha(mask.pixel(x,y)),
//                             cg.foreground()/*getColor(cg, ButtonArrow)*/, cg.background(), true);
//                 }
//             }

            p->setPen(cg.foreground());
            drawPrimitive((horiz ? PE_ArrowRight : PE_ArrowDown), p, r, cg, flags);
            break;
        }

    // CHECKBOXES
    // ----------
        case PE_Indicator: {
            QColor cContour = getColor(cg, ButtonContour, enabled);
            QColor cSurface = enabled?cg.base():cg.background();
            if(flags&Style_MouseOver) {
                cContour = cContour.dark(100+2*_contrast);
                cSurface = cSurface.dark(100+2*_contrast);
            }

            QBitmap bmp;
            bmp = QBitmap(checkbox_contour1_width, checkbox_contour1_height, checkbox_contour1_bits, true);
            bmp.setMask(bmp);
            p->setPen( cContour.dark(120) );
            p->drawPixmap( r_x, r_y, bmp );
            bmp = QBitmap(checkbox_contour2_width, checkbox_contour2_height, checkbox_contour2_bits, true);
            bmp.setMask(bmp);
            p->setPen( cContour.light(110) );
            p->drawPixmap( r_x, r_y, bmp );
            bmp = QBitmap(checkbox_surface_width, checkbox_surface_height, checkbox_surface_bits, true);
            bmp.setMask(bmp);
            p->setPen( cSurface );
            p->drawPixmap( r_x, r_y, bmp );

            drawPrimitive(PE_CheckMark, p, r, cg, flags);

            break;
        }

        case PE_IndicatorMask: {
            p->fillRect(r, color0);

            QBitmap bmp;
            bmp = QBitmap(checkbox_mask_width, checkbox_mask_height, checkbox_mask_bits, true);
            bmp.setMask(bmp);
            p->setPen(color1);
            p->drawPixmap(r_x, r_y, bmp);
            break;
        }

    // RADIOBUTTONS
    // ------------
        case PE_ExclusiveIndicator: {

            QColor cContour = getColor(cg, ButtonContour, enabled);
            QColor cSurface = enabled?cg.base():cg.background();
            if(flags&Style_MouseOver) {
                cContour = cContour.dark(100+2*_contrast);
                cSurface = cSurface.dark(100+2*_contrast);
            }

            QBitmap bmp;
            bmp = QBitmap(radiobutton_contour1_width, radiobutton_contour1_height, radiobutton_contour1_bits, true);
            bmp.setMask(bmp);
            p->setPen( cContour.dark(120) );
            p->drawPixmap( r_x, r_y, bmp );
            bmp = QBitmap(radiobutton_contour2_width, radiobutton_contour2_height, radiobutton_contour2_bits, true);
            bmp.setMask(bmp);
            p->setPen( cContour.light(110) );
            p->drawPixmap( r_x, r_y, bmp );
            bmp = QBitmap(radiobutton_surface_width, radiobutton_surface_height, radiobutton_surface_bits, true);
            bmp.setMask(bmp);
            p->setPen( cSurface );
            p->drawPixmap( r_x, r_y, bmp );

            break;
        }

        case PE_ExclusiveIndicatorMask: {
            p->fillRect(r, color0);

            QBitmap bmp;
            bmp = QBitmap(radiobutton_mask_width, radiobutton_mask_height, radiobutton_mask_bits, true);
            bmp.setMask(bmp);
            p->setPen(color1);
            p->drawPixmap(r_x, r_y, bmp);

            break;
        }

    // GENERAL PANELS
    // --------------
        case PE_Splitter: {
            // highlight on mouse over
            QColor color = (hoverWidget == p->device())?cg.background().dark(100+_contrast):cg.background();
            p->fillRect(r, color);
            if (r_w > r_h) {
                if (r_h > 4) {
                    int ycenter = r_h/2;
                    for(int k = 2*r_w/10; k < 8*r_w/10; k+=5) {
                        renderDot(p, k,ycenter-1, color, false, true);
                    }
                }
            } else {
                if (r_w > 4) {
                    int xcenter = r_w/2;
                    for(int k = 2*r_h/10; k < 8*r_h/10; k+=5) {
                        renderDot(p, xcenter-1,k, color, false, true);
                    }
                }
            }

                break;
        }

        case PE_PanelGroupBox:
        case PE_GroupBoxFrame: {
                renderPanel(p, r, cg, false);

            break;
        }

        case PE_WindowFrame: {
            renderContour(p, r, cg.background(), getColor(cg, PanelContour, enabled),
                    Draw_Left|Draw_Right|Draw_Top|Draw_Bottom, Edge_Plain, Edge_Plain, Edge_Plain, Edge_Plain );
            p->setPen( getColor(cg, PanelLight, enabled) );
            p->drawLine(r_x+1, r_y+1, r_x2-1, r_y+1 );
            p->setPen( getColor(cg, PanelMid, enabled) );
            p->drawLine(r_x+1, r_y+2, r_x+1, r_y2-2 );
            p->drawLine(r_x2-1, r_y+2, r_x2-1, r_y2-2 );
            p->setPen( getColor(cg, PanelDark, enabled) );
            p->drawLine(r_x+1, r_y2-1, r_x2-1, r_y2-1 );
            break;
        }

        case PE_Panel: {
            renderPanel(p, r, cg, true, sunken);
            break;
        }

        case PE_PanelLineEdit: {
            bool isReadOnly = false;
            bool isEnabled = true;
            if (p->device() && dynamic_cast<QLineEdit*>(p->device()))
            {
                QLineEdit* lineEdit = dynamic_cast<QLineEdit*>(p->device());
                isReadOnly = lineEdit->isReadOnly();
                isEnabled = lineEdit->isEnabled();
            }

            const QColor contentColor = enabled?cg.base():cg.background();
            if (_inputFocusHighlight && hasFocus && !isReadOnly && isEnabled)
                p->setPen(contentColor.dark(135) );
            else
                p->setPen(contentColor );
            p->drawLine(r_x+1, r_y+2, r_x+1, r_y2-2 );
            p->drawLine(r_x+2, r_y+1, r_x2-2, r_y+1 );
            if (_inputFocusHighlight && hasFocus && !isReadOnly && isEnabled)
                p->setPen(contentColor.dark(135) );
            else
                p->setPen(contentColor );
            p->drawLine(r_x+2, r_y2-1, r_x2-2, r_y2-1 );
            p->drawLine(r_x2-1, r_y+2, r_x2-1, r_y2-2 );
            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            if(!enabled) contourFlags|=Is_Disabled;

            // HACK!!
            //
            // In order to draw nice edges in khtml, we need to paint alpha-blended.
            // On the other hand, we can't paint alpha-blended in normal widgets.
            //
            // In this place there is no reliable way to detect if we are in khtml; the
            // only thing we know is that khtml buffers its widgets into a pixmap. So
            // when the paint device is a QPixmap, chances are high that we are in khtml.
            // It's possible that this breaks other things, so let's see how it works...
            if (p->device() && dynamic_cast<QPixmap*>(p->device() ) ) {
                contourFlags += Draw_AlphaBlend;
            }

            renderContour(p, r, cg.background(), getColor(cg, ButtonContour, enabled), contourFlags, Edge_Round3, Edge_Round3, Edge_Round3, Edge_Round3 );
            break;
        }

        case PE_StatusBarSection: {
            renderContour(p, r, cg.background(), cg.background().dark(130),
                          Draw_Left|Draw_Right|Draw_Top|Draw_Bottom, Edge_Soft, Edge_Soft, Edge_Soft, Edge_Soft);
            break;
        }

        case PE_TabBarBase: // Still not sure what this one does, seems not to be necessary
            break;
        case PE_PanelTabWidget: {
//             renderPanel(p, r, cg, true, sunken, true);
            // panel
            p->setPen( getColor(cg, TabPanelContour, enabled) );
            p->drawLine(r_x, r_y, r_x2, r_y );
            p->drawLine(r_x, r_y+1, r_x, r_y2-1 );
            p->drawLine(r_x2, r_y+1, r_x2, r_y2-1 );
            p->drawLine(r_x, r_y2, r_x2, r_y2 );
            // inner highlight...
            p->setPen( getColor(cg, PanelLight, enabled) );
            p->drawLine(r_x+1, r_y+1, r_x2-1, r_y+1 );
            p->setPen( getColor(cg, PanelMid, enabled) );
            p->drawLine(r_x+1, r_y+2, r_x+1, r_y2-2 );
            p->drawLine(r_x2-1, r_y+2, r_x2-1, r_y2-2 );
            p->setPen( getColor(cg, PanelDark, enabled) );
            p->drawLine(r_x+1, r_y2-1, r_x2-1, r_y2-1 );
            break;
        }

        case PE_PanelPopup: {
            renderContour(p, r, cg.background(), getColor(cg, PanelContour),
                          Draw_Left|Draw_Right|Draw_Top|Draw_Bottom, Edge_Plain, Edge_Plain, Edge_Plain, Edge_Plain);
            break;
        }

    // MENU / TOOLBAR PANEL
    // --------------------
        case PE_PanelMenuBar:
        case PE_PanelDockWindow: {
            // fix for toolbar lag (from Mosfet Liquid) 
            QWidget* w = dynamic_cast<QWidget*>(p->device());
            if(w && w->backgroundMode() == PaletteButton) 
                w->setBackgroundMode(PaletteBackground);
            p->fillRect(r, cg.brush(QColorGroup::Background));

            if ( _drawToolBarSeparator ) {
                if ( r_w > r_h ) {
                    p->setPen( getColor(cg, PanelLight) );
                    p->drawLine( r_x, r_y, r_x2, r_y );
                    p->setPen( getColor(cg, PanelDark) );
                    p->drawLine( r_x, r_y2, r_x2, r_y2 );
                }
                else {
                    p->setPen( getColor(cg, PanelLight) );
                    p->drawLine( r_x, r_y, r_x, r_y2 );
                    p->setPen( getColor(cg, PanelDark) );
                    p->drawLine( r_x2, r_y, r_x2, r_y2 );
                }
            }

            break;
        }

    // TOOLBAR/DOCK WINDOW HANDLE
    // --------------------------
        case PE_DockWindowResizeHandle: {
            renderButton(p, r, cg);
            break;
        }

        case PE_DockWindowHandle: {

            p->fillRect(r, cg.brush(QColorGroup::Background));

            int counter = 1;

            if(horiz) {
                int center = r_x+r_w/2;
                for(int j = r_y+2; j <= r_y2-3; j+=3) {
                    if(counter%2 == 0) {
                        renderDot(p, center+1,j, cg.background(), true, true);
                    } else {
                        renderDot(p, center-2,j, cg.background(), true, true);
                    }
                    counter++;
                }
            } else {
                int center = r_y+r_h/2;
                for(int j = r_x+2; j <= r_x2-3; j+=3) {
                    if(counter%2 == 0) {
                        renderDot(p, j,center+1, cg.background(), true, true);
                    } else {
                        renderDot(p, j,center-2, cg.background(), true, true);
                    }
                    counter++;
                }
            }

            break;
        }

    // TOOLBAR SEPARATOR
    // -----------------
        case PE_DockWindowSeparator: {
            // fix for toolbar lag
            QWidget* w = dynamic_cast<QWidget*>(p->device());
            if(w && w->backgroundMode() == PaletteButton) 
                w->setBackgroundMode(PaletteBackground);
            p->fillRect(r, cg.background());

            if(horiz) {
                int center = r_x+r_w/2;
                p->setPen( getColor(cg, PanelDark) );
                p->drawLine( center-1, r_y+3, center-1, r_y2-3 );
                p->setPen( getColor(cg, PanelLight) );
                p->drawLine( center, r_y+3, center, r_y2-3 );
            } else {
                int center = r_y+r_h/2;
                p->setPen( getColor(cg, PanelDark) );
                p->drawLine( r_x+3, center-1, r_x2-3, center-1 );
                p->setPen( getColor(cg, PanelLight) );
                p->drawLine( r_x+3, center, r_x2-3, center );
            }
            break;
        }

        case PE_CheckMark: {
            const QColor contentColor = enabled?cg.base():cg.background();
            QColor checkmarkColor = enabled?cg.foreground():cg.mid();
            if(flags & Style_Down) {
                checkmarkColor = alphaBlendColors(contentColor, checkmarkColor, 150);
            }

            if( flags & Style_On ) {
                int x = r.center().x() - checkmark_width/2,
                    y = r.center().y() - checkmark_height/2;
                QBitmap bmp = QBitmap(checkmark_width, checkmark_height, checkmark_bits, true);
                bmp.setMask(bmp);
                p->setPen( checkmarkColor );
                p->drawPixmap(x, y, bmp);
            } else if ( flags & Style_Off ) {
                // empty
            } else { // tristate
                int x = r.center().x() - checkmark_tristate_width/2,
                    y = r.center().y() - checkmark_tristate_height/2;
                QBitmap bmp = QBitmap(checkmark_tristate_width, checkmark_tristate_height, checkmark_tristate_bits, true);
                bmp.setMask(bmp);
                p->setPen( checkmarkColor );
                p->drawPixmap(x, y, bmp);
            }

            break;
        }

        case PE_SpinWidgetUp:
        case PE_SpinWidgetDown:
        case PE_HeaderArrow:
        case PE_ArrowUp:
        case PE_ArrowDown:
        case PE_ArrowLeft:
        case PE_ArrowRight: {
            QPointArray a;

            switch (pe) {
                case PE_SpinWidgetUp:
                case PE_ArrowUp: {
                    a.setPoints(7, u_arrow);
                    break;
                }
                case PE_SpinWidgetDown:
                case PE_ArrowDown: {
                    a.setPoints(7, d_arrow);
                    break;
                }
                case PE_ArrowLeft: {
                    a.setPoints(7, l_arrow);
                    break;
                }
                case PE_ArrowRight: {
                    a.setPoints(7, r_arrow);
                    break;
                }
                default: {
                    if (flags & Style_Up) {
                        a.setPoints(7, u_arrow);
                    } else {
                        a.setPoints(7, d_arrow);
                    }
                }
            }

            const QWMatrix oldMatrix( p->worldMatrix() );

            if (flags & Style_Down) {
                p->translate(pixelMetric(PM_ButtonShiftHorizontal),
                                pixelMetric(PM_ButtonShiftVertical));
            }

            a.translate((r_x+r_w/2), (r_y+r_h/2));
            // extra-pixel-shift, correcting some visual tics...
            switch(pe) {
                case PE_ArrowLeft:
                case PE_ArrowRight:
                    a.translate(0, -1);
                    break;
                case PE_SpinWidgetUp:
                case PE_SpinWidgetDown:
                    a.translate(+1, 0);
                    break;
                default:
                    a.translate(0, 0);
            }

            if (p->pen() == QPen::NoPen) {
                if (flags & Style_Enabled) {
                    p->setPen(cg.buttonText());
                } else {
                    p->setPen(cg.highlightedText());
                }
            }
            p->drawLineSegments(a, 0, 3);
            p->drawPoint(a[6]);

            p->setWorldMatrix( oldMatrix );
            break;
        }

        default: {
            return KStyle::drawPrimitive(pe, p, r, cg, flags, opt);
        }
    }
}


void LinspireClearStyle::drawControl(ControlElement element,
                              QPainter *p,
                              const QWidget *widget,
                              const QRect &r,
                              const QColorGroup &cg,
                              SFlags flags,
                              const QStyleOption& opt) const
{
    const bool reverseLayout = QApplication::reverseLayout();

    const bool enabled = (flags & Style_Enabled);

    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    switch (element) {

    // PROGRESSBAR
    // -----------
        case CE_ProgressBarGroove: {
            const QColor content = enabled?cg.base():cg.background();
            p->setPen(content/*.dark(105)*/ );
            p->drawLine(r_x+1, r_y+1, r_x2-1, r_y+1 );
            p->drawLine(r_x+1, r_y+2, r_x+1, r_y2-2);
            p->setPen(content/*.light(105)*/ );
            p->drawLine(r_x+1, r_y2-1, r_x2-1, r_y2-1 );
            p->drawLine(r_x2-1, r_y+2, r_x2-1, r_y2-2);
            renderContour(p, r, cg.background(), getColor(cg, ButtonContour, enabled), Draw_Left|Draw_Right|Draw_Top|Draw_Bottom,
                    Edge_Plain, Edge_Plain, Edge_Plain, Edge_Plain );
            break;
        }

        case CE_ProgressBarContents: {
            const QProgressBar *pb = dynamic_cast<const QProgressBar*>(widget);
            int steps = pb->totalSteps();

            const QColor bg = enabled?cg.base():cg.background(); // background
            const QColor fg = /*enabled?*/cg.button()/*:cg.background().dark(110)*/; // foreground

            if( steps == 0 ) { // Busy indicator
                static const int barWidth = 10;
                int progress = pb->progress() % (2*(r_w-barWidth));
                if( progress < 0)
                    progress = 0;
                if( progress > r_w-barWidth )
                    progress = (r_w-barWidth)-(progress-(r_w-barWidth));
                p->fillRect( QRect( r_x, r_y, r_w, r_h ), bg );
                uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
                if(!enabled)
                    contourFlags|=Is_Disabled;
                renderContour( p, QRect( r_x+progress, r_y, barWidth, r_h ),
                                bg, fg.dark(140),
                                contourFlags, Edge_Plain, Edge_Plain, Edge_Plain, Edge_Plain );
                uint surfaceFlags = /*Draw_Right|Draw_Left|Draw_Top|Draw_Bottom|*/Draw_Symmetric|Is_Horizontal|Draw_Soft;
                if(!enabled)
                    surfaceFlags|=Is_Disabled;
                renderSurface(p, QRect( r_x+progress+1, r_y+1, barWidth-2, r_h-2 ), fg,
                                    surfaceFlags);
            } else {
                double percent = static_cast<double>(pb->progress()) / static_cast<double>(steps);

                int w = static_cast<int>(r_w * percent);
                // renderContour/renderSurface handle small sizes not very well, so set a minimal
                // progressbar width...
                if(w<4) w = 4;
                int w2 = r_w-(r_w-w);

                QRect Rempty(reverseLayout?r_x:r_x+w-1, r_y, r_w-w+1, r_h );
                QRect Rcontour(reverseLayout?r_x2-w2+1:r_x, r_y, w2, r_h );
                QRect Rsurface(Rcontour.x()+1, Rcontour.y()+1, w2-2, Rcontour.height()-2);

                p->fillRect(Rempty, bg);

                uint contourFlags = reverseLayout ? Draw_Left|Draw_Right|Draw_Top|Draw_Bottom
                                : Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
                if(!enabled)
                    contourFlags|=Is_Disabled;
                renderContour(p, Rcontour,
                              bg, fg.dark(140),
                              contourFlags,
                              Edge_Plain, Edge_Plain, Edge_Plain, Edge_Plain);
                uint sourfaceFlags = /*Draw_Right|Draw_Left|Draw_Top|Draw_Bottom|*/Draw_Symmetric|Is_Horizontal|Draw_Soft;
                if(!enabled)
                    sourfaceFlags|=Is_Disabled;
                renderSurface(p, Rsurface, fg, sourfaceFlags);
            }

            break;
        }

        case CE_ProgressBarLabel: {
            // use QCommonStyle's progress bar label instead of KStyle's one
            QCommonStyle::drawControl(element, p, widget, r, cg, flags, opt);
            break;
        }

    // RADIOBUTTONS
    // ------------
        case CE_RadioButton: {
            drawPrimitive(PE_ExclusiveIndicator, p, r, cg, flags);

            const QColor contentColor = enabled?cg.base():cg.background();
            QColor radiomarkColor = enabled?cg.foreground():cg.mid();
            if(flags & Style_Down) {
                radiomarkColor = alphaBlendColors(contentColor, radiomarkColor, 150);
            }

            if (flags & Style_On || flags & Style_Down) {
                int x = r.center().x() - radiomark_width/2,
                    y = r.center().y() - radiomark_height/2;
                QBitmap bmp;
                bmp = QBitmap(radiomark_width, radiomark_height, radiomark_bits, true);
                bmp.setMask(bmp);
                p->setPen( radiomarkColor );
                p->drawPixmap(x, y, bmp);
            }

            break;
        }

    // TABS
    // ----
        case CE_TabBarTab: {
            const QTabBar * tb = (const QTabBar *) widget;
            bool cornerWidget = false;
            if( ::qt_cast<QTabWidget*>(tb->parent()) ) {
                const QTabWidget *tw = (const QTabWidget*)tb->parent();
                // is there a corner widget in the (top) left edge?
                QWidget *cw = tw->cornerWidget(Qt::TopLeft);
                if(cw)
                    cornerWidget = true;
            }
            QTabBar::Shape tbs = tb->shape();
            bool selected = false;
            if (flags & Style_Selected) selected = true;
            TabPosition pos;
            if (tb->count() == 1) {
                pos = Single;
            } else if ((tb->indexOf(opt.tab()->identifier()) == 0)) {
                pos = First;
            } else if (tb->indexOf(opt.tab()->identifier()) == tb->count() - 1) {
                pos = Last;
            } else {
                pos = Middle;
            }

            bool hover = enabled && (flags&Style_MouseOver);

            switch (tbs) {
                case QTabBar::TriangularAbove:
                    renderTab(p, r, cg, hover, selected, false, pos, /*true, */cornerWidget);
                    break;
                case QTabBar::RoundedAbove:
                    renderTab(p, r, cg, hover, selected, false, pos, /*false, */cornerWidget);
                    break;
                case QTabBar::TriangularBelow:
                    renderTab(p, r, cg, hover, selected, true, pos, /*true, */cornerWidget);
                    break;
                case QTabBar::RoundedBelow:
                    renderTab(p, r, cg, hover, selected, true, pos, /*false, */cornerWidget);
                    break;
                    default:
                            KStyle::drawControl(element, p, widget, r, cg, flags, opt);
            }

            break;
        }

        case CE_PushButton: {
            QPushButton *button = (QPushButton *)widget;

            const bool isDefault = enabled && button->isDefault();

            if (widget == hoverWidget)
                flags |= Style_MouseOver;

            // draw default button border...
            if (isDefault ) {
                drawPrimitive(PE_ButtonDefault, p, r, cg, flags);
            }

            // **hack**
            // pass the widget as opt argument so that drawPrimitive can access the AnimInfo from the QMap...
            drawPrimitive(PE_ButtonBevel, p,
                    isDefault?QRect(r_x+2,r_y+2,r_w-2*2,r_h-2*2):r,
                    cg/*g2*/, flags, QStyleOption(button) );

            break;
        }

        case CE_PushButtonLabel:
        {
            int x, y, w, h;
            r.rect( &x, &y, &w, &h );

            const QPushButton* button = static_cast<const QPushButton *>( widget );
            bool active = button->isOn() || button->isDown();
            bool cornArrow = false;

            // Shift button contents if pushed.
            if ( active )
            {
                x += pixelMetric(PM_ButtonShiftHorizontal, widget);
                y += pixelMetric(PM_ButtonShiftVertical, widget);
                flags |= Style_Sunken;
            }

            // Does the button have a popup menu?
            if ( button->isMenuButton() )
            {
                int dx = pixelMetric( PM_MenuButtonIndicator, widget );
                if ( button->iconSet() && !button->iconSet()->isNull()  &&
                    (dx + button->iconSet()->pixmap (QIconSet::Small, QIconSet::Normal, QIconSet::Off ).width()) >= w )
                {
                    cornArrow = true; //To little room. Draw the arrow in the corner, don't adjust the widget
                }
                else
                {
                    drawPrimitive( PE_ArrowDown, p, visualRect( QRect(x + w - dx - 8, y + 2, dx, h - 4), r ),
                                cg, flags, opt );
                    w -= dx;
                }
            }

            // Draw the icon if there is one
            if ( button->iconSet() && !button->iconSet()->isNull() )
            {
                QIconSet::Mode  mode  = QIconSet::Disabled;
                QIconSet::State state = QIconSet::Off;

                if (button->isEnabled())
                    mode = button->hasFocus() ? QIconSet::Active : QIconSet::Normal;
                if (button->isToggleButton() && button->isOn())
                    state = QIconSet::On;

                QPixmap pixmap = button->iconSet()->pixmap( QIconSet::Small, mode, state );

                if (button->text().isEmpty() && !button->pixmap())
                    p->drawPixmap( x + w/2 - pixmap.width()/2, y + h / 2 - pixmap.height() / 2,
                                    pixmap );
                else
                    p->drawPixmap( x + 4, y + h / 2 - pixmap.height() / 2, pixmap );

                if (cornArrow) //Draw over the icon
                    drawPrimitive( PE_ArrowDown, p, visualRect( QRect(x + w - 6, x + h - 6, 7, 7), r ),
                                cg, flags, opt );


                int  pw = pixmap.width();
                x += pw + 4;
                w -= pw + 4;
            }

            // Make the label indicate if the button is a default button or not
            drawItem( p, QRect(x, y, w, h), AlignCenter|ShowPrefix, button->colorGroup(),
                        button->isEnabled(), button->pixmap(), button->text(), -1,
                        &button->colorGroup().buttonText() );


            if ( flags & Style_HasFocus )
                drawPrimitive( PE_FocusRect, p,
                                visualRect( subRect( SR_PushButtonFocusRect, widget ), widget ),
                                cg, flags );
            break;
        }

    // MENUBAR ITEM (sunken panel on mouse over)
    // -----------------------------------------
        case CE_MenuBarItem: {
            QMenuItem *mi = opt.menuItem();
            bool active  = flags & Style_Active;
            bool focused = flags & Style_HasFocus;
            bool down = flags & Style_Down;
            const int text_flags =
                AlignVCenter | AlignHCenter | ShowPrefix | DontClip | SingleLine;

            p->fillRect(r, cg.background());

            if ( active && focused && (_mouseOver||down) ) {
                uint surfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Draw_Soft|
                        Round_UpperLeft|Round_UpperRight;
                if(_symmetricButtonSurface)
                    surfaceFlags|=Draw_Symmetric;
                surfaceFlags|=Is_Horizontal;
                if(down)
                    surfaceFlags|=Is_Sunken;
                else
                    surfaceFlags|=Round_BottomLeft|Round_BottomRight;
                renderSurface(p, QRect(r_x+1, r_y+1, r_w-2, r_h-2),
                        cg.button(),
                        surfaceFlags);

                uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
                renderContour(p, r, cg.background(),
                              getColor(cg,ButtonContour),
                              contourFlags, Edge_Round2, Edge_Round2, down?Edge_Plain:Edge_Round2, down?Edge_Plain:Edge_Round2);
            }
        
            p->setPen(cg.foreground() );
            p->drawText(r, text_flags, mi->text());
            break;
        }

    // POPUPMENU ITEM (highlighted on mouseover)
    // ------------------------------------------
        case CE_PopupMenuItem: {
            const QPopupMenu *popupmenu = static_cast< const QPopupMenu * >( widget );
            QMenuItem *mi = opt.menuItem();

            if ( !mi )
            {
                // Don't leave blank holes if we set NoBackground for the QPopupMenu.
                // This only happens when the popupMenu spans more than one column.
                if (! ( widget->erasePixmap() && !widget->erasePixmap()->isNull() ) )
                    p->fillRect( r, cg.background().light( 105 ) );

                break;
            }
            int  tab        = opt.tabWidth();
            int  checkcol   = opt.maxIconWidth();
            bool enabled    = mi->isEnabled();
            bool checkable  = popupmenu->isCheckable();
            bool active     = flags & Style_Active;
            bool etchtext   = styleHint( SH_EtchDisabledText );
            bool reverse    = QApplication::reverseLayout();
            if ( checkable )
                checkcol = QMAX( checkcol, 20 );

            // Draw the menu item background
            if (active) {
                if (enabled) {
                    p->fillRect( r_x, r_y, r_w, r_h, cg.highlight() );
                }
                else {
                    if ( widget->erasePixmap() && !widget->erasePixmap()->isNull() )
                        p->drawPixmap( r.topLeft(), *widget->erasePixmap(), r );
                    else p->fillRect( r, cg.background().light(105) );
                    p->drawWinFocusRect( r );
                }
            }
            // Draw the transparency pixmap
            else if ( widget->erasePixmap() && !widget->erasePixmap()->isNull() )
                p->drawPixmap( r.topLeft(), *widget->erasePixmap(), r );
            // Draw a solid background
            else
                p->fillRect( r, cg.background().light( 105 ) );
            // Are we a menu item separator?
            if ( mi->isSeparator() )
            {
                p->setPen( cg.mid() );
                p->drawLine( r_x+5, r_y + 1, r_x2-5, r_y + 1 );
                p->setPen( cg.light() );
                p->drawLine( r_x+5, r_y + 2, r_x2-5 , r_y + 2 );
                break;
            }

            QRect cr = visualRect( QRect( r_x + 2, r_y + 2, checkcol - 1, r_h - 4 ), r );
            // Do we have an icon?
            if ( mi->iconSet() )
            {
                QIconSet::Mode mode;

                // Select the correct icon from the iconset
                if (active)
                    mode = enabled?QIconSet::Active:QIconSet::Disabled;
                else
                    mode = enabled?QIconSet::Normal:QIconSet::Disabled;

                // Do we have an icon and are checked at the same time?
                // Then draw a "pressed" background behind the icon
                if ( checkable && /*!active &&*/ mi->isChecked() )
                    qDrawShadePanel( p, cr.x(), cr.y(), cr.width(), cr.height(),
                                        cg, true, 1, &cg.brush(QColorGroup::Midlight) );
                // Draw the icon
                QPixmap pixmap = mi->iconSet()->pixmap(QIconSet::Small, mode);
                QRect pmr( 0, 0, pixmap.width(), pixmap.height() );
                pmr.moveCenter( cr.center() );
                p->drawPixmap( pmr.topLeft(), pixmap );
            }

            // Are we checked? (This time without an icon)
            else if ( checkable && mi->isChecked() )
            {
                // We only have to draw the background if the menu item is inactive -
                // if it's active the "pressed" background is already drawn
            // if ( ! active )
                    qDrawShadePanel( p, cr.x(), cr.y(), cr.width(), cr.height(), cg, true, 1,
                                        &cg.brush(QColorGroup::Midlight) );

                // Draw the checkmark
                SFlags cflags = Style_Default;
                cflags |= active ? Style_Enabled : Style_On;

                drawPrimitive( PE_CheckMark, p, cr, cg, cflags );
            }

            // Time to draw the menu item label...
//             int xm = 2 + checkcol + 2; // X position margin
            int xm = itemFrame + checkcol + itemHMargin; // X position margin

            int xp = reverse ? // X position
//                     r_x + tab + 1 + 2 + 1 - 1 :
                    r_x + tab + rightBorder + itemHMargin + itemFrame - 1 :
                    r_x + xm;

            int offset = reverse ? -1 : 1; // Shadow offset for etched text

            // Label width (minus the width of the accelerator portion)
            int tw = r_w - xm - tab - arrowHMargin - itemHMargin * 3 - itemFrame + 1;
//             int tw = r_w - xm - tab - 5 - 2 * 3 - 1 + 1;

            // Set the color for enabled and disabled text
            // (used for both active and inactive menu items)
            p->setPen( enabled ? cg.buttonText() : cg.mid() );

            // This color will be used instead of the above if the menu item
            // is active and disabled at the same time. (etched text)
            QColor discol = cg.mid();

            // Does the menu item draw it's own label?
            if ( mi->custom() ) {
                int m = 2;
                // Save the painter state in case the custom
                // paint method changes it in some way
                p->save();

                // Draw etched text if we're inactive and the menu item is disabled
                if ( etchtext && !enabled && !active ) {
                    p->setPen( cg.foreground() );
                    mi->custom()->paint( p, cg, active, enabled, xp+offset, r_y+m+1, tw, r_h-2*m );
                    p->setPen( discol );
                }
                mi->custom()->paint( p, cg, active, enabled, xp, r_y+m, tw, r_h-2*m );
                p->restore();
            }
            else {
                // The menu item doesn't draw it's own label
                QString s = mi->text();
                // Does the menu item have a text label?
                if ( !s.isNull() ) {
                    int t = s.find( '\t' );
                    int m = 2;
                    int text_flags = AlignVCenter | ShowPrefix | DontClip | SingleLine;
                    text_flags |= reverse ? AlignRight : AlignLeft;

                    //QColor draw = cg.text();
                    QColor draw = (active && enabled) ? cg.highlightedText () : cg.foreground();
                    p->setPen(draw);


                    // Does the menu item have a tabstop? (for the accelerator text)
                    if ( t >= 0 ) {
                        int tabx = reverse ? r_x + rightBorder + itemHMargin + itemFrame :
                            r_x + r_w - tab - rightBorder - itemHMargin - itemFrame;
//                         int tabx = reverse ? r_x + 1 + 2 + 1 :
//                             r_x + r_w - tab - 1 - 2 - 1;

                        // Draw the right part of the label (accelerator text)
                        if ( etchtext && !enabled ) {
                            // Draw etched text if we're inactive and the menu item is disabled
                            p->setPen( cg.light() );
                            p->drawText( tabx+offset, r_y+m+1, tab, r_h-2*m, text_flags, s.mid( t+1 ) );
                            p->setPen( discol );
                        }
                        p->drawText( tabx, r_y+m, tab, r_h-2*m, text_flags, s.mid( t+1 ) );
                        s = s.left( t );
                    }

                    // Draw the left part of the label (or the whole label
                    // if there's no accelerator)
                    if ( etchtext && !enabled ) {
                        // Etched text again for inactive disabled menu items...
                        p->setPen( cg.light() );
                        p->drawText( xp+offset, r_y+m+1, tw, r_h-2*m, text_flags, s, t );
                        p->setPen( discol );
                    }


                    p->drawText( xp, r_y+m, tw, r_h-2*m, text_flags, s, t );

                    p->setPen(cg.text());

                }

                // The menu item doesn't have a text label
                // Check if it has a pixmap instead
                else if ( mi->pixmap() ) {
                    QPixmap *pixmap = mi->pixmap();

                    // Draw the pixmap
                    if ( pixmap->depth() == 1 )
                        p->setBackgroundMode( OpaqueMode );

                    int diffw = ( ( r_w - pixmap->width() ) / 2 )
                                    + ( ( r_w - pixmap->width() ) % 2 );
                    p->drawPixmap( r_x+diffw, r_y+1, *pixmap );

                    if ( pixmap->depth() == 1 )
                        p->setBackgroundMode( TransparentMode );
                }
            }

            // Does the menu item have a submenu?
            if ( mi->popup() ) {
                PrimitiveElement arrow = reverse ? PE_ArrowLeft : PE_ArrowRight;
                int dim = pixelMetric(PM_MenuButtonIndicator) - 1;
                QRect vr = visualRect( QRect( r_x + r_w - 5 - 1 - dim,
                            r_y + r_h / 2 - dim / 2, dim, dim), r );

                // Draw an arrow at the far end of the menu item
                if ( active ) {
                    if ( enabled )
                        discol = cg.buttonText();

                    QColorGroup g2( discol, cg.highlight(), white, white,
                                    enabled ? white : discol, discol, white );

                    drawPrimitive( arrow, p, vr, g2, Style_Enabled );
                } else
                    drawPrimitive( arrow, p, vr, cg,
                            enabled ? Style_Enabled : Style_Default );
            }
            break;
        }

    // Menu and dockwindow empty space
    //
        case CE_DockWindowEmptyArea:
            p->fillRect(r, cg.background());
            break;

        case CE_MenuBarEmptyArea:
            p->fillRect(r, cg.background());

            if ( _drawToolBarSeparator ) {
                p->setPen( getColor(cg, PanelDark) );
                p->drawLine( r_x, r_y2, r_x2, r_y2 );
            }

            break;

        default:
          KStyle::drawControl(element, p, widget, r, cg, flags, opt);
    }
}

void LinspireClearStyle::drawControlMask(ControlElement element,
                                  QPainter *p,
                                  const QWidget *w,
                                  const QRect &r,
                                  const QStyleOption &opt) const
{
    switch (element) {
        case CE_PushButton: {
                    p->fillRect (r, color0);
                    renderMask(p, r, color1,
                            Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight);
                    break;
        }

        default: {
            KStyle::drawControlMask (element, p, w, r, opt);
        }
    }
}

void LinspireClearStyle::drawComplexControlMask(ComplexControl c,
                                         QPainter *p,
                                         const QWidget *w,
                                         const QRect &r,
                                         const QStyleOption &o) const
{
    switch (c) {
        case CC_SpinWidget:
        case CC_ListView:
        case CC_ComboBox: {
                p->fillRect (r, color0);
                renderMask(p, r, color1,
                        Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight);
            break;
        }
        default: {
            KStyle::drawComplexControlMask (c, p, w, r, o);
        }
    }
}

void LinspireClearStyle::drawComplexControl(ComplexControl control,
                                     QPainter *p,
                                     const QWidget *widget,
                                     const QRect &r,
                                     const QColorGroup &cg,
                                     SFlags flags,
                                     SCFlags controls,
                                     SCFlags active,
                                     const QStyleOption& opt) const
{
    const bool reverseLayout = QApplication::reverseLayout();
    const bool sunken = active&Style_Sunken;
    const bool enabled = flags & Style_Enabled;

    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    switch(control) {
    // COMBOBOX
    // --------
        case CC_ComboBox: {
            static const unsigned int handleWidth = 17;

            QComboBox *cb = (QComboBox *)(widget); // **hack**: cast away const-ness for animation support!
            // at the moment cb is only needed to check if the combo box is editable or not.
            // if cb doesn't exist, just assume false and the app (gideon! ;) ) at least doesn't crash.
            bool editable = false;
            bool hasFocus = false;
            if (cb) {
                editable = cb->editable();
                hasFocus = cb->hasFocus();
            }

            int highlightStep = 0;
            if( _mouseOverAnimation && animWidgets.contains(cb) ) {
                highlightStep = (int)animWidgets[cb].step;
            } else if ( _mouseOver && (cb == hoverWidget) ) {
                highlightStep = (int)ANIMATIONSTEPS;
            }
            const QColor buttonColor = cg.button().light(100+highlightStep) /*enabled?cg.button():cg.background()*/;
            const QColor contourColor = getColor(cg, ButtonContour, enabled).light(100+highlightStep);
            const QColor inputColor = enabled?(editable?cg.base():cg.button())
                                              :cg.background();

            const QRect RbuttonSurface(reverseLayout?r_x+1:r_x2-handleWidth, r_y+1,
                                        handleWidth, r_h-2);
            const QRect RcontentSurface(reverseLayout?r_x+1+handleWidth+1:r_x+1, r_y+1,
                                         r_w-handleWidth-3, r_h-2);

            // handler

            uint surfaceFlags = Draw_Top|Draw_Bottom|Is_Horizontal|Draw_Soft;
            if(_symmetricButtonSurface)
                surfaceFlags|=Draw_Symmetric;
            if(!enabled) surfaceFlags|=Is_Disabled;
            if(reverseLayout) {
                surfaceFlags |= Round_UpperLeft|Round_BottomLeft;
                if(editable)
                    surfaceFlags |= Draw_Left|Draw_Right;
                else
                    surfaceFlags |= Draw_Left;
                
            } else {
                surfaceFlags |= Round_UpperRight|Round_BottomRight;
                if(editable)
                    surfaceFlags |= Draw_Left|Draw_Right;
                else
                    surfaceFlags |= Draw_Right;
            }

            if (widget == hoverWidget) {
                surfaceFlags |= Is_Highlight;
//                 if(editable) surfaceFlags |= Highlight_Left|Highlight_Right;
//                 surfaceFlags |= Highlight_Top|Highlight_Bottom;
            }
            if( sunken ) surfaceFlags |= Is_Sunken;
            renderSurface(p, RbuttonSurface, buttonColor, surfaceFlags);

            if(!editable) {
                surfaceFlags = Draw_Top|Draw_Bottom|Is_Horizontal|Draw_Soft;
                if(_symmetricButtonSurface)
                    surfaceFlags|=Draw_Symmetric;
                if(!enabled) surfaceFlags|=Is_Disabled;
                if(reverseLayout) {
                    surfaceFlags |= Round_UpperRight|Round_BottomRight|Draw_Right;
                } else {
                    surfaceFlags |= Round_UpperLeft|Round_BottomLeft|Draw_Left;
                }

                if (widget == hoverWidget) {
                    surfaceFlags |= Is_Highlight;
//                     surfaceFlags |= Highlight_Top|Highlight_Bottom;
                }
                if( sunken ) surfaceFlags |= Is_Sunken;
                renderSurface(p, RcontentSurface, buttonColor, surfaceFlags);
                    if (hasFocus) {
                        drawPrimitive(PE_FocusRect, p,
                            QRect(RcontentSurface.x() + 2,
                                RcontentSurface.y() + 2,
                                RcontentSurface.width() - 4,
                                RcontentSurface.height() - 4), cg);
                    }
            } else {
                // thin frame around the input area
                if (_inputFocusHighlight && hasFocus && editable && enabled)
                    p->setPen(inputColor.dark(135) );
                else
                    p->setPen(inputColor);
                p->drawLine(RcontentSurface.x(), reverseLayout?RcontentSurface.y():RcontentSurface.y()+1,
                        RcontentSurface.x(), reverseLayout?RcontentSurface.bottom():RcontentSurface.bottom()-1);
                p->drawLine(RcontentSurface.x()+1, RcontentSurface.y(),
                        reverseLayout?RcontentSurface.right()-1:RcontentSurface.right(), RcontentSurface.y() );
                if (_inputFocusHighlight && hasFocus && editable && enabled)
                    p->setPen(inputColor.dark(135) );
                else
                    p->setPen(inputColor);
                p->drawLine(reverseLayout?RcontentSurface.x():RcontentSurface.x()+1, RcontentSurface.bottom(),
                        reverseLayout?RcontentSurface.right()-1:RcontentSurface.right(), RcontentSurface.bottom() );
                p->drawLine(RcontentSurface.right(), RcontentSurface.top()+1,
                        RcontentSurface.right(), RcontentSurface.bottom()-1 );

                // input area
                p->fillRect(RcontentSurface.x()+1, RcontentSurface.y()+1,
                        RcontentSurface.width()-2, RcontentSurface.height()-2, inputColor );
            }

            // contour...
//             const WidgetState s = enabled?/*(sunken?IsPressed:*/IsEnabled/*)*/:IsDisabled;
            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            EdgeType e1 = Edge_Round2, e2 = Edge_Round2, e3 = Edge_Round2, e4 = Edge_Round2;
            if(editable) {
                if(reverseLayout) {
                    e2 = Edge_Round3;
                    e4 = Edge_Round3;
                } else {
                    e1 = Edge_Round3;
                    e3 = Edge_Round3;
                }
            }
            if( khtmlWidgets.contains(cb) )
                contourFlags |= Draw_AlphaBlend;
            renderContour(p, r, cg.background(), contourColor, contourFlags, e1, e2, e3, e4);
            //extend the contour: between input and handler...
            p->setPen(alphaBlendColors(cg.background(), contourColor, 50) );
            if(reverseLayout) {
                p->drawLine(r_x+1+handleWidth, r_y+1, r_x+1+handleWidth, r_y2-1);
            } else {
                p->drawLine(r_x2-handleWidth-1, r_y+1, r_x2-handleWidth-1, r_y2-1);
            }


            p->setPen(cg.foreground());
            drawPrimitive(PE_SpinWidgetDown, p, RbuttonSurface, cg, Style_Default|Style_Enabled|Style_Raised);

            // QComboBox draws the text using cg.text(), we can override this
            // from here
            p->setPen( cg.buttonText() );
            p->setBackgroundColor( cg.button() );
            break;
        }

    // SPINWIDGETS
    // -----------
        case CC_SpinWidget: {
            static const unsigned int handleWidth = 17;

            QSpinWidget *sw = (QSpinWidget*)(widget); // **hack**: cast away const-ness for animation support!
            SFlags sflags = flags;
            PrimitiveElement pe;

            bool hasFocus = false;
            if (sw)
                hasFocus = sw->hasFocus();

            int highlightStep = 0;
            if( _mouseOverAnimation && animWidgets.contains(sw) ) {
                highlightStep = (int)animWidgets[sw].step;
            } else if ( _mouseOver && (sw == hoverWidget) ) {
                highlightStep = (int)ANIMATIONSTEPS;
            }
            const QColor buttonColor = cg.button().light(100+highlightStep);
            const QColor inputColor = enabled?cg.base():cg.background();

            const bool heightDividable = ((r_h%2) == 0),
                       upEnabled = sw->isUpEnabled(),
                       downEnabled = sw->isDownEnabled();;
    
            if ( controls & SC_SpinWidgetUp || controls & SC_SpinWidgetDown ) {
                // surface
                QRect upRect = QRect(reverseLayout?r_x+1:r_x2-handleWidth, r_y+1,
                        handleWidth, (r_h-2)/2);
                QRect downRect = QRect(reverseLayout?r_x+1:r_x2-handleWidth,
                        heightDividable?r_y+1+((r_h-2)/2):r_y+1+((r_h-2)/2)+1,
                        handleWidth, ((r_h-2)/2) );
                if(heightDividable) {
                    upRect = QRect(upRect.left(), upRect.top(), upRect.width(), upRect.height()-1 );
                    downRect = QRect(downRect.left(), downRect.top()+1, downRect.width(), downRect.height()-1 );
                }
    
                uint surfaceFlags;
                if ( controls & SC_SpinWidgetUp ) {
                    surfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Is_Horizontal|Draw_Soft;
                    if(_symmetricButtonSurface)
                        surfaceFlags|=Draw_Symmetric;
                    if(reverseLayout) {
                        surfaceFlags |= Round_UpperLeft;
                    } else {
                        surfaceFlags |= Round_UpperRight;
                    }
                    if ((widget == hoverWidget) || (sflags & Style_MouseOver)) {
                        surfaceFlags |= Is_Highlight;
                    }
                    if (active==SC_SpinWidgetUp) surfaceFlags|=Is_Sunken;
                    if(!enabled) surfaceFlags|=Is_Disabled;
                    if( sw && !upEnabled ) surfaceFlags|=Is_Disabled;
                    renderSurface(p, upRect, upEnabled?buttonColor:cg.button(), surfaceFlags);
                }
                if ( controls & SC_SpinWidgetDown ) {
                    surfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Is_Horizontal|Draw_Soft;
                    if(_symmetricButtonSurface)
                        surfaceFlags|=Draw_Symmetric;
                    if(reverseLayout) {
                        surfaceFlags |= Round_BottomLeft;
                    } else {
                        surfaceFlags |= Round_BottomRight;
                    }
                    if ((widget == hoverWidget) || (sflags & Style_MouseOver)) {
                        surfaceFlags |= Is_Highlight;
                    }
                    if (active==SC_SpinWidgetDown) surfaceFlags|=Is_Sunken;
                    if(!enabled) surfaceFlags|=Is_Disabled;
                    if( sw && !downEnabled ) surfaceFlags|=Is_Disabled;
                    renderSurface(p, downRect, downEnabled?buttonColor:cg.button(), surfaceFlags);
                }
    
                // icons...
                if ( controls & SC_SpinWidgetUp ) {
                    sflags = Style_Default | Style_Enabled;
                    if (active == SC_SpinWidgetUp) {
                        sflags |= Style_On;
                        sflags |= Style_Sunken;
                    } else
                        sflags |= Style_Raised;
                    if (sw->buttonSymbols() == QSpinWidget::PlusMinus)
                        pe = PE_SpinWidgetPlus;
                    else
                        pe = PE_SpinWidgetUp;
                    p->setPen( upEnabled?cg.foreground():(sw->palette().disabled().foreground()) );
                    drawPrimitive(pe, p, upRect, cg, sflags);
                }
    
                if ( controls & SC_SpinWidgetDown ) {
                    sflags = Style_Default | Style_Enabled;
                    if (active == SC_SpinWidgetDown) {
                        sflags |= Style_On;
                        sflags |= Style_Sunken;
                    } else
                        sflags |= Style_Raised;
                    if (sw->buttonSymbols() == QSpinWidget::PlusMinus)
                        pe = PE_SpinWidgetMinus;
                    else
                        pe = PE_SpinWidgetDown;
                    p->setPen( downEnabled?cg.foreground():(sw->palette().disabled().foreground()) );
                    drawPrimitive(pe, p, downRect, cg, sflags);
                }
            }

            if ( controls & SC_SpinWidgetFrame ) {
                // thin frame around the input area
                const QRect Rcontent = QRect(reverseLayout?r_x+1+handleWidth+1:r_x+1, r_y+1,
                        r_w-1-2-handleWidth, r_h-2);
                if (_inputFocusHighlight && hasFocus && enabled)
                    p->setPen(inputColor.dark(135) );
                else
                    p->setPen(inputColor);
                p->drawLine(Rcontent.left(), reverseLayout?Rcontent.top():Rcontent.top()+1,
                        Rcontent.left(), reverseLayout?Rcontent.bottom():Rcontent.bottom()-1 );
                p->drawLine(Rcontent.left()+1, Rcontent.top(),
                        reverseLayout?Rcontent.right()-1:Rcontent.right(), Rcontent.top() );
                if (_inputFocusHighlight && hasFocus && enabled)
                    p->setPen(inputColor.dark(135) );
                else
                    p->setPen(inputColor);
                p->drawLine(Rcontent.left()+1, Rcontent.bottom(), Rcontent.right()-1, Rcontent.bottom() );
                p->drawLine(Rcontent.right(), Rcontent.top()+1,
                        Rcontent.right(), reverseLayout?Rcontent.bottom()-1:Rcontent.bottom() );
    
                // contour
                EdgeType e1,e2,e3,e4;
                if(reverseLayout) {
                    e1 = Edge_Round2;
                    e3 = Edge_Round2;
                    e2 = Edge_Round3;
                    e4 = Edge_Round3;
                } else {
                    e1 = Edge_Round3;
                    e3 = Edge_Round3;
                    e2 = Edge_Round2;
                    e4 = Edge_Round2;
                }
                renderContour(p, querySubControlMetrics(control, widget, SC_SpinWidgetFrame),
                            cg.background(), getColor(cg, ButtonContour, enabled), Draw_Left|Draw_Right|Draw_Top|Draw_Bottom, e1,e2,e3,e4 );
                p->setPen(alphaBlendColors(cg.background(), getColor(cg, ButtonContour, enabled), 50) );
                p->drawLine(reverseLayout?r_x+1+handleWidth:r_x2-handleWidth-1, r_y+1,
                        reverseLayout?r_x+1+handleWidth:r_x2-handleWidth-1, r_y2-1);
                p->drawLine(reverseLayout?r_x+1:r_x2-handleWidth, r_y+1+(r_h-2)/2,
                        reverseLayout?r_x+handleWidth:r_x2-1, r_y+1+(r_h-2)/2);
                if(heightDividable)
                    p->drawLine(reverseLayout?r_x+1:r_x2-handleWidth, r_y+1+(r_h-2)/2-1,
                            reverseLayout?r_x+handleWidth:r_x2-1, r_y+1+(r_h-2)/2-1);
            }

            break;
        }

    // TOOLBUTTON
    // ----------
        case CC_ToolButton: {
            QToolButton *tb = (QToolButton *) widget; // **hack**: cast away const-ness for animation support!

            int highlightStep = 0;
            if( _mouseOverAnimation && animWidgets.contains(tb) ) {
                highlightStep = (int)animWidgets[tb].step;
            }

            QRect button, menuarea;
            button   = querySubControlMetrics(control, widget, SC_ToolButton, opt);
            menuarea = querySubControlMetrics(control, widget, SC_ToolButtonMenu, opt);

            SFlags bflags = flags,
                    mflags = flags;

            if (kornMode) {
                drawPrimitive(PE_ButtonTool, p, button, cg, bflags, QStyleOption(tb)/*opt*/);
                break;
            } else {
//                 // don't want to have the buttoncolor as the background...
//                 p->fillRect(r, cg.background());
                bflags &= ~Style_MouseOver;
            }

            if (active & SC_ToolButton)
                bflags |= Style_Down;

            if (active & SC_ToolButtonMenu)
                mflags |= Style_Down;

            if (controls & SC_ToolButton) {
                // draw the toolbutton specially (blend in and out smoothly) if it's a child of a QToolBar.
                if(tb->parent() && ::qt_cast<QToolBar*>(tb->parent()) ) {
                    if( highlightStep>0 || _mouseOver?(bflags & (Style_Down | Style_On) || (_toolbarMouseOver && (bflags & Style_Raised)))
                                                    :(bflags & (Style_Down | Style_On)) ) {
                        bool sunken2 = bflags&Style_Down || bflags&Style_On;
                        int alpha = (sunken2||((bflags&Style_Raised)&&!_mouseOverAnimation))?0:255-(255/(int)ANIMATIONSTEPS*highlightStep);

                        const WidgetState s = enabled?(sunken2?IsPressed:IsEnabled):IsDisabled;
                        uint surfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|
                                            Round_UpperLeft|Round_UpperRight|Round_BottomLeft|
                                            Is_Horizontal|Round_BottomRight|Draw_Soft;
                        if(_symmetricButtonSurface)
                            surfaceFlags|=Draw_Symmetric;
                        if(!enabled) surfaceFlags|=Is_Disabled;
                        else {
                            if(sunken2) surfaceFlags|=Is_Sunken;
                        }
                        renderSurface(p, QRect(r_x+1, r_y+1, r_w-2, r_h-2),
                                    alphaBlendColors(cg.background(), cg.button().light(100+highlightStep),
                                            alpha),
                                    surfaceFlags);

                        uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom/*|
                                Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight*/;
                        if(!enabled) contourFlags|=Is_Disabled;
                        renderContour(p, r, cg.background(),
                                    alphaBlendColors(cg.background(), getColor(cg,ButtonContour,s).light(100+highlightStep),
                                            alpha),
                                    contourFlags);
                    } else {
                        p->fillRect(r,cg.background() );
                    }

                } else if (bflags & (Style_Down | Style_On | Style_Raised)) { // If we're pressed, on, or raised...
                    drawPrimitive( PE_ButtonTool, p, button, cg, bflags, QStyleOption(tb) );
                } else if (tb->parentWidget() &&
                            tb->parentWidget()->backgroundPixmap() &&
                            !tb->parentWidget()->backgroundPixmap()->isNull()) {
                    QPixmap pixmap = *(tb->parentWidget()->backgroundPixmap());
                    p->drawTiledPixmap( r, pixmap, tb->pos() );
                }
            }

            // Draw a toolbutton menu indicator if required
            if (controls & SC_ToolButtonMenu) {
                if (mflags & (Style_Down | Style_On | Style_Raised)) {
                    drawPrimitive(PE_ButtonDropDown, p, menuarea, cg, mflags, opt);
                }
                drawPrimitive(PE_ArrowDown, p, menuarea, cg, mflags, opt);
            }

            if (tb->hasFocus() && !tb->focusProxy()) {
                QRect fr = tb->rect();
                fr.addCoords(2, 2, -2, -2);
                drawPrimitive(PE_FocusRect, p, fr, cg);
            }

            // Set the color for the ToolButton menu indicator
            p->setPen(cg.buttonText() );

            break;
        }

        default:
            KStyle::drawComplexControl(control, p, widget,
                                        r, cg, flags, controls,
                                        active, opt);
            break;
    }
}


QRect LinspireClearStyle::subRect(SubRect r, const QWidget *widget) const
{
    switch (r) {
        case SR_ComboBoxFocusRect: {
            return querySubControlMetrics( CC_ComboBox, widget, SC_ComboBoxEditField );
        }

        // Don't use KStyles progressbar subrect
        // TODO:
        case SR_ProgressBarGroove: {
            return QRect(widget->rect());
        }
        case SR_ProgressBarContents:
        case SR_ProgressBarLabel: {
            QRect rw = widget->rect();
            return QRect(rw.x()+2, rw.y()+2, rw.width()-4, rw.height()-4 );
        }

        default: {
            return KStyle::subRect(r, widget);
        }
    }
}

QRect LinspireClearStyle::querySubControlMetrics(ComplexControl control,
                                          const QWidget *widget,
                                          SubControl subcontrol,
                                          const QStyleOption &opt) const
{
    if (!widget) {
        return QRect();
    }

    QRect r(widget->rect());

    int r_w = r.width();
    int r_h = r.height();
    int r_x, r_y, r_x2, r_y2;
    r.coords(&r_x, &r_y, &r_x2, &r_y2);

    switch (control) {
        case CC_ComboBox: {
            switch (subcontrol) {
                case SC_ComboBoxEditField: {
                    // TODO: see if the handler width is in pixelmetric too.
                    return QRect(r_x+2, r_y+2, r_w-4-17-1, r_h-4);
                }
                default: {
                    return KStyle::querySubControlMetrics(control, widget, subcontrol, opt);
                }
            }
            break;
        }
        case CC_SpinWidget: {
            const int fw = 2; // Frame width...

            const bool heightDividable = ((r_h%2) == 0);

            QSize bs;
            if(heightDividable) {
                bs.setHeight(QMAX(8, (r_h-2)/2));
            } else {
                bs.setHeight(QMAX(8, (r_h-2-1)/2));
            }
            bs.setWidth(17);

            const int buttonsLeft = /*reverseLayout?r_x+1:*/r_x2-bs.width();

            switch (subcontrol) {
                case SC_SpinWidgetUp: {
                    return QRect(buttonsLeft, r_y+1, bs.width(), bs.height() );
                }
                case SC_SpinWidgetDown: {
                    if(heightDividable) {
                        return QRect(buttonsLeft, r_y+1+bs.height(),
                                bs.width(), r_h-(bs.height()+2) );
                    } else {
                        return QRect(buttonsLeft, r_y+1+bs.height()+1,
                                bs.width(), r_h-(bs.height()+2+1) );
                    }
                }
                case SC_SpinWidgetFrame: {
                    return QRect(r_x, r_y, r_w, r_h );
                }
                case SC_SpinWidgetEditField: {
                    return QRect(r_x+fw, r_y+fw,
                            r_w-(bs.width()+1+2*fw), r_h-2*fw);
                }
                case SC_SpinWidgetButtonField: {
                    return QRect(buttonsLeft, r_y+1, bs.width(), r_h-2);
                }
                default: {
                    return KStyle::querySubControlMetrics(control, widget, subcontrol, opt);
                }
            }
            break;
        }
        default: {
            return KStyle::querySubControlMetrics(control, widget, subcontrol, opt);
        }
    }
}

QStyle::SubControl LinspireClearStyle::querySubControl( ComplexControl control,
                                                  const QWidget* widget,
                                                  const QPoint& pos,
                                                  const QStyleOption& opt ) const
{
    switch (control) {
        case CC_ScrollBar: {
            // reimplement to fix KStyle's hackish CC_ScrollBar implementation which enables the
            // third button but somethimes returns a wrong result...
            QStyle::SubControl ret = QCommonStyle::querySubControl(control, widget, pos, opt);

            // also check if the point actually is in the widget geometry!
            if (ret == SC_None && widget->rect().contains(pos) )
                ret = SC_ScrollBarSubLine;

            return ret;
        }
        case CC_Slider: {
            QRect r = querySubControlMetrics(control, widget, SC_SliderHandle, opt);
            if (r.isValid() && r.contains(pos))
                return SC_SliderHandle;
            else
                return SC_SliderGroove;
        }
        default:
            return KStyle::querySubControl( control, widget, pos, opt );
    }
}

int LinspireClearStyle::pixelMetric(PixelMetric m, const QWidget *widget) const
{
    switch(m) {
    // TABS
    // ----
        case PM_TabBarTabVSpace: {
            const QTabBar * tb = (const QTabBar *) widget;
            if (tb->shape() == QTabBar::RoundedAbove ||
                tb->shape() == QTabBar::RoundedBelow)
                return 10;
            else
                return 4;
        }

        case PM_TabBarTabOverlap: {
            return 1;
        }

#if (QT_VERSION >= 0x030300) // requires Qt 3.3
    // extra space between menubar items
        case PM_MenuBarItemSpacing: {
            return 6;
        }
#endif

// TODO: another thing worth considering. this extra border gives a cleaner look to popup menus...
//         case PM_PopupMenuFrameHorizontalExtra:
//             return 1;
//         case PM_PopupMenuFrameVerticalExtra:
//             return 1;

// #if (QT_VERSION >= 0x030300) // requires Qt 3.3
//     // extra space between toolbar items
//         case PM_ToolBarItemSpacing: {
//             return 4;
//         }
// #endif

    // SCROLL BAR
        case PM_ScrollBarSliderMin: {
            return 21;
        }
        case PM_ScrollBarExtent: {
            return 16;
        }

        case PM_DockWindowSeparatorExtent:
            return 6;

    // SPLITTERS
    // ---------
        case PM_SplitterWidth: {
            return 6;
        }

    // PROGRESSBARS
    // ------------
        case PM_ProgressBarChunkWidth:
            return 10;

    // SLIDER
    // ------
        case PM_SliderLength:
            return 11;

    // MENU INDICATOR
    // --------------
        case PM_MenuButtonIndicator:
            return 8;

    // CHECKBOXES / RADIO BUTTONS
    // --------------------------
        case PM_ExclusiveIndicatorWidth:    // Radiobutton size
            return radiobutton_mask_width;
        case PM_ExclusiveIndicatorHeight:
            return radiobutton_mask_height;
        case PM_IndicatorWidth:             // Checkbox size
            return checkbox_mask_width;
        case PM_IndicatorHeight:
            return checkbox_mask_height;

    // FRAMES
    // ------
        case PM_SpinBoxFrameWidth:
            return 1;

        case PM_MenuBarFrameWidth:
            return 0;

        case PM_DefaultFrameWidth: {
            if(widget && ::qt_cast<QPopupMenu*>(widget))
                return 1;
            else
                return 2;
        }

        case PM_ButtonDefaultIndicator:
            return 0;

        case PM_ButtonMargin:
            return 2;

        case PM_ButtonShiftVertical:
            return 1;

        case PM_ButtonShiftHorizontal:
            return 0;

        default:
            return KStyle::pixelMetric(m, widget);
    }
}


QSize LinspireClearStyle::sizeFromContents(ContentsType t,
                                    const QWidget *widget,
                                    const QSize &s,
                                    const QStyleOption &opt) const
{
    switch (t) {
        case CT_PopupMenuItem: {
            if (!widget || opt.isDefault())
                return s;

            const QPopupMenu *popup = dynamic_cast<const QPopupMenu *>(widget);
            QMenuItem *mi = opt.menuItem();
            int maxpmw = opt.maxIconWidth();
            int w = s.width(), h = s.height();
            bool checkable = popup->isCheckable();

            if (mi->custom()) {
                w = mi->custom()->sizeHint().width();
                h = mi->custom()->sizeHint().height();
                if (!mi->custom()->fullSpan() )
                    h += 4;
            } else if (mi->widget()) {
            // don't change the size in this case.
            } else if (mi->isSeparator()) {
                w = 20;
                h = 2;
            } else {
                if (mi->pixmap()) {
                    h = QMAX(h, mi->pixmap()->height() + 2);
                } else {
                    h = QMAX(h, 16 + 2 );
                    h = QMAX(h, popup->fontMetrics().height() + 4 );
                }

                if (mi->iconSet()) {
                    h = QMAX(h, mi->iconSet()->pixmap(QIconSet::Small, QIconSet::Normal).height() + 2);
                }
            }

            if (!mi->text().isNull() && (mi->text().find('\t') >= 0)) {
                w += itemHMargin + itemFrame*2 + 7;
//                 w += 12;
            } else if (mi->popup()) {
                w += 2 * arrowHMargin;
//                 w += 24;
            }

            if (maxpmw) {
                w += maxpmw + 6;
            }
            if (checkable && maxpmw < 20) {
                w += 20 - maxpmw;
            }
            if (checkable || maxpmw > 0) {
                w += 12;
            }

            w += rightBorder;
//             w += 20;

            return QSize(w, h);
        }

        case CT_PushButton:
        {
            const QPushButton* btn = static_cast<const QPushButton*>(widget);

            int w = s.width() + 2 * pixelMetric(PM_ButtonMargin, widget);
            int h = s.height() + 2 * pixelMetric(PM_ButtonMargin, widget);
            if ( btn->text().isEmpty() && s.width() < 32 ) return QSize(w, h);
//             if ( btn->isDefault() ) {
//                 w += 2*pixelMetric(PM_ButtonDefaultIndicator);
//                 h += 2*pixelMetric(PM_ButtonDefaultIndicator);
//             }

            return QSize( w+25, h+5 );
        }

        case CT_ToolButton:
        {
            if(widget->parent() && ::qt_cast<QToolBar*>(widget->parent()) )
                return QSize( s.width()+2*4, s.height()+2*4 );
            else
                return KStyle::sizeFromContents (t, widget, s, opt);
        }

        default:
            return KStyle::sizeFromContents (t, widget, s, opt);
    }

    return KStyle::sizeFromContents (t, widget, s, opt);
}

int LinspireClearStyle::styleHint( StyleHint stylehint,
                                  const QWidget *widget,
                                  const QStyleOption &option,
                                  QStyleHintReturn* returnData ) const
{
    switch (stylehint) {
        case SH_PopupMenu_SubMenuPopupDelay:
            return 96; // Motif-like delay...

// this motif defaults seems to be very good. unfortunately it doesn't look good with the current panel/scrollbar style.
//         case SH_ScrollView_FrameOnlyAroundContents:
//             return true;

        default:
            return KStyle::styleHint(stylehint, widget, option, returnData);
    }
}

bool LinspireClearStyle::eventFilter(QObject *obj, QEvent *ev)
{
    if (KStyle::eventFilter(obj, ev) )
        return true;

    if (!obj->isWidgetType() ) return false;

    // focus highlight
    if ( ::qt_cast<QLineEdit*>(obj) ) {
        QWidget* widget = static_cast<QWidget*>(obj);

        if ( ::qt_cast<QSpinWidget*>(widget->parentWidget()) )
        {
            QWidget* spinbox = widget->parentWidget();
            if ((ev->type() == QEvent::FocusIn) || (ev->type() == QEvent::FocusOut))
            {
                spinbox->repaint(false);
            }
            return false;
        }

        if ((ev->type() == QEvent::FocusIn) || (ev->type() == QEvent::FocusOut))
        {
            widget->repaint(false);
        }
        return false;
    }

    //Hover highlight... use qt_cast to check if the widget inheits one of the classes.
    if ( _mouseOver &&
            (::qt_cast<QPushButton*>(obj) || ::qt_cast<QComboBox*>(obj) || ::qt_cast<QSpinWidget*>(obj) ||
            ::qt_cast<QToolButton*>(obj) ) ) {
        if( _mouseOverAnimation ) {
            if ( (ev->type() == QEvent::Enter) && static_cast<QWidget*>(obj)->isEnabled() ) {
                QWidget *widget = static_cast<QWidget*>(obj);
                hoverWidget = widget;
                animWidgets[widget].status = Starting;
                // if timer isn't running, start it now for glow-in animation.
                if( !animTimer->isActive() )
                    animTimer->start( ANIMATIONTIMEOUT, false );
            } else if ( (ev->type() == QEvent::Leave) && (obj == hoverWidget) ) {
                QWidget *widget = static_cast<QWidget*>(obj);
                hoverWidget = 0;
                animWidgets[widget].status = Stopping;
                // if timer isn't running, start it now for glow-out animation.
                if( !animTimer->isActive() )
                    animTimer->start( ANIMATIONTIMEOUT, false );
            }
        } else {
            if ((ev->type() == QEvent::Enter) && static_cast<QWidget*>(obj)->isEnabled())
            {
                QWidget* widget = static_cast<QWidget*>(obj);
                hoverWidget = widget;
                widget->repaint(false);
            }
            else if ((ev->type() == QEvent::Leave) && (obj == hoverWidget) )
            {
                QWidget* widget = static_cast<QWidget*>(obj);
                hoverWidget = 0;
                widget->repaint(false);
            }
        }
    }

    if ( _mouseOver && ( ::qt_cast<QCheckBox*>(obj) || ::qt_cast<QRadioButton*>(obj) || obj->inherits("QSplitterHandle") ) ) {
        if ((ev->type() == QEvent::Enter) && static_cast<QWidget*>(obj)->isEnabled())
        {
            QWidget* widget = static_cast<QWidget*>(obj);
            hoverWidget = widget;
            widget->repaint(false);
        }
        else if ((ev->type() == QEvent::Leave) && (obj == hoverWidget) )
        {
            QWidget* widget = static_cast<QWidget*>(obj);
            hoverWidget = 0;
            widget->repaint(false);
        }
    } else if ( _mouseOver && ::qt_cast<QTabBar*>(obj) ) {
        if ((ev->type() == QEvent::Enter) && static_cast<QWidget*>(obj)->isEnabled())
        {
            QWidget* tabbar = static_cast<QWidget*>(obj);
            hoverWidget = tabbar;
            hoverTab = 0;
            tabbar->repaint(false);
        }
        else if (ev->type() == QEvent::MouseMove)
        {
            QTabBar *tabbar = dynamic_cast<QTabBar*>(obj);
            QMouseEvent *me = dynamic_cast<QMouseEvent*>(ev);

            if (tabbar && me) {
                // avoid unnecessary repaints (which otherwise would occour on every
                // MouseMove event causing high cpu load).

                bool repaint = true;

                // go through the tabbar and see which tabs are hovered by the mouse.
                // tabs are overlapping 1 px, so it's possible that 2 tabs are under the mouse.
                int tabCount = 0;
                for (int i = 0; i < tabbar->count(); ++i) {
                    QTab *tab = tabbar->tab(i);
                    if (tab && tab->rect().contains(me->pos() ) ) {
                        ++ tabCount;

                        if (tabCount < 2) {
                                // good, only one tab under the mouse.
                            if (hoverTab==tab)
                                repaint = false; // has been updated before, no repaint necessary
                            hoverTab = tab;
                        } else {
                                // uhh! there's another tab under the mouse, repaint...
                            repaint = true;
                            hoverTab = 0; // make sure the tabbar is repainted next time too.
                        }

                    }
                }

                if (repaint)
                    tabbar->repaint(false);
            }
        }
        else if (ev->type() == QEvent::Leave)
        {
            QWidget* tabbar = static_cast<QWidget*>(obj);
            hoverWidget = 0;
            hoverTab = 0;
            tabbar->repaint(false);
        }
    } else if ( _mouseOver && ::qt_cast<QScrollBar*>(obj) ) {
        if ((ev->type() == QEvent::Enter) && static_cast<QWidget*>(obj)->isEnabled())
        {
            QWidget* widget = static_cast<QWidget*>(obj);
            hoverSubControl = SC_None;
            hoverWidget = widget;
//             qDebug("Enter: %s", widget->className() );
            // we don't have the mouse position and thus the sub control
            // yet, so we also don't need to repaint.
            // widget->repaint(false);
        }
        else if ((ev->type() == QEvent::MouseMove) && (obj == hoverWidget) )
        {
            QScrollBar* widget = static_cast<QScrollBar*>(obj);
            QMouseEvent * me = ( QMouseEvent * ) ev;
            SubControl oldHoverSubControl = hoverSubControl;
            hoverSubControl = querySubControl( CC_ScrollBar, widget, me->pos() );
            if( oldHoverSubControl != hoverSubControl ) {
                if( oldHoverSubControl==SC_ScrollBarAddLine ||
                    oldHoverSubControl==SC_ScrollBarSubLine ||
                    oldHoverSubControl==SC_ScrollBarSlider ) {
                    // subcontrol metric returns inaccurate information for SC_ScrollBarSubLine because
                    // with the 3rd button there are 2 areas to repaint...
                    if( oldHoverSubControl == SC_ScrollBarSubLine ) {
                        // normal SC_ScrollBarSubLine.
                        widget->update(querySubControlMetrics(CC_ScrollBar, widget, oldHoverSubControl));
                        // the third button...
                        QRect r1 = querySubControlMetrics(CC_ScrollBar, widget, SC_ScrollBarAddPage);
                        QRect r2 = querySubControlMetrics(CC_ScrollBar, widget, SC_ScrollBarAddLine);
                        if(widget->orientation() == Horizontal)
                            widget->update( r1.right(), r1.top(), r2.left(), r1.bottom() );
                        else
                            widget->update( r1.left(), r1.bottom(), r1.right(), r2.top() );
                    } else
                        widget->update(querySubControlMetrics(CC_ScrollBar, widget, oldHoverSubControl));
                }
                if( hoverSubControl==SC_ScrollBarAddLine ||
                    hoverSubControl==SC_ScrollBarSubLine ||
                    hoverSubControl==SC_ScrollBarSlider ) {
                    if( hoverSubControl == SC_ScrollBarSubLine ) {
                        // normal SC_ScrollBarSubLine.
                        widget->update(querySubControlMetrics(CC_ScrollBar, widget, hoverSubControl));
                        // the third button...
                        QRect r1 = querySubControlMetrics(CC_ScrollBar, widget, SC_ScrollBarAddPage);
                        QRect r2 = querySubControlMetrics(CC_ScrollBar, widget, SC_ScrollBarAddLine);
                        if(widget->orientation() == Horizontal)
                            widget->update( r1.right(), r1.top(), r2.left(), r1.bottom() );
                        else
                            widget->update( r1.left(), r1.bottom(), r1.right(), r2.top() );
                    } else
                        widget->update(querySubControlMetrics(CC_ScrollBar, widget, hoverSubControl));
                }
            }
        }
        else if ((ev->type() == QEvent::Leave) && (obj == hoverWidget) )
        {
            QScrollBar* widget = static_cast<QScrollBar*>(obj);
            SubControl oldHoverSubControl = hoverSubControl;
            hoverSubControl = SC_None;
            hoverWidget = 0;
            if( oldHoverSubControl==SC_ScrollBarAddLine ||
                    oldHoverSubControl==SC_ScrollBarSubLine ||
                    oldHoverSubControl==SC_ScrollBarSlider ) {
                if( oldHoverSubControl == SC_ScrollBarSubLine ) {
                    // normal SC_ScrollBarSubLine.
                    widget->update(querySubControlMetrics(CC_ScrollBar, widget, oldHoverSubControl));
                    // the third button...
                    QRect r1 = querySubControlMetrics(CC_ScrollBar, widget, SC_ScrollBarAddPage);
                    QRect r2 = querySubControlMetrics(CC_ScrollBar, widget, SC_ScrollBarAddLine);
                    if(widget->orientation() == Horizontal)
                        widget->update( r1.right(), r1.top(), r2.left(), r1.bottom() );
                    else
                        widget->update( r1.left(), r1.bottom(), r1.right(), r2.top() );
                } else
                    widget->update(querySubControlMetrics(CC_ScrollBar, widget, oldHoverSubControl));
            }
        }
    } else if ( _mouseOver && ::qt_cast<QSlider*>(obj) ) {
        if ((ev->type() == QEvent::Enter) && static_cast<QWidget*>(obj)->isEnabled())
        {
            QWidget* widget = static_cast<QWidget*>(obj);
            hoverSubControl = SC_None;
            hoverWidget = widget;
            // we don't have the mouse position and thus the sub control
            // yet, so we also don't need to repaint.
            // widget->repaint(false);
        }
        else if ((ev->type() == QEvent::MouseMove) && (obj == hoverWidget) )
        {
            QWidget* widget = static_cast<QWidget*>(obj);
            QMouseEvent * me = ( QMouseEvent * ) ev;
            SubControl oldHoverSubControl = hoverSubControl;
            hoverSubControl = querySubControl( CC_Slider, widget, me->pos() );
            if( oldHoverSubControl != hoverSubControl ) {
                bool repaint = false;
                if( oldHoverSubControl==SC_SliderHandle )
                    repaint = true;
                if( hoverSubControl==SC_SliderHandle )
                    repaint = true;
                if( repaint )
                    widget->update(querySubControlMetrics(CC_Slider, widget, SC_SliderHandle));
            }
        }
        else if ((ev->type() == QEvent::Leave) && (obj == hoverWidget) )
        {
            QWidget* widget = static_cast<QWidget*>(obj);
            SubControl oldHoverSubControl = hoverSubControl;
            hoverSubControl = SC_None;
            hoverWidget = 0;
            if( oldHoverSubControl==SC_SliderHandle )
                widget->update(querySubControlMetrics(CC_Slider, widget, SC_SliderHandle));
        }
    } else if ( _mouseOver && ::qt_cast<QHeader*>(obj) ) {
        QHeader * hw = ( QHeader * ) obj;
        if ( ev->type() == QEvent::Enter ) {
            hoverHeaderCurrent = hw;
            hoverHeaderId = -1;
        } else if ( ev->type() == QEvent::Leave ) {
            hoverHeaderCurrent = 0;
            if ( hoverHeaderId != -1 ) {
                hw->update();
            }
            hoverHeaderId = -1;
        } else if ( ev->type() == QEvent::MouseMove ) {
            QMouseEvent * me = ( QMouseEvent * ) ev;
            int oldHeader = hoverHeaderId;
            hoverHeaderId = hw->sectionAt( me->x() );
            if ( oldHeader != hoverHeaderId ) {
                // reset old header
                bool repaint = false;
                if ( oldHeader != -1 ) {
                    repaint = true;
                }
                if ( hoverHeaderId != -1 ) {
                    repaint = true;
                }
                if( repaint )
                    hw->update();
            }
        }
    }

    // fix background of widgets which are plugged into the toolbar (e.g. QSlider in Juk)
    if ( !qstrcmp(obj->name(), "kde toolbar widget") )
    {
        QWidget* lb = static_cast<QWidget*>(obj);
        if (lb->backgroundMode() == Qt::PaletteButton)
            lb->setBackgroundMode(Qt::PaletteBackground);
        lb->removeEventFilter(this);
    }

    return false;
}

void LinspireClearStyle::animTimerTimeout()
{
    for (AnimWidgetMap::iterator it = animWidgets.begin(); it != animWidgets.end(); ++it) {
        QWidget *widget = it.key();

        if ( (*it).status == Starting ) {
            if ( (*it).step < ANIMATIONSTEPS ) {
                double s = (*it).step;
                (*it).step += 2 + 0.14*s; // make it a bit smooth...
                // make sure we stay in the range...
                if( (*it).step > ANIMATIONSTEPS )
                    (*it).step = ANIMATIONSTEPS;
//                 qDebug("glowing in '%s/%s', step: %f", widget->className(), widget->name(), (*it).step );
                widget->repaint(false);
            } else {
                (*it).step = ANIMATIONSTEPS;
                (*it).status = Paused;
                // in case no other animation requires the timer currently, stop it.
                bool tmrUsed = false;
                for (AnimWidgetMap::iterator it2 = animWidgets.begin(); it2 != animWidgets.end(); ++it2) {
                    if( ((*it2).status==Starting) || ((*it2).status==Stopping) ) {
                        tmrUsed = true;
//                         break; // leave the for loop...
                    }
                }
                if( !tmrUsed && animTimer->isActive() )
                    animTimer->stop();
            }
        } else if ( (*it).status == Stopping ) {
            if ( (*it).step > 0 ) {
                double s = (*it).step;
                (*it).step -= 2 + 0.1*s; // make it a bit smooth...
                // make sure we stay in the range...
                if( (*it).step < 0 )
                    (*it).step = 0;
                widget->repaint(false);
            } else {
                (*it).step = 0;
                (*it).status = Stopped;
                // in case no other animation requires the timer currently, stop it.
                bool tmrUsed = false;
                for (AnimWidgetMap::iterator it2 = animWidgets.begin(); it2 != animWidgets.end(); ++it2) {
                    if( ((*it2).status==Starting) || ((*it2).status==Stopping) ) {
                        tmrUsed = true;
//                         break; // leave the for loop...
                    }
                }
                if( !tmrUsed && animTimer->isActive() )
                    animTimer->stop();
            }
        }
    }
}

QColor LinspireClearStyle::getColor(const QColorGroup &cg, const ColorType t, const bool enabled)const
{
    return getColor(cg, t, enabled?IsEnabled:IsDisabled);
}

QColor LinspireClearStyle::getColor(const QColorGroup &cg, const ColorType t, const WidgetState s)const
{
    const bool enabled = (s != IsDisabled) &&
            ((s == IsEnabled) || (s == IsPressed) || (s == IsHighlighted));
    const bool pressed = (s == IsPressed);
    const bool highlighted = (s == IsHighlighted);
    switch(t) {
        case ButtonContour:
            if(enabled) {
                if(pressed)
                    return hsvRelative(cg.button(), 0, + /*6**/_contrast, - 14*_contrast);
                else
                    return hsvRelative(cg.button(), 0, 0/*+ 6*_contrast*/, - 8*_contrast);
            } else {
                return hsvRelative(cg.button(), 0, - 7*_contrast, - 9*_contrast);
            }
        case ButtonSurface: {
            if(enabled) {
                    return cg.button();
            } else {
                return hsvRelative(cg.button(), 0, - 7*_contrast, 0 );
            }
        }
//         case LightButtonContour:
//             return hsvRelative(cg.button(), 0, + 3*_contrast, - 4*_contrast);
//         case LightButtonSurface:
//             return hsvRelative(cg.button(), 0, - 4*_contrast, - 2*_contrast);
        case DragButtonContour:
            if(enabled) {
                if(pressed)
                    return hsvRelative(cg.button(), 0, + 2*_contrast, - 14*_contrast).light(100+_contrast);
//                 else if( highlighted )
                else
                    return hsvRelative(cg.button(), 0, /*+ 6**/_contrast, - 8*_contrast);
            } else {
                return hsvRelative(cg.button(), 0, - 7*_contrast, - 9*_contrast);
            }
        case DragButtonSurface: {
            if(enabled) {
                if(pressed)
                    return cg.button().dark(100+2*_contrast);
                else if(highlighted)
                    return cg.button().light(100+(int)ANIMATIONSTEPS);
                else
                    return cg.button();                      // normal
            } else {
                return hsvRelative(cg.button(), 0, - 7*_contrast, 0 );
            }
        }
        case PanelContour: // fall through, the same as TabPanelContour for now
        case TabPanelContour:
            return hsvRelative(cg.button(), 0, 0/*+ 6*_contrast*/, - 8*_contrast);
        case InactiveTabContour: {
            // saturation a little bit more than the background, less v...
            int h1, h2, h3, s1, s2, s3, v1, v2, v3;
            cg.button().getHsv(&h1, &s1, &v1);
            cg.background().getHsv(&h2, &s2, &v2);
            h3 = h1; s3 = s2+2*_contrast; v3 = v1-11*_contrast;
            // make sure the hsv value stays in range...
            if(h3 < 0) h3 = 0;
            else if(h3 > 359) h3 = 359;
            if(s3 < 0) s3 = 0;
            else if(s3 > 255) s3 = 255;
            if(v3 < 0) v3 = 0;
            else if(v3 > 255) v3 = 255;
            QColor result;
            result.setHsv(h3, s3, v3);
            return result;
        }
        case TabSurfaceInactive: {
            // saturation a little bit more than the background, h+v stay the button ones...
            int h1, h2, h3, s1, s2, s3, v1, v2, v3;
            cg.button().getHsv(&h1, &s1, &v1);
            cg.background().getHsv(&h2, &s2, &v2);
            h3 = h1; s3 = s2+2*_contrast; v3 = v1-2*_contrast;
            // make sure the hsv value stays in range...
            if(h3 < 0) h3 = 0;
            else if(h3 > 359) h3 = 359;
            if(s3 < 0) s3 = 0;
            else if(s3 > 255) s3 = 255;
            if(v3 < 0) v3 = 0;
            else if(v3 > 255) v3 = 255;
            QColor result;
            result.setHsv(h3, s3, v3);
            if(highlighted)
                return result.dark(100+_contrast);
            else
                return result;
        }
        case ButtonSurfaceInactive: {
            // saturation a little bit more than the background, h+v stay the button ones...
            int h1, h2, h3, s1, s2, s3, v1, v2, v3;
            cg.button().getHsv(&h1, &s1, &v1);
            cg.background().getHsv(&h2, &s2, &v2);
            h3 = h1; s3 = s2+/*2**/_contrast; v3 = v1/*-2*_contrast*/;
            // make sure the hsv value stays in range...
            if(h3 < 0) h3 = 0;
            else if(h3 > 359) h3 = 359;
            if(s3 < 0) s3 = 0;
            else if(s3 > 255) s3 = 255;
            if(v3 < 0) v3 = 0;
            else if(v3 > 255) v3 = 255;
            QColor result;
            result.setHsv(h3, s3, v3);
            if(highlighted)
                return result.dark(100+2*_contrast);
            else
                return result;
        }
        case PanelLight:
            return hsvRelative(cg.button(), 0, -5*_contrast/*, +3*_contrast*/).light(100+2*_contrast);
        case PanelMid:
            return hsvRelative(cg.button(), 0, -5*_contrast/*, 0*/);
        case PanelDark:
            return hsvRelative(cg.button(), 0, -5*_contrast/*, -3*_contrast*/).dark(100+2*_contrast);

        case ButtonArrow:
            return hsvRelative(cg.button(), 0, + 10*_contrast, - 17*_contrast);

//         case PanelDark:
//             return alphaBlendColors(cg.background(), cg.background().dark(120+_contrast*5), 110);
//         case PanelDark2:
//             return alphaBlendColors(cg.background(), cg.background().dark(110+_contrast*5), 110);
//         case PanelLight:
//             return alphaBlendColors(cg.background(), cg.background().light(120+_contrast*5), 110);
//         case PanelLight2:
//             return alphaBlendColors(cg.background(), cg.background().light(110+_contrast*5), 110);
        default:
            return cg.background();
    }
}
